@@ -109,6 +109,12 @@
 
 #undef HAVE_VSNPRINTF_DECL
 
+#undef HAVE_STRCASECMP_DECL
+
+#undef HAVE_STRPTIME_DECL
+
+#undef HAVE_RANDOM_DECL
+
 /* Define to `int' if <sys/socket.h> doesn't define.  */
 #undef socklen_t
 
@@ -124,6 +130,9 @@
 /* build modular lftp */
 #undef WITH_MODULES
 
+/* md5 support is present in libc or libcrypt */
+#undef HAVE_MD5
+
 @BOTTOM@
 
 #if !defined(HAVE_LSTAT)
@@ -210,4 +219,8 @@ CDECL void SOCKSinit(const char *);
 # include <socks.h>
 #endif
 
+#if defined __MSDOS__ || defined __CYGWIN32__
+# define NATIVE_CRLF 1
+#endif
+
 #endif /* CONFIG_H */
