@@ -18,43 +18,34 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: CatJob.h,v 1.4 1999/05/31 12:34:48 lav Exp $ */
+/* $Id: CatJob.h,v 1.6 1999/12/22 20:06:55 lav Exp $ */
 
 #ifndef CATJOB_H
 #define CATJOB_H
 
-#include "XferJob.h"
+#include "CopyJob.h"
 
 class ArgV;
 class FDStream;
 
-class CatJob : public XferJob
+class CatJob : public CopyJobEnv
 {
-   void Init();
-
 protected:
    FDStream *global;
-   FDStream *local;
-   FDStream *filter_wait;
-
-   ArgV	 *args;
-
-   void	 NextFile();
-
    const char *for_each;
+   bool ascii;
+
+   void	NextFile();
 
 public:
-   int	 Do();
-   int	 Done();
-   int	 ExitCode() { return failed; }
-   int	 AcceptSig(int sig);
-
-   void	 ShowRunStatus(StatusLine *);
-   void	 PrintStatus(int);
+   int Do();
+   int Done();
+   int AcceptSig(int sig);
 
    CatJob(FileAccess *s,FDStream *global,ArgV *args);
-   CatJob(FDStream *global,char *buffer,int size);
-   virtual ~CatJob();
+   ~CatJob();
+
+   void Ascii() { ascii=true; }
 };
 
 #endif /* CATJOB_H */
