@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: CmdExec.cc,v 1.39.2.1 2000/03/09 14:17:19 lav Exp $ */
+/* $Id: CmdExec.cc,v 1.59.2.1 2000/06/08 10:13:29 lav Exp $ */
 
 #include <config.h>
 #include <errno.h>
@@ -40,6 +40,7 @@
 #define super SessionJob
 
 static ResDecl
+   res_default_proto	   ("cmd:default-protocol","ftp",0,0),
    res_long_running	   ("cmd:long-running",	"30",ResMgr::UNumberValidate,0),
    res_remote_completion   ("cmd:remote-completion","on",ResMgr::BoolValidate,0),
    res_prompt		   ("cmd:prompt",	"lftp> ",0,0),
@@ -49,7 +50,8 @@ static ResDecl
    res_verify_host	   ("cmd:verify-host",	"yes",ResMgr::BoolValidate,0),
    res_at_exit		   ("cmd:at-exit",	"",   0,0),
    res_fail_exit	   ("cmd:fail-exit",	"no", ResMgr::BoolValidate,0),
-   res_verbose		   ("cmd:verbose",	"no", ResMgr::BoolValidate,0);
+   res_verbose		   ("cmd:verbose",	"no", ResMgr::BoolValidate,0),
+   res_interactive	   ("cmd:interactive",	"no", ResMgr::BoolValidate,0);
 
 CmdExec	 *CmdExec::cwd_owner=0;
 CmdExec	 *CmdExec::chain=0;
@@ -93,6 +95,7 @@ int  CmdExec::RestoreCWD()
       eprintf(_("Warning: chdir(%s) failed: %s\n"),cwd,strerror(errno));
    fail:
       // can't run further commands in wrong directory
+      eprintf("No directory to execute commands in - terminating\n");
       while(!Done())
 	 RemoveFeeder();
       exit_code=1;
@@ -238,7 +241,6 @@ restart:
       }
       if(waiting)
       {
-	 waiting->parent=this;
 	 if(waiting->jobno<0)
 	    waiting->AllocJobno();
 	 if(cmd && waiting->cmdline==0)
@@ -246,33 +248,30 @@ restart:
 	    waiting->cmdline=cmd;
 	    cmd=0;
       	 }
-	 if(fg && !background)
-	    waiting->Fg();
+	 waiting->SetParentFg(this,!background);
       }
       if(background)
       {
-	 exit_code=0;
 	 if(waiting)
 	 {
-	    while(waiting->Do()==MOVED)
-	       ;
+	    Roll(waiting);
 	    if(!waiting->Done())
-	    {
-	       waiting->Bg();
-	       if(interactive)
-	       {
-		  printf("[%d] %s &\n",waiting->jobno,
-		     waiting->cmdline?waiting->cmdline:"?");
-		  waiting->PrintStatus(1);
-	       }
-	       last_bg=waiting->jobno;
-	       waiting=0;
-	    }
+	       SuspendJob();
 	 }
       } // background
    }
 }
 
+void CmdExec::SuspendJob()
+{
+   waiting->Bg();
+   if(interactive)
+      waiting->ListOneJob(0,0,"&");
+   last_bg=waiting->jobno;
+   exit_code=0;
+   waiting=0;
+}
+
 void CmdExec::ExecParsed(ArgV *a,FDStream *o,bool b)
 {
    bool old_running=running;
@@ -281,8 +280,7 @@ void CmdExec::ExecParsed(ArgV *a,FDStream *o,bool b)
    if(args)
       delete args;
    args=a;
-   if(cmd)
-      free(cmd);
+   xfree(cmd);
    cmd=args->Combine();
    if(output)
       delete output;
@@ -324,6 +322,20 @@ void CmdExec::RemoveFeeder()
       cwd_history.Set(session,session->GetCwd());
 }
 
+void CmdExec::ReuseSavedSession()
+{
+   Reuse(saved_session);
+   saved_session=0;
+}
+void CmdExec::RevertToSavedSession()
+{
+   if(saved_session==0)
+      return;
+   Reuse(session);
+   session=saved_session;
+   saved_session=0;
+}
+
 int CmdExec::Do()
 {
    int m=STALL;
@@ -375,6 +387,7 @@ int CmdExec::Do()
 	    if(status_line)
 	       status_line->Clear();
 	    session->Close();
+	    ReuseSavedSession();
 	    waiting=0;
 	    builtin=BUILTIN_NONE;
 	    beep_if_long();
@@ -388,6 +401,7 @@ int CmdExec::Do()
 	       status_line->Clear();
 	    eprintf("%s: %s\n",args->getarg(0),session->StrError(res));
 	    session->Close();
+	    RevertToSavedSession();
 	    waiting=0;
 	    builtin=BUILTIN_NONE;
 	    beep_if_long();
@@ -433,9 +447,7 @@ int CmdExec::Do()
 	       exec_parsed_command();
 	       return MOVED;
 	    }
-	    glob=session->MakeGlob(pat);
-	    if(!glob)
-	       glob=new NoGlob(pat);
+	    glob=new GlobURL(session,pat);
 	    m=MOVED;
 	 }
 	 break;
@@ -461,7 +473,9 @@ int CmdExec::Do()
 	       if(builtin==BUILTIN_CD)
 	       {
 		  // accept the path
-		  session->Chdir(session->GetFile(),false);
+		  const char *f=session->GetFile();
+		  f=alloca_strdup(f);
+		  session->Chdir(f,false);
 	       }
 	       session->Close();
 	       exit_code=0;
@@ -476,13 +490,13 @@ int CmdExec::Do()
 	 }
 	 if(SignalHook::GetCount(SIGHUP))
 	 {
-	    interactive=0;
+	    SetInteractive(false);
 	    return MOVED;
 	 }
       }
       if(status_line)
 	 waiting->ShowRunStatus(status_line);
-      return STALL;
+      return m;
    }
 
    if(waiting)
@@ -495,7 +509,7 @@ int CmdExec::Do()
  	 if(interactive || verbose)
 	    waiting->SayFinal(); // final phrase like 'rm succeed'
 	 exit_code=waiting->ExitCode();
-	 delete waiting;
+	 Delete(waiting);
 	 waiting=0;
 	 beep_if_long();
       	 return MOVED;
@@ -512,24 +526,18 @@ int CmdExec::Do()
 	 }
 	 if(SignalHook::GetCount(SIGTSTP))
 	 {
-	    waiting->Bg();
-	    if(status_line)
-	       status_line->WriteLine("[%d] %s &",waiting->jobno,waiting->cmdline);
-	    waiting->PrintStatus(1);
-	    exit_code=0;
-	    last_bg=waiting->jobno;
-	    waiting=0;
+	    SuspendJob();
 	    return MOVED;
 	 }
 	 if(SignalHook::GetCount(SIGHUP))
 	 {
-	    interactive=0;
+	    SetInteractive(false);
 	    return MOVED;
 	 }
       }
       if(status_line)
 	 waiting->ShowRunStatus(status_line);
-      return STALL;
+      return m;
    }
 
    if(!interactive)
@@ -644,17 +652,23 @@ void CmdExec::ShowRunStatus(StatusLine *s)
 void CmdExec::PrintStatus(int v)
 {
    SessionJob::PrintStatus(v);
-   if(v<1)
-      return;
    if(waiting==this)
    {
       char *s=args->Combine();
       printf(_("\tExecuting builtin `%s' [%s]\n"),s,session->CurrentStatus());
-      free(s);
+      xfree(s);
       return;
    }
    if(is_queue)
    {
+      if(waiting)
+      {
+	 printf("\t%s ",_("Now executing:"));
+	 if(v==0)
+	    waiting->ListOneJob(v);
+	 else
+	    waiting->PrintJobTitle();
+      }
       if(!(next_cmd && next_cmd[0]))
 	 return;
       printf(_("\tCommands queued:\n"));
@@ -671,6 +685,11 @@ void CmdExec::PrintStatus(int v)
 	 cmd=cmd_end+1;
 	 if(*cmd==0)
 	    break;
+	 if(v<2 && n>4)
+	 {
+	    printf("\t%2d. ...\n",n);
+	    break;
+	 }
       }
       return;
    }
@@ -696,6 +715,7 @@ CmdExec::CmdExec(FileAccess *f) : SessionJob(f)
    background=false;
 
    interactive=false;
+   top_level=false;
    status_line=0;
    feeder=0;
    feeder_called=false;
@@ -714,14 +734,11 @@ CmdExec::CmdExec(FileAccess *f) : SessionJob(f)
    SaveCWD();
 
    var_ls=xstrdup("");
-   var_prompt=xstrdup("lftp> ");
    remote_completion=false;
    long_running=0;
    csh_history=false;
    verify_host=verify_path=true;
 
-   Reconfig();
-
    start_time=0;
    old_cwd=0;
    old_lcwd=0;
@@ -732,6 +749,10 @@ CmdExec::CmdExec(FileAccess *f) : SessionJob(f)
    is_queue=false;
    queue_cwd=0;
    queue_lcwd=0;
+
+   saved_session=0;
+
+   Reconfig();
 }
 
 CmdExec::~CmdExec()
@@ -765,6 +786,8 @@ CmdExec::~CmdExec()
 
    xfree(queue_cwd);
    xfree(queue_lcwd);
+
+   Reuse(saved_session);
 }
 
 char *CmdExec::MakePrompt()
@@ -782,7 +805,7 @@ char *CmdExec::MakePrompt()
 
    char *store=prompt;
 
-   char *scan=var_prompt;
+   const char *scan=res_prompt.Query(getenv("TERM"));
    char ch;
    char str[3];
    const char *to_add;
@@ -895,11 +918,14 @@ char *CmdExec::MakePrompt()
 	 continue;
 
       int store_index=store-prompt;
-      int need=store_index+(int)strlen(to_add)+1;
-      while(prompt_size<need)
-	 prompt_size*=2;
-      prompt=(char*)xrealloc(prompt,prompt_size);
-      store=prompt+store_index;
+      int need=store_index+strlen(to_add)+1;
+      if(prompt_size<need)
+      {
+	 while(prompt_size<need)
+	    prompt_size*=2;
+	 prompt=(char*)xrealloc(prompt,prompt_size);
+	 store=prompt+store_index;
+      }
 
       strcpy(store,to_add);
       store+=strlen(to_add);
@@ -915,7 +941,7 @@ void CmdExec::beep_if_long()
       write(1,"\007",1);
 }
 
-void CmdExec::Reconfig(const char *)
+void CmdExec::Reconfig(const char *name)
 {
    const char *c=0;
    if(session)
@@ -926,11 +952,12 @@ void CmdExec::Reconfig(const char *)
    csh_history = res_csh_history.Query(0);
    xfree(var_ls);
    var_ls=xstrdup(res_default_ls.Query(c));
-   xfree(var_prompt);
-   var_prompt=xstrdup(res_prompt.Query(c));
    verify_path=res_verify_path.Query(c);
    verify_host=res_verify_host.Query(c);
    verbose=res_verbose.Query(0);
+   // only allow explicit setting of cmd:interactive to change interactiveness.
+   if(top_level && name && !strcmp(name,"cmd:interactive"))
+      SetInteractive(res_interactive.Query(0));
 }
 
 void CmdExec::pre_stdout()
@@ -939,6 +966,7 @@ void CmdExec::pre_stdout()
       status_line->Clear();
    if(feeder_called)
       feeder->clear();
+   current->TimeoutS(1);
 }
 
 void CmdExec::top_vfprintf(FILE *file,const char *f,va_list v)
@@ -965,8 +993,11 @@ int CmdExec::AcceptSig(int sig)
       switch(builtin)
       {
       case(BUILTIN_CD):
+	 session->Close();
+	 break;
       case(BUILTIN_OPEN):
 	 session->Close();
+	 RevertToSavedSession();
 	 break;
       case(BUILTIN_GLOB):
 	 delete glob;
@@ -989,10 +1020,13 @@ int CmdExec::AcceptSig(int sig)
       if(res==WANTDIE)
       {
 	 exit_code=1;
-	 Job *new_waiting=waiting->waiting;
-	 waiting->waiting=0;
-	 delete waiting;
-	 waiting=new_waiting;
+	 int j=-1;
+	 if(waiting->waiting)
+	    j=waiting->waiting->jobno;
+	 Delete(waiting);
+	 waiting=0;
+	 if(j>=0)
+	    waiting=FindJob(j);
       }
       if(waiting==0 && parent!=0)
 	 return WANTDIE;
@@ -1158,6 +1192,7 @@ void CmdExec::ChangeSession(FileAccess *new_session)
 {
    Reuse(session);
    session=new_session;
+   session->SetPriority(fg?1:0);
    Reconfig(0);
 }
 
