@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: CmdExec.h,v 1.29 1999/08/31 13:58:02 lav Exp $ */
+/* $Id: CmdExec.h,v 1.33.2.1 2000/06/08 10:13:30 lav Exp $ */
 
 #ifndef CMDEXEC_H
 #define CMDEXEC_H
@@ -120,7 +120,7 @@ private:
    char *old_cwd;
    char *old_lcwd;
 
-   Glob *glob;
+   GlobURL *glob;
    ArgV *args_glob;
 
    static CmdExec *chain;
@@ -131,6 +131,10 @@ private:
    char *queue_lcwd;
    CmdExec *FindQueue();
 
+   FileAccess *saved_session;
+   void ReuseSavedSession();
+   void RevertToSavedSession();
+
 public:
    void FeedCmd(const char *c);
    void FeedArgV(const ArgV *,int start=0);
@@ -141,6 +145,8 @@ public:
    void FeedQuoted(const char *c);
    void AtExit();
 
+   void SuspendJob();
+
    CmdExec(FileAccess *s);
    ~CmdExec();
 
@@ -155,6 +161,7 @@ public:
    char *MakePrompt();
 
    bool interactive;
+   bool top_level;
    bool verbose;
    StatusLine *status_line;
    void SetCmdFeeder(CmdFeeder *new_feeder);
@@ -164,7 +171,6 @@ public:
    static const char *GetFullCommandName(const char *);
 
    char	 *var_ls;
-   char	 *var_prompt;
    bool	 remote_completion;
    int	 long_running;
    bool	 csh_history;
@@ -187,6 +193,11 @@ public:
    void top_vfprintf(FILE *file,const char *f,va_list v);
 
    void SetInteractive(bool i);
+   void SetTopLevel()
+      {
+	 top_level=true;
+	 Reconfig(0);
+      }
 
    static void RegisterCommand(const char *name,cmd_creator_t creator,
       const char *short_name=0,const char *long_name=0);
