@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileAccess.cc,v 1.30.2.1 2000/02/20 14:51:56 lav Exp $ */
+/* $Id: FileAccess.cc,v 1.61.2.1 2000/07/31 16:42:19 lav Exp $ */
 
 #include <config.h>
 
@@ -27,9 +27,11 @@
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
-#include <netinet/tcp.h>
+#ifdef HAVE_NETINET_TCP_H
+# include <netinet/tcp.h>
+#endif
 #include <errno.h>
-#include <ctype.h>
+#include "ascii_ctype.h"
 #include <fcntl.h>
 #include <fnmatch.h>
 #include "LsCache.h"
@@ -37,6 +39,9 @@
 #include "log.h"
 #include "url.h"
 #include "misc.h"
+#include "DummyProto.h"
+#include "netrc.h"
+#include "ArgV.h"
 #ifdef WITH_MODULES
 # include "module.h"
 #endif
@@ -50,6 +55,7 @@ void FileAccess::Init()
    hostname=0;
    group=gpass=0;
    user=pass=0;
+   pass_open=false;
 
    file=0;
    file1=0;
@@ -68,16 +74,19 @@ void FileAccess::Init()
    saved_errno=0;
    mkdir_p=false;
    ascii=false;
+   norest_manual=false;
 
    url=0;
 
-   entity_size=-1;
-   entity_date=(time_t)-1;
+   entity_size=NO_SIZE;
+   entity_date=NO_DATE;
 
    closure=0;
 
    chmod_mode=0644;
 
+   priority=0;
+
    next=chain;
    chain=this;
 }
@@ -94,6 +103,7 @@ FileAccess::FileAccess(const FileAccess *fa)
    user=xstrdup(fa->user);
    xfree(pass);
    pass=xstrdup(fa->pass);
+   pass_open=fa->pass_open;
    xfree(group);
    group=xstrdup(fa->group);
    xfree(gpass);
@@ -209,13 +219,13 @@ int   FileAccess::CheckHangup(struct pollfd *pfd,int num)
       {
 	 sprintf(str,_("Socket error (%s) - reconnecting"),
 				    strerror(errno?errno:s_errno));
-	 DebugPrint("**** ",str);
+	 DebugPrint("**** ",str,0);
 	 return 1;
       }
 #endif /* SO_ERROR */
       if(pfd[i].revents&POLLERR)
       {
-	 DebugPrint("**** ","POLLERR");
+	 DebugPrint("**** ","POLLERR",0);
 	 return 1;
       }
    } /* end for */
@@ -236,7 +246,11 @@ void FileAccess::CloseOnExec(int fd)
 
 void  FileAccess::Open(const char *fn,int mode,long offs)
 {
-   Close();
+#ifdef OPEN_DEBUG
+   printf("FA::Open(%s)\n",fn);
+#endif
+   if(IsOpen())
+      Close();
    Resume();
    file=xstrdup(fn);
    real_pos=-1;
@@ -244,6 +258,7 @@ void  FileAccess::Open(const char *fn,int mode,long offs)
    this->mode=mode;
    mkdir_p=false;
    try_time=0;
+   Timeout(0);
 
    switch((open_mode)mode)
    {
@@ -326,16 +341,18 @@ void FileAccess::Close()
    opt_date=0;
    opt_size=0;
    array_for_info=0;
-   entity_size=-1;
-   entity_date=(time_t)-1;
+   entity_size=NO_SIZE;
+   entity_date=NO_DATE;
    ascii=false;
+   norest_manual=false;
    ClearError();
 }
 
 void FileAccess::Rename(const char *f,const char *f1)
 {
-   Open(f,RENAME);
+   Close();
    file1=xstrdup(f1);
+   Open(f,RENAME);
 
    LsCache::TreeChanged(this,file);
    LsCache::FileChanged(this,file);
@@ -365,6 +382,7 @@ const char *FileAccess::GetFileURL(const char *f,int flags)
    const char *proto=GetProto();
    if(proto[0]==0)
       return "";
+
    len+=strlen(proto)+strlen("://");
    if(user)
    {
@@ -376,33 +394,24 @@ const char *FileAccess::GetFileURL(const char *f,int flags)
       len+=strlen(hostname)*3;
    if(portname)
       len+=1+strlen(portname)*3;
-   f=dir_file(cwd?cwd:"~",f);
+   if(!f || (f[0]!='/' && f[0]!='~'))
+      f=dir_file(cwd?cwd:"~",f);
    len+=1+strlen(f)*3;
    url=(char*)xrealloc(url,len);
-   sprintf(url,"%s://",proto);
-   if(user)
-   {
-      url::encode_string(user,url+strlen(url),"/:@"URL_UNSAFE);
-      if(pass && (flags&WITH_PASSWORD))
-      {
-	 strcat(url,":");
-	 url::encode_string(pass,url+strlen(url),"/:@"URL_UNSAFE);
-      }
-      strcat(url,"@");
-   }
-   if(hostname)
-      url::encode_string(hostname,url+strlen(url),"/:"URL_UNSAFE);
-   if(portname)
-   {
-      strcat(url,":");
-      url::encode_string(portname,url+strlen(url),"/"URL_UNSAFE);
-   }
-   if(strcmp(f,"~") && !(flags&NO_PATH))
-   {
-      if(f[0]!='/') // e.g. ~/path
-	 strcat(url,"/");
-      url::encode_string(f,url+strlen(url));
-   }
+
+   ParsedURL u("");
+
+   u.proto=(char*)proto;
+   u.user=user;
+   if((pass_open || (flags&WITH_PASSWORD)) && !(flags&NO_PASSWORD))
+      u.pass=pass;
+   u.host=hostname;
+   u.port=portname;
+   if(!(flags&NO_PATH))
+      u.path=(char*)f;
+
+   u.Combine(url,home);
+
    return url;
 }
 
@@ -421,7 +430,6 @@ void FileAccess::Connect(const char *host1,const char *port1)
    xfree(cwd);
    cwd=xstrdup(default_cwd);
    xfree(home); home=0;
-   try_time=0;
    DontSleep();
 }
 
@@ -432,10 +440,54 @@ void FileAccess::Login(const char *user1,const char *pass1)
    user=xstrdup(user1);
    xfree(pass);
    pass=xstrdup(pass1);
+   pass_open=false;
    xfree(cwd);
    cwd=xstrdup(default_cwd);
    xfree(home);
    home=0;
+
+   if(user && pass==0)
+   {
+      for(FileAccess *o=chain; o!=0; o=o->next)
+      {
+	 pass=o->pass;
+	 if(SameSiteAs(o) && o->pass)
+	 {
+	    pass=xstrdup(o->pass);
+	    break;
+	 }
+	 pass=0;
+      }
+      if(pass==0 && hostname) // still no pass? Try .netrc
+      {
+	 NetRC::Entry *nrc=NetRC::LookupHost(hostname);
+	 if(nrc)
+	 {
+	    if(nrc->user && !strcmp(nrc->user,user))
+	       pass=xstrdup(nrc->pass);
+	 }
+      }
+   }
+}
+
+void FileAccess::SetPasswordGlobal(const char *p)
+{
+   xfree(pass);
+   pass=xstrdup(p);
+   for(FileAccess *o=chain; o!=0; o=o->next)
+   {
+      if(o==this)
+	 continue;
+      char *save_pass=o->pass;	 // cheat SameSiteAs.
+      o->pass=pass;
+      if(SameSiteAs(o))
+      {
+	 xfree(save_pass);
+	 o->pass=xstrdup(pass);
+      }
+      else
+	 o->pass=save_pass;
+   }
 }
 
 void FileAccess::GroupLogin(const char *group,const char *pass)
@@ -452,6 +504,7 @@ void FileAccess::AnonymousLogin()
    Disconnect();
    xfree(user); user=0;
    xfree(pass); pass=0;
+   pass_open=false;
    xfree(group); group=0;
    xfree(gpass); gpass=0;
    xfree(cwd);
@@ -492,11 +545,29 @@ void  FileAccess::ExpandTildeInCWD()
 	 expand_tilde(&cwd,home);
       if(real_cwd)
 	 expand_tilde(&real_cwd,home);
-      if(mode==CHANGE_DIR && file)
+      if(file)
 	 expand_tilde(&file,home);
+      if(file1)
+	 expand_tilde(&file1,home);
    }
 }
 
+const char *FileAccess::ExpandTildeStatic(const char *s)
+{
+   if(!home || !(s[0]=='~' && (s[1]=='/' || s[1]==0)))
+      return s;
+
+   static char *buf=0;
+   static int buf_len=0;
+
+   int len=strlen(s)+1;
+   if(len>buf_len)
+      buf=(char*)xrealloc(buf,buf_len=len);
+   strcpy(buf,s);
+   expand_tilde(&buf,home);
+   return buf;
+}
+
 static inline
 int last_element_is_doubledot(char *path,char *end)
 {
@@ -504,6 +575,16 @@ int last_element_is_doubledot(char *path,char *end)
         || (end>path+2 && !strncmp(end-3,"/..",3)));
 }
 
+int FileAccess::device_prefix_len(const char *path)
+{
+   int i=0;
+   while(path[i] && (is_ascii_alnum(path[i]) || strchr("$_-",path[i])))
+      i++;
+   if(i>0 && path[i]==':')
+      return i+1+(path[i+1]=='/');
+   return 0;
+}
+
 void FileAccess::Chdir(const char *path,bool verify)
 {
    char	 *newcwd=(char*)alloca(strlen(cwd)+strlen(path)+2);
@@ -513,25 +594,25 @@ void FileAccess::Chdir(const char *path,bool verify)
 
    if(cwd[0]=='/')
       prefix_size=1;
-   else if(isalpha(cwd[0]) && cwd[1]==':')
-      prefix_size=2+(cwd[2]=='/');
    else if(cwd[0]=='~')
    {
       prefix_size=1;
       while(cwd[prefix_size]!='/' && cwd[prefix_size]!='\0')
 	 prefix_size++;
    }
+   else
+   {
+      // handle VMS and DOS devices.
+      prefix_size=device_prefix_len(cwd);
+   }
+
+   int dev_prefix=0;
 
    if(path[0]=='/')
    {
       strcpy(newcwd,path);
       prefix_size=1;
    }
-   else if(isalpha(path[0]) && path[1]==':')
-   {
-      strcpy(newcwd,path);
-      prefix_size=2+(path[2]=='/');
-   }
    else if(path[0]=='~')
    {
       strcpy(newcwd,path);
@@ -539,6 +620,11 @@ void FileAccess::Chdir(const char *path,bool verify)
       while(newcwd[prefix_size]!='/' && newcwd[prefix_size]!='\0')
 	 prefix_size++;
    }
+   else if((dev_prefix=device_prefix_len(path))>0)
+   {
+      strcpy(newcwd,path);
+      prefix_size=dev_prefix;
+   }
    else
    {
       if(cwd[0])
@@ -629,6 +715,8 @@ void FileAccess::Chmod(const char *file,int m)
 
 void FileAccess::SetError(int ec,const char *e)
 {
+   if(ec==SEE_ERRNO)
+      saved_errno=errno;
    xfree(error);
    error=xstrdup(e);
    error_code=ec;
@@ -651,12 +739,15 @@ FileAccess *SessionPool::pool[pool_size];
 
 void SessionPool::Reuse(FileAccess *f)
 {
+   if(f==0)
+      return;
    if(f->GetHostName()==0)
    {
-      delete f;
+      SMTask::Delete(f);
       return;
    }
    f->Close();
+   f->SetPriority(0);
    int i;
    for(i=0; i<pool_size; i++)
    {
@@ -670,12 +761,12 @@ void SessionPool::Reuse(FileAccess *f)
    {
       if(f->IsBetterThan(pool[i]))
       {
-	 delete pool[i];
+	 SMTask::Delete(pool[i]);
 	 pool[i]=f;
 	 return;
       }
    }
-   delete f;
+   SMTask::Delete(f);
 }
 
 void SessionPool::Print(FILE *f)
@@ -767,8 +858,31 @@ ResValue FileAccess::Query(const char *name,const char *closure)
    return ResMgr::Query(fullname,closure);
 }
 
-void FileAccess::Reconfig(const char *)
+bool FileAccess::IsBetterThan(FileAccess *fa)
 {
+   return(SameProtoAs(fa) && this->IsConnected() > fa->IsConnected());
+}
+
+void FileAccess::Reconfig(const char *) {}
+void FileAccess::ConnectVerify() {}
+const char *FileAccess::CurrentStatus() { return ""; }
+int FileAccess::Buffered() { return 0; }
+bool FileAccess::IOReady() { return IsOpen(); }
+int FileAccess::IsConnected() { return 0; }
+void FileAccess::Disconnect() {}
+void FileAccess::UseCache(bool) {}
+bool FileAccess::NeedSizeDateBeforehand() { return false; }
+void FileAccess::Cleanup() {}
+void FileAccess::CleanupThis() {}
+ListInfo *FileAccess::MakeListInfo() { return 0; }
+Glob *FileAccess::MakeGlob(const char *pattern) { return 0; }
+DirList *FileAccess::MakeDirList(ArgV *a) { if(a) delete a; return 0; }
+
+DirList::~DirList()
+{
+   Delete(buf);
+   if(args)
+      delete args;
 }
 
 void FileAccess::CleanupAll()
@@ -777,6 +891,35 @@ void FileAccess::CleanupAll()
       o->CleanupThis();
 }
 
+FileAccess *FileAccess::NextSameSite(FA *scan)
+{
+   if(scan==0)
+      scan=chain;
+   else
+      scan=scan->next;
+   for( ; scan; scan=scan->next)
+      if(scan!=this && SameSiteAs(scan))
+	 return scan;
+   return 0;
+}
+
+
+FileAccess *FileAccess::New(const ParsedURL *u,bool dummy)
+{
+   FileAccess *s=New(u->proto);
+   if(!s)
+   {
+      if(!dummy)
+	 return 0;
+      return new DummyNoProto(u->proto);
+   }
+   s->Connect(u->host,u->port);
+   if(u->user)
+      s->Login(u->user,u->pass);
+   // path?
+   return s;
+}
+
 // FileAccess::Protocol implementation
 FileAccess::Protocol *FileAccess::Protocol::chain=0;
 
@@ -882,6 +1025,9 @@ Glob::Glob(const char *p)
    list=0;
    list_size=0;
    list_alloc=0;
+   dirs_only=false;
+   files_only=false;
+   match_period=true;
 }
 
 void Glob::free_list()
@@ -914,8 +1060,12 @@ void Glob::add(const char *ptr,int len)
    memcpy(s,ptr,len);
    s[len]=0;
 
+   int flags=FNM_PATHNAME;
+   if(match_period)
+      flags|=FNM_PERIOD;
+
    if(pattern[0]!=0
-   && fnmatch(pattern, s, FNM_PATHNAME|FNM_PERIOD)!=0)
+   && fnmatch(pattern, s, flags)!=0)
       return; // unmatched
 
    add_force(ptr,len);
@@ -933,6 +1083,7 @@ bool Glob::HasWildcards(const char *s)
 	 break;
       case '*':
       case '[':
+      case ']':
       case '?':
 	 return true;
       }
@@ -950,6 +1101,7 @@ void Glob::UnquoteWildcards(char *s)
       {
 	 if(s[1]=='*'
 	 || s[1]=='['
+	 || s[1]==']'
 	 || s[1]=='?'
 	 || s[1]=='\\')
 	    s++;
@@ -966,6 +1118,12 @@ int NoGlob::Do()
 {
    if(!done)
    {
+      if(!HasWildcards(pattern))
+      {
+	 char *p=alloca_strdup(pattern);
+	 UnquoteWildcards(p);
+	 add(p);
+      }
       done=true;
       return MOVED;
    }
@@ -973,9 +1131,52 @@ int NoGlob::Do()
 }
 NoGlob::NoGlob(const char *p) : Glob(p)
 {
-   char *p1=pattern;
-   pattern=xstrdup("");
-   UnquoteWildcards(p1);
-   add(p1);
-   xfree(p1);
+}
+
+GlobURL::GlobURL(FileAccess *s,const char *p)
+{
+   session=s;
+   reuse=false;
+   glob=0;
+   url_prefix=xstrdup(p);
+   url_prefix[url::path_index(p)]=0;
+
+   ParsedURL p_url(p,true);
+   if(p_url.proto && p_url.path)
+   {
+      session=FA::New(&p_url);
+      if(session)
+      {
+	 glob=session->MakeGlob(p_url.path);
+	 reuse=true;
+      }
+   }
+   else
+   {
+      glob=session->MakeGlob(p);
+   }
+   if(!glob)
+      glob=new NoGlob(p);
+}
+GlobURL::~GlobURL()
+{
+   SMTask::Delete(glob);
+   if(session && reuse)
+      SessionPool::Reuse(session);
+   xfree(url_prefix);
+}
+char **GlobURL::GetResult()
+{
+   char **list=glob->GetResult();
+   if(!list)
+      return 0;
+   if(!reuse)
+      return list;
+   for(int i=0; list[i]; i++)
+   {
+      const char *n=url_file(url_prefix,list[i]);
+      list[i]=(char*)xrealloc(list[i],strlen(n)+1);
+      strcpy(list[i],n);
+   }
+   return list;
 }
