@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileAccess.h,v 1.24 1999/10/08 15:13:28 lav Exp $ */
+/* $Id: FileAccess.h,v 1.45 2000/03/07 00:34:05 lav Exp $ */
 
 #ifndef FILEACCESS_H
 #define FILEACCESS_H
@@ -41,6 +41,11 @@
 #include "xmalloc.h"
 #include "ResMgr.h"
 
+#define NO_SIZE	     (-1L)
+#define NO_SIZE_YET  (-2L)
+#define NO_DATE	     ((time_t)-1L)
+#define NO_DATE_YET  ((time_t)-2L)
+
 class ListInfo;
 class Glob;
 class NoGlob;
@@ -83,6 +88,7 @@ protected:
    char	 *portname;
    char  *user;
    char  *pass;
+   bool	 pass_open;
    char	 *group;
    char	 *gpass;
 
@@ -103,7 +109,7 @@ protected:
    static void NonBlock(int fd);
    static void CloseOnExec(int fd);
 
-   void  DebugPrint(const char *prefix,const char *str,int level=0);
+   void  DebugPrint(const char *prefix,const char *str,int level=9);
 
    time_t   try_time;
    time_t   event_time;
@@ -118,6 +124,7 @@ protected:
    int	 saved_errno;
 
    void	 ExpandTildeInCWD();
+   const char *ExpandTildeStatic(const char *s);
 
    char *real_cwd;
    void set_real_cwd(const char *c)
@@ -136,6 +143,9 @@ protected:
 
    int chmod_mode;
    bool ascii;
+   bool norest_manual;
+
+   int	priority;   // higher priority can take over other session.
 
    bool Error() { return error_code!=OK; }
    void ClearError();
@@ -146,6 +156,12 @@ protected:
 
    FileAccess *next;
    static FileAccess *chain;
+   FileAccess *FirstSameSite() { return NextSameSite(0); }
+   FileAccess *NextSameSite(FileAccess *);
+
+   static int device_prefix_len(const char *path);
+
+   virtual ~FileAccess();
 
 public:
    virtual const char *GetProto() = 0; // http, ftp, file etc
@@ -158,10 +174,10 @@ public:
    const char  *GetPort() { return portname; }
    const char  *GetConnectURL(int flags=0);
    const char  *GetFileURL(const char *file,int flags=0);
-   enum { NO_PATH=1,WITH_PASSWORD=2 };
+   enum { NO_PATH=1,WITH_PASSWORD=2,NO_PASSWORD=4 };
 
    virtual void Connect(const char *h,const char *p);
-   virtual void ConnectVerify() {}
+   virtual void ConnectVerify();
 
    virtual void AnonymousLogin();
 
@@ -191,15 +207,17 @@ public:
 	 return array_ptr*100/array_cnt;
       }
 
-   virtual const char *CurrentStatus() { return(""); }
+   virtual const char *CurrentStatus();
 
    virtual int Read(void *buf,int size) = 0;
    virtual int Write(const void *buf,int size) = 0;
-   virtual int Buffered() { return 0; }
+   virtual int Buffered();
    virtual int StoreStatus() = 0;
+   virtual bool IOReady();
    long GetPos() { return pos; }
    long GetRealPos() { return real_pos<0?pos:real_pos; }
    void SeekReal() { pos=GetRealPos(); }
+   void RereadManual() { norest_manual=true; }
 
    const char *GetCwd() { return cwd; }
    const char *GetFile() { return file; }
@@ -209,22 +227,22 @@ public:
 
    virtual bool SameLocationAs(FileAccess *fa);
    virtual bool SameSiteAs(FileAccess *fa);
-   virtual bool IsBetterThan(FileAccess *fa) { (void)fa; return false; }
+   virtual bool IsBetterThan(FileAccess *fa);
 
    void Init();
    FileAccess() { Init(); }
    FileAccess(const FileAccess *);
-   virtual ~FileAccess();
 
    void	 DontSleep() { try_time=0; }
 
    bool	 IsClosed() { return mode==CLOSED; }
    bool	 IsOpen() { return !IsClosed(); }
+   int	 OpenMode() { return mode; }
 
-   virtual void CopyOptions(FileAccess *fa) { (void)fa; }
-   virtual bool IsConnected() { return false; }
-   virtual void Disconnect() {}
-   virtual void UseCache(bool) {}
+   virtual int  IsConnected(); // level of connection (0 - not connected).
+   virtual void Disconnect();
+   virtual void UseCache(bool);
+   virtual bool NeedSizeDateBeforehand();
 
    enum status
    {
@@ -242,14 +260,14 @@ public:
       NOT_SUPP
    };
 
-   const char *StrError(int err);
-   virtual void Cleanup() {}
-   virtual void CleanupThis() {}
+   virtual const char *StrError(int err);
+   virtual void Cleanup();
+   virtual void CleanupThis();
    void CleanupAll();
       // ^^ close idle connections, etc.
-   virtual ListInfo *MakeListInfo() { return 0; }
-   virtual Glob *MakeGlob(const char *pattern) { return 0; }
-   virtual DirList *MakeDirList(ArgV *arg) { return 0; }
+   virtual ListInfo *MakeListInfo();
+   virtual Glob *MakeGlob(const char *pattern);
+   virtual DirList *MakeDirList(ArgV *a);
 
    static bool NotSerious(int err);
 
@@ -279,6 +297,20 @@ public:
       {
 	 return Protocol::NewSession(proto);
       }
+   static FileAccess *New(const class ParsedURL *u,bool dummy=true);
+
+   void SetPasswordGlobal(const char *p);
+   void InsecurePassword(bool i)
+      {
+	 pass_open=i;
+      }
+   void SetPriority(int p)
+      {
+	 if(p==priority)
+	    return;
+	 priority=p;
+	 current->Timeout(0);
+      }
 };
 
 class FileAccessOperation : public SMTask
@@ -290,9 +322,10 @@ protected:
 
    bool use_cache;
 
+   virtual ~FileAccessOperation();
+
 public:
    FileAccessOperation();
-   virtual ~FileAccessOperation();
 
    virtual int Do() = 0;
    bool Done() { return done; }
@@ -311,15 +344,21 @@ protected:
    char	 **list;
    int	 list_size;
    int	 list_alloc;
+   bool	 dirs_only;
+   bool	 files_only;
+   bool	 match_period;
    void	 add(const char *ptr,int len);
    void	 add_force(const char *ptr,int len);
    void	 add(const char *ptr) { add(ptr,strlen(ptr)); }
    void	 free_list();
+   virtual ~Glob();
 public:
    const char *GetPattern() { return pattern; }
    char **GetResult() { return list; }
    Glob(const char *p);
-   virtual ~Glob();
+   void DirectoriesOnly() { dirs_only=true; }
+   void FilesOnly() { files_only=true; }
+   void NoMatchPeriod() { match_period=false; }
 
    static bool HasWildcards(const char *);
    static void UnquoteWildcards(char *);
@@ -332,6 +371,21 @@ public:
    int Do();
 };
 
+class GlobURL
+{
+   FileAccess *session;
+   bool reuse;
+   char *url_prefix;
+public:
+   Glob *glob;
+   GlobURL(FileAccess *s,const char *p);
+   ~GlobURL();
+   char **GetResult();
+   bool Done()  { return glob->Done(); }
+   bool Error() { return glob->Error(); }
+   const char *ErrorText() { return glob->ErrorText(); }
+   const char *Status() { return glob->Status(); }
+};
 
 #include "FileSet.h"
 
@@ -347,9 +401,10 @@ protected:
    unsigned need;
    bool follow_symlinks;
 
+   virtual ~ListInfo();
+
 public:
    ListInfo();
-   virtual ~ListInfo();
 
    virtual void SetExclude(const char *p,regex_t *x,regex_t *i);
 
@@ -374,17 +429,13 @@ class DirList : public FileAccessOperation
 protected:
    Buffer *buf;
    ArgV *args;
+   ~DirList();
 public:
    DirList(ArgV *a)
       {
 	 buf=new Buffer();
 	 args=a;
       }
-   ~DirList()
-      {
-	 if(buf)
-	    delete buf;
-      }
 
    virtual int Do() = 0;
    virtual const char *Status() = 0;
