@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileSet.cc,v 1.7 1999/09/08 16:06:54 lav Exp $ */
+/* $Id: FileSet.cc,v 1.9.2.2 2000/07/27 17:52:14 lav Exp $ */
 
 #include <config.h>
 #include "FileSet.h"
@@ -27,6 +27,7 @@
 #include <sys/stat.h>
 #include <utime.h>
 #include "misc.h"
+#include "ResMgr.h"
 
 void  FileInfo::Merge(const FileInfo& f)
 {
@@ -91,7 +92,7 @@ void FileSet::Merge(const FileSet *set)
       }
       if(j==fnum)
       {
-	 Add(set->files[i]);
+	 Add(new FileInfo(*set->files[i]));
       }
    }
 }
@@ -137,12 +138,13 @@ FileSet::~FileSet()
    xfree(files);
 }
 
-void FileSet::SubtractSame(const FileSet *set,bool only_newer,time_t prec,int ignore)
+void FileSet::SubtractSame(const FileSet *set,bool only_newer,
+      const TimeInterval *prec,const TimeInterval *loose_prec,int ignore)
 {
    for(int i=0; i<fnum; i++)
    {
       FileInfo *f=set->FindByName(files[i]->name);
-      if(f && files[i]->SameAs(f,only_newer,prec,ignore))
+      if(f && files[i]->SameAs(f,only_newer,prec,loose_prec,ignore))
 	 Sub(i--);
    }
 }
@@ -188,7 +190,8 @@ void FileSet::ExcludeDots()
    }
 }
 
-bool  FileInfo::SameAs(const FileInfo *fi,bool only_newer,time_t prec,int ignore)
+bool  FileInfo::SameAs(const FileInfo *fi,bool only_newer,
+	 const TimeInterval *prec,const TimeInterval *loose_prec,int ignore)
 {
    if(defined&NAME && fi->defined&NAME)
       if(strcmp(name,fi->name))
@@ -204,11 +207,24 @@ bool  FileInfo::SameAs(const FileInfo *fi,bool only_newer,time_t prec,int ignore
    if(defined&SYMLINK_DEF && fi->defined&SYMLINK_DEF)
       return (strcmp(symlink,fi->symlink)==0);
 
-   if(defined&DATE && fi->defined&DATE && !(ignore&DATE))
+   if(defined&(DATE|DATE_UNPREC) && fi->defined&(DATE|DATE_UNPREC)
+   && !(ignore&DATE))
    {
+      time_t p;
+      bool inf;
+      if((defined&DATE_UNPREC) || (fi->defined&DATE_UNPREC))
+      {
+	 p=loose_prec->Seconds();
+	 inf=loose_prec->IsInfty();
+      }
+      else
+      {
+	 p=prec->Seconds();
+	 inf=prec->IsInfty();
+      }
       if(only_newer && date<fi->date)
 	    return true;
-      if(abs((long)date-(long)(fi->date))>prec)
+      if(!inf && abs((long)date-(long)(fi->date))>p)
 	 return false;
    }
 
@@ -259,8 +275,8 @@ void  FileSet::Exclude(const char *prefix,regex_t *exclude,regex_t *include)
    for(int i=0; i<fnum; i++)
    {
       const char *name=dir_file(prefix,files[i]->name);
-      if(!(include && regexec(include,name,0,0,REG_NOSUB)==0)
-       && ((exclude && regexec(exclude,name,0,0,REG_NOSUB)==0)
+      if(!(include && regexec(include,name,0,0,0)==0)
+       && ((exclude && regexec(exclude,name,0,0,0)==0)
 	   || (include && !exclude)))
       {
 	 Sub(i);
