@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1999 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1999-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FtpDirList.cc,v 1.5 1999/08/18 14:16:08 lav Exp $ */
+/* $Id: FtpDirList.cc,v 1.11.2.1 2000/09/07 15:21:35 lav Exp $ */
 
 #include <config.h>
 
@@ -26,6 +26,13 @@
 #include "FtpDirList.h"
 #include "LsCache.h"
 #include "ArgV.h"
+#include "misc.h"
+
+#include <sys/types.h>
+#include <time.h>
+#ifdef TM_IN_SYS_TIME
+# include <sys/time.h>
+#endif
 
 #define super DirList
 
@@ -42,7 +49,7 @@ int FtpDirList::Do()
 
    if(!ubuf)
    {
-      char *cache_buffer=0;
+      const char *cache_buffer=0;
       int cache_buffer_size=0;
       if(use_cache && LsCache::Find(session,pattern,FA::LONG_LIST,
 				    &cache_buffer,&cache_buffer_size))
@@ -50,7 +57,6 @@ int FtpDirList::Do()
 	 ubuf=new Buffer();
 	 ubuf->Put(cache_buffer,cache_buffer_size);
 	 ubuf->PutEOF();
-	 xfree(cache_buffer);
       }
       else
       {
@@ -67,7 +73,7 @@ int FtpDirList::Do()
    if(b==0) // eof
    {
       buf->PutEOF();
-      
+
       const char *cache_buffer;
       int cache_buffer_size;
       ubuf->GetSaved(&cache_buffer,&cache_buffer_size);
@@ -79,40 +85,111 @@ int FtpDirList::Do()
 
       return MOVED;
    }
-   
+
    int m=STALL;
 
-   while(len>0)
+   if(len>0)
    {
-      const char *cr=(const char *)memchr(b,'\r',len);
-      if(!cr)
+      const char *eol=find_char(b,len,'\n');
+      if(!eol && !ubuf->Eof() && len<0x1000)
+	 return m;
+      int skip_len=len;
+      if(eol)
       {
-	 buf->Put(b,len);
-	 ubuf->Skip(len);
-	 m=MOVED;
-	 break;
-      }
-      else
-      {
-	 if(cr-b>0)
+	 skip_len=eol+1-b;
+	 len=skip_len;
+
+	 const char *name=0;
+	 int name_len=0;
+	 long size=NO_SIZE;
+	 time_t date=NO_DATE;
+	 long date_l;
+	 bool dir=false;
+	 int perms=-1;
+	 // check for EPLF listing
+	 if(eol>b+1 && b[0]=='+')
 	 {
-	    buf->Put(b,cr-b);
-	    ubuf->Skip(cr-b);
-	    m=MOVED;
-	    len-=cr-b;
-	    b=cr;
-	 }
-	 if(len==1)
-	    break;
-	 if(b[1]!='\n')
-	 {
-	    buf->Put(b,1);
-	    m=MOVED;
-	 }
-	 ubuf->Skip(1);
-	 b++,len--;
-      }
+	    const char *scan=b+1;
+	    int scan_len=len-1;
+	    while(scan && scan_len>0)
+	    {
+	       switch(*scan)
+	       {
+		  case '\t':  // the rest is file name.
+		     name=scan+1;
+		     name_len=scan_len-1;
+		     scan=0;
+		     break;
+		  case 's':
+		     sscanf(scan+1,"%ld",&size);
+		     break;
+		  case 'm':
+		     if(1 != sscanf(scan+1,"%ld",&date_l))
+			break;
+		     date = date_l;
+		     break;
+		  case '/':
+		     dir=true;
+		     break;
+		  case 'r':
+		     dir=false;
+		     break;
+		  case 'i':
+		     break;
+		  case 'u':
+		     if(scan[1]=='p')  // permissions.
+			sscanf(scan+2,"%o",&perms);
+		     break;
+		  default:
+		     name=0;
+		     scan=0;
+		     break;
+	       }
+	       if(scan==0 || scan_len==0)
+		  break;
+	       const char *comma=find_char(scan,scan_len,',');
+	       if(comma)
+	       {
+		  scan_len-=comma+1-scan;
+		  scan=comma+1;
+	       }
+	       else
+		  break;
+	    }
+	    if(name && name_len>0)
+	    {
+	       // ok, this is EPLF. Format new string.
+	       char *line_add=string_alloca(80+name_len);
+	       if(perms==-1)
+		  perms=(dir?0755:0644);
+	       char size_str[32];
+	       if(size==NO_SIZE)
+		  strcpy(size_str,"-");
+	       else
+		  sprintf(size_str,"%ld",size);
+	       char date_str[21];
+	       if(date == NO_DATE)
+		  strcpy(date_str,"-");
+	       else {
+		  struct tm *t=localtime(&date);
+		  sprintf(date_str, "%04d-%02d-%02d %02d:%02d",
+			  t->tm_year+1900,t->tm_mon+1,t->tm_mday,
+			  t->tm_hour,t->tm_min);
+	       }
+	       sprintf(line_add, "%c%s  %10s  %16s  %.*s", 
+		       dir ? 'd':'-', format_perms(perms), size_str,
+		       date_str, name_len, name);
+	       b=line_add;
+	       len=strlen(b);
+	    }
+	 } // end if(+).
+      } // end if(eol)
+
+      buf->Put(b,len);
+      ubuf->Skip(skip_len);
+      m=MOVED;
    }
+
    if(ubuf->Error())
    {
       SetError(ubuf->ErrorText());
@@ -131,7 +208,7 @@ FtpDirList::FtpDirList(ArgV *a,FileAccess *fa)
 
 FtpDirList::~FtpDirList()
 {
-   delete ubuf;
+   Delete(ubuf);
    xfree(pattern);
 }
 
