@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FtpListInfo.cc,v 1.15 1999/08/04 23:15:17 lav Exp $ */
+/* $Id: FtpListInfo.cc,v 1.27 2000/03/18 23:46:15 lav Exp $ */
 
 #include <config.h>
 #include "FtpListInfo.h"
@@ -29,6 +29,8 @@
 #include "xstring.h"
 #include <ctype.h>
 #include "xalloca.h"
+#include "FtpSplitList.h"
+#include "misc.h"
 
 #define need_size (need&FileInfo::SIZE)
 #define need_time (need&FileInfo::DATE)
@@ -38,14 +40,13 @@ FtpListInfo::FtpListInfo(Ftp *s)
    session=s;
    get_info=0;
    get_info_cnt=0;
-   glob=0;
+   slist=0;
    state=INITIAL;
 }
 
 FtpListInfo::~FtpListInfo()
 {
-   if(glob)
-      delete glob;
+   Delete(slist);
    session->Close();
    xfree(get_info);
 }
@@ -56,7 +57,7 @@ int FtpListInfo::Do()
    FileInfo *file;
    int res;
    int m=STALL;
-   char	**glob_res;
+   char	**slist_res;
    int err;
 
    if(done)
@@ -65,63 +66,63 @@ int FtpListInfo::Do()
    switch(state)
    {
    case(INITIAL):
-      glob=new FtpGlob(session,"",session->LONG_LIST);
-      glob->UseCache(use_cache);
+      slist=new FtpSplitList(session,session->LONG_LIST);
+      slist->UseCache(use_cache);
       state=GETTING_LONG_LIST;
       m=MOVED;
 
    case(GETTING_LONG_LIST):
-      if(!glob->Done())
+      if(!slist->Done())
 	 return m;
-      if(glob->Error())
+      if(slist->Error())
       {
-	 SetError(glob->ErrorText());
-      	 delete glob;
-	 glob=0;
+	 SetError(slist->ErrorText());
+      	 Delete(slist);
+	 slist=0;
 	 return MOVED;
       }
-      glob_res=glob->GetResult();
+      slist_res=slist->GetResult();
 
       // don't consider empty list to be valid
-      if(glob_res && glob_res[0])
-	 result=ParseFtpLongList(glob_res,&err);
+      if(slist_res && slist_res[0])
+	 result=ParseFtpLongList(slist_res,&err);
       else
 	 err=1;
 
       if(!result)
 	 result=new FileSet;
 
-      delete glob;
-      glob=0;
-      glob_res=0; // note: glob_res is pointer to part of glob
+      Delete(slist);
+      slist=0;
+      slist_res=0; // note: slist_res is pointer to part of slist
 
       if(err==0)
 	 goto pre_GETTING_INFO;
 
       // there were parse errors, try another method
-      glob=new FtpGlob(session,"",session->LIST);
-      glob->UseCache(use_cache);
-      glob->NoLongList();
+      slist=new FtpSplitList(session,session->LIST);
+      slist->UseCache(use_cache);
       state=GETTING_SHORT_LIST;
       m=MOVED;
 
    case(GETTING_SHORT_LIST):
-      if(!glob->Done())
+      if(!slist->Done())
 	 return m;
-      if(glob->Error())
+      if(slist->Error())
       {
-	 SetError(glob->ErrorText());
-      	 delete glob;
-	 glob=0;
+	 SetError(slist->ErrorText());
+      	 Delete(slist);
+	 slist=0;
 	 return MOVED;
       }
-      glob_res=glob->GetResult();
-      result->Merge(glob_res);
-      delete glob;
-      glob=0;
-      glob_res=0;
+      slist_res=slist->GetResult();
+      result->Merge(slist_res);
+      Delete(slist);
+      slist=0;
+      slist_res=0;
 
    pre_GETTING_INFO:
+      result->ExcludeDots();
       if(rxc_exclude || rxc_include)
 	 result->Exclude(path,rxc_exclude,rxc_include);
 
@@ -216,6 +217,7 @@ int FtpListInfo::Do()
 const char *FtpListInfo::Status()
 {
    static char s[256];
+   const char *status;
    switch(state)
    {
    case(DONE):
@@ -223,11 +225,21 @@ const char *FtpListInfo::Status()
       return "";
    case(GETTING_LONG_LIST):
    case(GETTING_SHORT_LIST):
+      if(!slist)
+	 return "";
       sprintf(s,_("Getting directory contents (%ld)"),session->GetPos());
+      if(slist->RateValid())
+	 sprintf(s+strlen(s)," %s",slist->GetRateStr());
+      status=session->CurrentStatus();
+      if(status && status[0])
+	 sprintf(s+strlen(s)," [%s]",status);
       return s;
    case(GETTING_INFO):
       // xgettext:c-format
       sprintf(s,_("Getting files information (%d%%)"),session->InfoArrayPercentDone());
+      status=session->CurrentStatus();
+      if(status && status[0])
+	 sprintf(s+strlen(s)," [%s]",status);
       return s;
    }
    // cat't happen
@@ -235,93 +247,14 @@ const char *FtpListInfo::Status()
 }
 
 
-
-
-mode_t	 parse_perms(const char *s)
-{
-   mode_t   p=0;
-
-   if(strlen(s)!=9)
-      bad: return (mode_t)-1;
-
-   switch(s[0])
-   {
-   case('r'): p|=S_IRUSR; break;
-   case('-'): break;
-   default: goto bad;
-   }
-   switch(s[1])
-   {
-   case('w'): p|=S_IWUSR; break;
-   case('-'): break;
-   default: goto bad;
-   }
-   switch(s[2])
-   {
-   case('S'): p|=S_ISUID; break;
-   case('s'): p|=S_ISUID; // fall-through
-   case('x'): p|=S_IXUSR; break;
-   case('-'): break;
-   default: goto bad;
-   }
-   s+=3;
-   switch(s[0])
-   {
-   case('r'): p|=S_IRGRP; break;
-   case('-'): break;
-   default: goto bad;
-   }
-   switch(s[1])
-   {
-   case('w'): p|=S_IWGRP; break;
-   case('-'): break;
-   default: goto bad;
-   }
-   switch(s[2])
-   {
-   case('S'): p|=S_ISGID; break;
-   case('s'): p|=S_ISGID; // fall-through
-   case('x'): p|=S_IXGRP; break;
-   case('-'): break;
-   default: goto bad;
-   }
-   s+=3;
-   switch(s[0])
-   {
-   case('r'): p|=S_IROTH; break;
-   case('-'): break;
-   default: goto bad;
-   }
-   switch(s[1])
-   {
-   case('w'): p|=S_IWOTH; break;
-   case('-'): break;
-   default: goto bad;
-   }
-   switch(s[2])
-   {
-   case('T'): case('t'): p|=S_ISVTX; break;
-   case('l'): case('L'): p|=S_ISGID; p&=~S_IXGRP; break;
-   case('x'): p|=S_IXOTH; break;
-   case('-'): break;
-   default: goto bad;
-   }
-
-   return p;
-}
-
-int   parse_month(char *m)
-{
-   static const char *months[]={
-      "Jan","Feb","Mar","Apr","May","Jun",
-      "Jul","Aug","Sep","Oct","Nov","Dec",0
-   };
-   for(int i=0; months[i]; i++)
-      if(!strcasecmp(months[i],m))
-	 return(i%12);
-   return -1;
-}
-
+/*
+total 123
+-rwxr-xr-x   1 lav      root         4771 Sep 12  1996 install-sh
+-rw-r--r--   1 lav      root         1349 Feb  2 14:10 lftp.lsm
+drwxr-xr-x   4 lav      root         1024 Feb 22 15:32 lib
+lrwxrwxrwx   1 lav      root           33 Feb 14 17:45 ltconfig -> /usr/share/libtool/ltconfig
+NOTE: group may be missing.
+*/
 static
 FileInfo *ParseFtpLongList_UNIX(const char *line_c,int *err)
 {
@@ -456,6 +389,7 @@ FileInfo *ParseFtpLongList_UNIX(const char *line_c,int *err)
 	    fi.SetSymlink(arrow+4);
 	    break;
 	 }
+	 arrow++;
       }
    }
    fi.SetName(name);
@@ -463,6 +397,15 @@ FileInfo *ParseFtpLongList_UNIX(const char *line_c,int *err)
    return new FileInfo(fi);
 }
 
+/*
+07-13-98  09:06PM       <DIR>          aix
+07-13-98  09:06PM       <DIR>          hpux
+07-13-98  09:06PM       <DIR>          linux
+07-13-98  09:06PM       <DIR>          ncr
+07-13-98  09:06PM       <DIR>          solaris
+03-18-98  06:01AM              2109440 nlxb318e.tar
+07-02-98  11:17AM                13844 Whatsnew.txt
+*/
 static
 FileInfo *ParseFtpLongList_NT(const char *line_c,int *err)
 {
@@ -501,10 +444,10 @@ FileInfo *ParseFtpLongList_NT(const char *line_c,int *err)
 	 hour=0;
    }
    struct tm tms;
-   tms.tm_sec=0;	      /* seconds after the minute [0, 61]  */
+   tms.tm_sec=0;	   /* seconds after the minute [0, 61]  */
    tms.tm_min=minute;      /* minutes after the hour [0, 59] */
-   tms.tm_hour=hour;	      /* hour since midnight [0, 23] */
-   tms.tm_mday=day;	      /* day of the month [1, 31] */
+   tms.tm_hour=hour;	   /* hour since midnight [0, 23] */
+   tms.tm_mday=day;	   /* day of the month [1, 31] */
    tms.tm_mon=month-1;     /* months since January [0, 11] */
    tms.tm_year=year-1900;  /* years since 1900 */
    tms.tm_isdst=0;
@@ -533,10 +476,187 @@ FileInfo *ParseFtpLongList_NT(const char *line_c,int *err)
    return new FileInfo(fi);
 }
 
+/*
++i774.71425,m951188401,/,	users
++i774.49602,m917883130,r,s79126,	jgr_www2.exe
+
+starts with +
+comma separated
+first character of field is type:
+ i - ?
+ m - modification time
+ / - means directory
+ r - means plain file
+ s - size
+ up - permissions in octal
+ \t - file name follows.
+*/
+static
+FileInfo *ParseFtpLongList_EPLF(const char *line,int *err)
+{
+   int len=strlen(line);
+   const char *b=line;
+
+   if(len<2 || b[0]!='+')
+      ERR;
+
+   const char *name=0;
+   int name_len=0;
+   long size=NO_SIZE;
+   time_t date=NO_DATE;
+   long date_l;
+   bool dir=false;
+   bool type_known=false;
+   int perms=-1;
+
+   const char *scan=b+1;
+   int scan_len=len-1;
+   while(scan && scan_len>0)
+   {
+      switch(*scan)
+      {
+	 case '\t':  // the rest is file name.
+	    name=scan+1;
+	    name_len=scan_len-1;
+	    scan=0;
+	    break;
+	 case 's':
+	    sscanf(scan+1,"%ld",&size);
+	    break;
+	 case 'm':
+	    if(1 != sscanf(scan+1,"%ld",&date_l))
+	       break;
+	    date = date_l;
+	    break;
+	 case '/':
+	    dir=true;
+	    type_known=true;
+	    break;
+	 case 'r':
+	    dir=false;
+	    type_known=true;
+	    break;
+	 case 'i':
+	    break;
+	 case 'u':
+	    if(scan[1]=='p')  // permissions.
+	       sscanf(scan+2,"%o",&perms);
+	    break;
+	 default:
+	    name=0;
+	    scan=0;
+	    break;
+      }
+      if(scan==0 || scan_len==0)
+	 break;
+      const char *comma=find_char(scan,scan_len,',');
+      if(comma)
+      {
+	 scan_len-=comma+1-scan;
+	 scan=comma+1;
+      }
+      else
+	 break;
+   }
+   if(name==0 && type_known)
+      ERR;
+
+   FileInfo *fi=new FileInfo();
+   fi->SetName(name);
+   if(size!=NO_SIZE)
+      fi->SetSize(size);
+   if(date!=NO_DATE)
+      fi->SetDate(date);
+   if(type_known)
+   {
+      if(dir)
+	 fi->SetType(fi->DIRECTORY);
+      else
+	 fi->SetType(fi->NORMAL);
+   }
+   if(perms!=-1)
+      fi->SetMode(perms);
+
+   return fi;
+}
+
+/*
+                 0          DIR  06-27-96  11:57  PROTOCOL
+               169               11-29-94  09:20  SYSLEVEL.MPT
+*/
+static
+FileInfo *ParseFtpLongList_OS2(const char *line_c,int *err)
+{
+   char	 *line=alloca_strdup(line_c);
+   int 	 len=strlen(line);
+
+   if(len==0)
+      return 0;
+   char *t = FIRST_TOKEN;
+   if(t==0)
+      ERR;
+
+   FileInfo fi;
+
+   unsigned long size;
+   if(sscanf(t,"%lu",&size)!=1)
+      ERR;
+   fi.SetSize(size);
+
+   t = NEXT_TOKEN;
+   if(t==0)
+      ERR;
+   fi.SetType(fi.NORMAL);
+   if(!strcmp(t,"DIR"))
+   {
+      fi.SetType(fi.DIRECTORY);
+      t = NEXT_TOKEN;
+      if(t==0)
+	 ERR;
+   }
+   int month,day,year;
+   if(sscanf(t,"%2d-%2d-%2d",&month,&day,&year)!=3)
+      ERR;
+   if(year>=70)
+      year+=1900;
+   else
+      year+=2000;
+
+   t = NEXT_TOKEN;
+   if(t==0)
+      ERR;
+   int hour,minute;
+   if(sscanf(t,"%2d:%2d",&hour,&minute)!=3)
+      ERR;
+
+   struct tm tms;
+   tms.tm_sec=0;	   /* seconds after the minute [0, 61]  */
+   tms.tm_min=minute;      /* minutes after the hour [0, 59] */
+   tms.tm_hour=hour;	   /* hour since midnight [0, 23] */
+   tms.tm_mday=day;	   /* day of the month [1, 31] */
+   tms.tm_mon=month-1;     /* months since January [0, 11] */
+   tms.tm_year=year-1900;  /* years since 1900 */
+   tms.tm_isdst=0;
+   fi.SetDateUnprec(mktime(&tms));
+
+   t=strtok(NULL,"");
+   if(t==0)
+      ERR;
+   while(*t==' ')
+      t++;
+   if(*t==0)
+      ERR;
+   fi.SetName(t);
+
+   return new FileInfo(fi);
+}
+
 typedef FileInfo *(*ListParser)(const char *line,int *err);
 static ListParser list_parsers[]={
    ParseFtpLongList_UNIX,
    ParseFtpLongList_NT,
+   ParseFtpLongList_EPLF,
+   ParseFtpLongList_OS2,
    0
 };
 
@@ -562,7 +682,10 @@ FileSet *FtpListInfo::ParseFtpLongList(const char * const *lines_c,int *err_ret)
       {
 	 FileInfo *fi=(*parser)(line,&err);
 	 if(fi)
-	    result->Add(fi);
+	 {
+	    if(!strchr(fi->name,'/'))
+	       result->Add(fi);
+	 }
 	 if(err>=best_err)
 	    break;
       }
