@@ -18,13 +18,12 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FtpListInfo.h,v 1.4 1999/04/22 13:41:33 lav Exp $ */
+/* $Id: FtpListInfo.h,v 1.5 1999/11/02 16:39:04 lav Exp $ */
 
 #ifndef FTPLISTINFO_H
 #define FTPLISTINFO_H
 
 #include "ftpclass.h"
-#include "FtpGlob.h"
 
 class FtpListInfo : public ListInfo
 {
@@ -42,7 +41,7 @@ class FtpListInfo : public ListInfo
    Ftp::fileinfo *get_info;
    int get_info_cnt;
 
-   FtpGlob *glob;
+   class FtpSplitList *slist;
 
 public:
    static FileSet *ParseFtpLongList(const char * const *lines,int *err);
