@@ -18,39 +18,36 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: GetJob.h,v 1.2 1998/09/18 19:12:32 lav Exp $ */
+/* $Id: GetJob.h,v 1.4 1999/12/26 20:49:24 lav Exp $ */
 
 #ifndef GETJOB_H
 #define GETJOB_H
 
-#include "FileXfer.h"
+#include "CopyJob.h"
 
-class GetJob : public FileXfer
+class GetJob : public CopyJobEnv
 {
 protected:
    void	 NextFile();
 
-   time_t set_file_time;
-   time_t file_time;
-   bool delete_files:1;
-   bool	deleting:1;
-   bool made_backup:1;
+   bool delete_files;
+   char *backup_file;
+   FDStream *local;
+   bool reverse;
 
    void RemoveBackupFile();
 
+   FileCopyPeer *NoProtoSrc(const char *src);
+   FileCopyPeer *NoProtoDst(const char *dst);
+
 public:
+   GetJob(FileAccess *s,ArgV *a,bool c=false);
+   ~GetJob();
+
    int	 Do();
 
-   void SetTime(time_t t) { set_file_time=t; }
    void DeleteFiles() { delete_files=true; }
-
-   GetJob(FileAccess *s,ArgV *args,bool cont=false)
-      : FileXfer(s,args,cont)
-   {
-      set_file_time=(time_t)-1;
-      file_time=(time_t)-2;
-      delete_files=deleting=made_backup=false;
-   }
+   void Reverse() { reverse=true; } // put
 };
 
 #endif /* GETJOB_H */
