@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1999 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1999-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,13 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Http.cc,v 1.37.2.3 2000/01/09 00:13:10 lav Exp $ */
-
-#ifdef __linux__
-/* to get prototype for strptime, we need this */
-# define _XOPEN_SOURCE 500
-# define _XOPEN_SOURCE_EXTENDED 1
-#endif
+/* $Id: Http.cc,v 1.65.2.2 2000/06/20 06:38:49 lav Exp $ */
 
 #include <config.h>
 #include <unistd.h>
@@ -43,6 +37,10 @@
 
 #include "ascii_ctype.h"
 
+#ifndef HAVE_STRPTIME_DECL
+CDECL char *strptime(const char *buf, const char *format, struct tm *tm);
+#endif
+
 #define super NetAccess
 
 #define max_buf 0x10000
@@ -67,6 +65,7 @@ void Http::Init()
    send_buf=0;
    recv_buf=0;
    body_size=-1;
+   bytes_received=0;
    line=0;
    status=0;
    status_code=0;
@@ -86,11 +85,16 @@ void Http::Init()
    chunk_size=-1;
    chunk_pos=0;
 
+   no_ranges=false;
+   seen_ranges_bytes=false;
+
    no_cache_this=false;
    no_cache=false;
 
    hftp=false;
    use_head=true;
+
+   user_agent=0;
 }
 
 Http::Http() : super()
@@ -114,10 +118,8 @@ Http::Http(const Http *f) : super(f)
 
 Http::~Http()
 {
-   if(send_buf)
-      delete send_buf;
-   if(recv_buf)
-      delete recv_buf;
+   Delete(send_buf);
+   Delete(recv_buf);
    if(sock!=-1)
       close(sock);
    xfree(line);
@@ -125,31 +127,12 @@ Http::~Http()
    xfree(location);
 }
 
-bool Http::CheckTimeout()
-{
-   if(now-event_time>=timeout)
-   {
-      DebugPrint("**** ",_("Timeout - reconnecting"));
-      Disconnect();
-      event_time=now;
-      return(true);
-   }
-   block+=TimeOut((timeout-(now-event_time))*1000);
-   return(false);
-}
-
 void Http::Disconnect()
 {
-   if(send_buf)
-   {
-      delete send_buf;
-      send_buf=0;
-   }
-   if(recv_buf)
-   {
-      delete recv_buf;
-      recv_buf=0;
-   }
+   Delete(send_buf);
+   send_buf=0;
+   Delete(recv_buf);
+   recv_buf=0;
    if(rate_limit)
    {
       delete rate_limit;
@@ -175,6 +158,7 @@ void Http::Disconnect()
    chunked=false;
    chunk_size=-1;
    chunk_pos=0;
+   seen_ranges_bytes=false;
    if(mode==STORE && state!=DONE && !Error())
       SetError(STORE_FAILED,0);
    else
@@ -183,25 +167,47 @@ void Http::Disconnect()
 
 void Http::Close()
 {
-   retries=0;
    Disconnect();
    array_send=0;
    no_cache_this=false;
+   no_ranges=false;
+   bytes_received=0;
    super::Close();
 }
 
+#if defined(HAVE_VSNPRINTF) && !defined(HAVE_VSNPRINTF_DECL)
+CDECL int vsnprintf(char *,size_t,const char *,va_list);
+#endif
+
 void Http::Send(const char *format,...)
 {
-   char *str=(char*)alloca(max_send);
    va_list va;
    va_start(va,format);
 
+   char *str;
+#ifdef HAVE_VSNPRINTF
+   static int max_send=256;
+   for(;;)
+   {
+      str=string_alloca(max_send);
+      int res=vsnprintf(str,max_send,format,va);
+      if(res>=0 && res<max_send)
+      {
+	 if(res<max_send/16)
+	    max_send/=2;
+	 break;
+      }
+      max_send*=2;
+   }
+#else // !HAVE_VSNPRINTF
+   str=string_alloca(2048);
    vsprintf(str,format,va);
+#endif
+
+   DebugPrint("---> ",str,5);
+   send_buf->Put(str);
 
    va_end(va);
-
-   DebugPrint("---> ",str,3);
-   send_buf->Put(str);
 }
 
 void Http::SendMethod(const char *method,const char *efile)
@@ -212,7 +218,8 @@ void Http::SendMethod(const char *method,const char *efile)
       method="GET";
    Send("%s %s HTTP/1.1\r\n",method,efile);
    Send("Host: %s\r\n",ehost);
-   Send("User-Agent: %s/%s\r\n","lftp",VERSION);
+   if(user_agent && user_agent[0])
+      Send("User-Agent: %s\r\n",user_agent);
    Send("Accept: */*\r\n");
 }
 
@@ -240,12 +247,12 @@ bool Http::ModeSupported()
    switch((open_mode)mode)
    {
    case CLOSED:
-   case CONNECT_VERIFY:
    case QUOTE_CMD:
    case RENAME:
    case LIST:
    case CHANGE_MODE:
       return false;
+   case CONNECT_VERIFY:
    case RETRIEVE:
    case STORE:
    case MAKE_DIR:
@@ -262,14 +269,14 @@ bool Http::ModeSupported()
 void Http::SendRequest(const char *connection,const char *f)
 {
    char *efile=string_alloca(strlen(f)*3+1);
-   url::encode_string(f,efile);
+   url::encode_string(f,efile,URL_PATH_UNSAFE);
    char *ecwd=string_alloca(strlen(cwd)*3+1);
-   url::encode_string(cwd,ecwd);
+   url::encode_string(cwd,ecwd,URL_PATH_UNSAFE);
    int efile_len;
 
    char *pfile=(char*)alloca(4+3+xstrlen(user)*6+3+xstrlen(pass)*3+1+
 			      strlen(hostname)*3+1+strlen(cwd)*3+1+
-			      strlen(efile)+1+1);
+			      strlen(efile)+1+6+1);
 
    if(proxy)
    {
@@ -281,7 +288,7 @@ void Http::SendRequest(const char *connection,const char *f)
 	    strcpy(pfile,"ftp://");
 	    url::encode_string(user,pfile+strlen(pfile),URL_USER_UNSAFE);
 	    strcat(pfile,"@");
-	    url::encode_string(hostname,pfile+strlen(pfile),URL_PATH_UNSAFE);
+	    url::encode_string(hostname,pfile+strlen(pfile),URL_HOST_UNSAFE);
 	    goto add_path;
 	 }
 	 proto="ftp";
@@ -300,25 +307,40 @@ void Http::SendRequest(const char *connection,const char *f)
    }
 
 add_path:
-   if(ecwd[0]=='~' && ecwd[1]=='/')
-      ecwd+=1;
+
+   char *path_base=pfile+strlen(pfile);
 
    if(efile[0]=='/')
-      strcat(pfile,efile);
+      strcpy(path_base,efile);
    else if(efile[0]=='~')
-      sprintf(pfile+strlen(pfile),"/%s",efile);
-   else if(cwd[0]==0 || ((cwd[0]=='/' || cwd[0]=='~') && cwd[1]==0))
-      sprintf(pfile+strlen(pfile),"/%s",efile);
+      sprintf(path_base,"/%s",efile);
+   else if(cwd[0]==0 || ((cwd[0]=='/' || (!hftp && cwd[0]=='~')) && cwd[1]==0))
+      sprintf(path_base,"/%s",efile);
    else if(cwd[0]=='~')
-      sprintf(pfile+strlen(pfile),"/%s/%s",ecwd,efile);
+      sprintf(path_base,"/%s/%s",ecwd,efile);
    else
-      sprintf(pfile+strlen(pfile),"%s/%s",ecwd,efile);
+      sprintf(path_base,"%s/%s",ecwd,efile);
+
+   if(path_base[1]=='~' && path_base[2]=='/')
+      memmove(path_base,path_base+2,strlen(path_base+2)+1);
+   else if(hftp && path_base[1]!='~')
+   {
+      // root directory in ftp urls needs special encoding. (/%2Fpath)
+      memmove(path_base+4,path_base+1,strlen(path_base+1)+1);
+      memcpy(path_base+1,"%2F",3);
+   }
+
+   if(hftp && mode!=LONG_LIST && mode!=CHANGE_DIR && mode!=MAKE_DIR)
+   {
+      if(ascii)
+	 strcat(pfile,";type=a");
+      else
+	 strcat(pfile,";type=i");
+   }
 
    efile=pfile;
    efile_len=strlen(efile);
 
-   max_send=efile_len+40;
-
    if(pos==0)
       real_pos=0;
    if(mode==STORE)    // can't seek before writing
@@ -337,7 +359,7 @@ add_path:
    case RETRIEVE:
    retrieve:
       SendMethod("GET",efile);
-      if(pos>0)
+      if(pos>0 && !no_ranges)
 	 Send("Range: bytes=%ld-\r\n",pos);
       break;
 
@@ -349,12 +371,28 @@ add_path:
 	 Send("Range: bytes=%ld-\r\n",pos);
       else if(pos>0)
 	 Send("Range: bytes=%ld-%ld/%ld\r\n",pos,entity_size-1,entity_size);
+      if(entity_date!=(time_t)-1)
+      {
+	 char d[256];
+	 static const char weekday_names[][4]={
+	    "Sun","Mon","Tue","Wed","Thu","Fri","Sat"
+	 };
+	 static const char month_names[][4]={
+	    "Jan","Feb","Mar","Apr","May","Jun",
+	    "Jul","Aug","Sep","Oct","Nov","Dec",
+	 };
+	 struct tm *t=gmtime(&entity_date);
+	 sprintf(d,"%s, %2d %s %04d %02d:%02d:%02d GMT",
+	    weekday_names[t->tm_wday],t->tm_mday,month_names[t->tm_mon],
+	    t->tm_year+1900,t->tm_hour,t->tm_min,t->tm_sec);
+	 Send("Last-Modified: %s\r\n",d);
+      }
       break;
 
    case CHANGE_DIR:
    case LONG_LIST:
    case MAKE_DIR:
-      if(efile[0]==0 || efile[efile_len-1]!='/')
+      if(efile_len<1 || efile[efile_len-1]!='/')
 	 strcat(efile,"/");
       if(mode==CHANGE_DIR)
 	 SendMethod("HEAD",efile);
@@ -379,6 +417,12 @@ add_path:
       Send("Pragma: no-cache\r\n"); // for HTTP/1.0 compatibility
       Send("Cache-Control: no-cache\r\n");
    }
+   if(!hftp)
+   {
+      const char *cookie=Query("cookie",hostname);
+      if(cookie && cookie[0])
+	 Send("Cookie: %s\r\n",cookie);
+   }
    if(mode==ARRAY_INFO && !use_head)
       connection="close";
    if(mode!=ARRAY_INFO || connection)
@@ -389,6 +433,7 @@ add_path:
    chunked=false;
    chunk_size=-1;
    chunk_pos=0;
+   no_ranges=false;
 }
 
 void Http::SendArrayInfoRequest()
@@ -408,6 +453,8 @@ void Http::HandleHeaderLine(const char *name,const char *value)
    if(!strcasecmp(name,"Content-length"))
    {
       sscanf(value,"%ld",&body_size);
+      if(pos==0 && mode!=STORE)
+	 entity_size=body_size;
       if(pos==0 && opt_size)
 	 *opt_size=body_size;
 
@@ -426,6 +473,8 @@ void Http::HandleHeaderLine(const char *name,const char *value)
 	 return;
       real_pos=first;
       body_size=last-first+1;
+      if(mode!=STORE)
+	 entity_size=fsize;
       if(opt_size)
 	 *opt_size=fsize;
       return;
@@ -465,6 +514,7 @@ void Http::HandleHeaderLine(const char *name,const char *value)
 	 keep_alive=true;
       else if(!strcasecmp(value,"close"))
 	 keep_alive=false;
+      return;
    }
    if(!strcasecmp(name,"Transfer-Encoding"))
    {
@@ -474,6 +524,15 @@ void Http::HandleHeaderLine(const char *name,const char *value)
 	 chunk_size=-1;	// to indicate "before first chunk"
 	 chunk_pos=0;
       }
+      return;
+   }
+   if(!strcasecmp(name,"Accept-Ranges"))
+   {
+      if(!strcasecmp(value,"none"))
+	 no_ranges=true;
+      if(strstr(value,"bytes"))
+	 seen_ranges_bytes=true;
+      return;
    }
 }
 
@@ -499,19 +558,20 @@ int Http::Do()
    Buffer *data_buf;
 
    // check if idle time exceeded
+   // NOT USED YET.
    if(mode==CLOSED && sock!=-1 && idle>0)
    {
       if(now-idle_start>=idle)
       {
-	 DebugPrint("---- ",_("Closing idle connection"),2);
+	 DebugPrint("---- ",_("Closing idle connection"),1);
 	 Disconnect();
 	 return m;
       }
-      Timeout((idle_start+idle-now)*1000);
+      TimeoutS(idle_start+idle-now);
    }
 
    if(home==0)
-      home=xstrdup("/");
+      home=xstrdup(default_cwd);
    ExpandTildeInCWD();
 
    if(Error())
@@ -536,34 +596,20 @@ int Http::Do()
 	    return MOVED;
 	 }
       }
-      if(try_time!=0 && now-try_time<reconnect_interval)
-      {
-	 block+=TimeOut(1000*(reconnect_interval-(now-try_time)));
-	 return m;
-      }
 
-      if(peer==0 || relookup_always)
-      {
-	 if(Resolve(HTTP_DEFAULT_PORT,"http","tcp")==MOVED)
-	    m=MOVED;
-	 if(!peer)
-	    return m;
-      }
+      if(!ReconnectAllowed())
+	 return m;
+
+      if(Resolve(HTTP_DEFAULT_PORT,"http","tcp")==MOVED)
+	 m=MOVED;
+      if(!peer)
+	 return m;
 
       if(mode==CONNECT_VERIFY)
 	 return m;
 
-      try_time=now;
-
-      if(max_retries>0 && retries>=max_retries)
-      {
-	 Fatal(_("max-retries exceeded"));
+      if(!NextTry())
       	 return MOVED;
-      }
-      retries++;
-
-      assert(peer!=0);
-      assert(peer_curr<peer_num);
 
       sock=socket(peer[peer_curr].sa.sa_family,SOCK_STREAM,IPPROTO_TCP);
       if(sock==-1)
@@ -577,7 +623,7 @@ int Http::Do()
 	 if(errno==ENFILE || errno==EMFILE)
 	 {
 	    // file table overflow - it could free sometime
-	    block+=TimeOut(1000);
+	    TimeoutS(1);
 	    return m;
 	 }
 	 char str[256];
@@ -634,7 +680,7 @@ int Http::Do()
       send_buf=new FileOutputBuffer(new FDStream(sock,"<output-socket>"));
       recv_buf=new FileInputBuffer(new FDStream(sock,"<input-socket>"));
 
-      DebugPrint("---- ","Sending request...",9);
+      DebugPrint("---- ","Sending request...");
       if(mode==ARRAY_INFO)
       {
 	 SendArrayInfoRequest();
@@ -655,10 +701,12 @@ int Http::Do()
    case RECEIVING_HEADER:
       if(send_buf->Error() || recv_buf->Error())
       {
+	 if(mode==STORE && status_code && !H_20X(status_code))
+	    goto pre_RECEIVING_BODY;   // assume error.
 	 if(send_buf->Error())
-	    DebugPrint("**** ",send_buf->ErrorText());
+	    DebugPrint("**** ",send_buf->ErrorText(),0);
 	 if(recv_buf->Error())
-	    DebugPrint("**** ",recv_buf->ErrorText());
+	    DebugPrint("**** ",recv_buf->ErrorText(),0);
 	 Disconnect();
 	 return MOVED;
       }
@@ -670,7 +718,7 @@ int Http::Do()
       if(!buf)
       {
 	 // eof
-	 DebugPrint("**** ","Hit EOF while fetching headers");
+	 DebugPrint("**** ","Hit EOF while fetching headers",0);
 	 Disconnect();
 	 return MOVED;
       }
@@ -681,7 +729,7 @@ int Http::Do()
 	 {
 	    if(eol==buf)
 	    {
-	       DebugPrint("<--- ","",2);
+	       DebugPrint("<--- ","",4);
 	       recv_buf->Skip(2);
 	       if(mode==ARRAY_INFO)
 	       {
@@ -723,7 +771,7 @@ int Http::Do()
 		  if(!sent_eot && H_20X(status_code))
 		  {
 		     // should never happen
-		     DebugPrint("**** ","Success, but did nothing??");
+		     DebugPrint("**** ","Success, but did nothing??",0);
 		     Disconnect();
 		     return MOVED;
 		  }
@@ -739,7 +787,7 @@ int Http::Do()
 
 	    recv_buf->Skip(len+2);
 
-	    DebugPrint("<--- ",line,2);
+	    DebugPrint("<--- ",line,4);
 	    m=MOVED;
 
 	    if(status==0)
@@ -751,11 +799,16 @@ int Http::Do()
 	       if(3!=sscanf(status,"HTTP/%d.%d %n%d",&ver_major,&ver_minor,
 		     &status_consumed,&status_code))
 	       {
-		  DebugPrint("**** ","Could not parse HTTP status line",1);
 		  // simple 0.9 ?
-		  proto_version=0x09;
-		  //FIXME: STORE
-		  goto pre_RECEIVING_BODY;
+		  ver_major=0;
+		  ver_minor=9;
+		  status_code=200;
+		  if(1!=sscanf(status,"HTTP %n%d",&status_consumed,&status_code))
+		  {
+		     DebugPrint("**** ","Could not parse HTTP status line",0);
+		     //FIXME: STORE
+		     goto pre_RECEIVING_BODY;
+		  }
 	       }
 	       proto_version=(ver_major<<4)+ver_minor;
 	       if(!H_20X(status_code))
@@ -823,7 +876,8 @@ int Http::Do()
 	    if(file && file[0])
 	       sprintf(err,"%s (%s)",status+status_consumed,file);
 	    else
-	       sprintf(err,"%s (%s/)",status+status_consumed,cwd);
+	       sprintf(err,"%s (%s%s)",status+status_consumed,cwd,
+		  (cwd[0] && cwd[strlen(cwd)-1]=='/')?"":"/");
 	 }
 	 Disconnect();
 	 SetError(NO_FILE,err);
@@ -838,11 +892,19 @@ int Http::Do()
 	 return MOVED;
       }
 
-      DebugPrint("---- ","Receiving body...",9);
+      DebugPrint("---- ","Receiving body...");
       assert(rate_limit==0);
       rate_limit=new RateLimit();
       if(real_pos<0) // assume Range: did not work
+      {
+	 if(mode!=STORE && body_size>=0)
+	 {
+	    entity_size=body_size;
+	    if(opt_size)
+	       *opt_size=entity_size;
+	 }
 	 real_pos=0;
+      }
       state=RECEIVING_BODY;
       m=MOVED;
    case RECEIVING_BODY:
@@ -850,9 +912,9 @@ int Http::Do()
       if(recv_buf->Error() || send_buf->Error())
       {
 	 if(send_buf->Error())
-	    DebugPrint("**** ",send_buf->ErrorText());
+	    DebugPrint("**** ",send_buf->ErrorText(),0);
 	 if(recv_buf->Error())
-	    DebugPrint("**** ",recv_buf->ErrorText());
+	    DebugPrint("**** ",recv_buf->ErrorText(),0);
 	 Disconnect();
 	 return MOVED;
       }
@@ -875,8 +937,13 @@ int Http::Do()
 	    m=MOVED;
 	 else
 	 {
+	    // check if ranges were emulated by squid
+	    bool no_ranges_if_timeout=(bytes_received==0 && !seen_ranges_bytes);
 	    if(CheckTimeout())
+	    {
+	       no_ranges|=no_ranges_if_timeout;
 	       return MOVED;
+	    }
 	 }
       }
       return m;
@@ -893,12 +960,14 @@ system_error:
       Timeout(1000);
       return m;
    }
-   saved_errno=errno;
+   SetError(SEE_ERRNO,0);
    Disconnect();
-   SetError(SEE_ERRNO,strerror(saved_errno));
    return MOVED;
 }
 
+FileAccess *Http::New() { return new Http(); }
+FileAccess *HFtp::New() { return new HFtp(); }
+
 void  Http::ClassInit()
 {
    // register the class
@@ -927,10 +996,10 @@ int Http::Read(void *buf,int size)
       recv_buf->Get(&buf1,&size1);
       if(buf1==0) // eof
       {
-	 DebugPrint("---- ","Hit EOF",9);
+	 DebugPrint("---- ","Hit EOF");
 	 if(bytes_received<body_size || chunked)
 	 {
-	    DebugPrint("**** ","Received not enough data, retrying",1);
+	    DebugPrint("**** ","Received not enough data, retrying",0);
 	    Disconnect();
 	    return DO_AGAIN;
 	 }
@@ -939,7 +1008,15 @@ int Http::Read(void *buf,int size)
       if(size1==0)
 	 return DO_AGAIN;
       if(body_size>=0 && bytes_received>=body_size)
+      {
+	 DebugPrint("---- ","Received all");
 	 return 0; // all received
+      }
+      if(entity_size>=0 && pos>=entity_size)
+      {
+	 DebugPrint("---- ","Received all (total)");
+	 return 0;
+      }
       if(chunked)
       {
 	 if(chunk_size==-1) // expecting first/next chunk
@@ -992,6 +1069,8 @@ int Http::Read(void *buf,int size)
 	 size1=bytes_allowed;
       if(size1==0)
 	 return DO_AGAIN;
+      if(norest_manual && real_pos==0 && pos>0)
+	 return DO_AGAIN;
       if(real_pos<pos)
       {
 	 long to_skip=pos-real_pos;
@@ -1028,6 +1107,8 @@ int Http::Done()
       return error_code;
    if(state==DONE)
       return OK;
+   if(mode==CONNECT_VERIFY && peer)
+      return OK;
    return IN_PROGRESS;
 }
 
@@ -1060,13 +1141,12 @@ int Http::Write(const void *buf,int size)
 	 return DO_AGAIN;
       if(NotSerious(errno) || errno==EPIPE)
       {
-	 DebugPrint("**** ",strerror(errno));
+	 DebugPrint("**** ",strerror(errno),0);
 	 Disconnect();
 	 return STORE_FAILED;
       }
-      saved_errno=errno;
+      SetError(SEE_ERRNO,0);
       Disconnect();
-      SetError(SEE_ERRNO,strerror(saved_errno));
       return error_code;
    }
    retries=0;
@@ -1111,8 +1191,8 @@ const char *Http::CurrentStatus()
       {
 	 if(resolver)
 	    return(_("Resolving host address..."));
-	 if(now-try_time<reconnect_interval)
-	    return(_("Delaying before reconnect"));
+	 if(!ReconnectAllowed())
+	    return DelayingMessage();
       }
       return "";
    case CONNECTING:
@@ -1138,12 +1218,17 @@ void Http::Reconfig(const char *name)
    super::Reconfig(name);
 
    no_cache = !(bool)Query("cache",c);
-   SetProxy(Query("proxy",c));
+   if(!hftp && NoProxy())
+      SetProxy(0);
+   else
+      SetProxy(Query("proxy",c));
 
    if(sock!=-1)
       SetSocketBuffer(sock,socket_buffer);
-   if(proxy_port==0)
+   if(proxy && proxy_port==0)
       proxy_port=xstrdup(HTTP_DEFAULT_PROXY_PORT);
+
+   user_agent=ResMgr::Query("http:user-agent",c);
 }
 
 bool Http::SameSiteAs(FileAccess *fa)
@@ -1169,10 +1254,7 @@ void Http::Connect(const char *new_host,const char *new_port)
 {
    super::Connect(new_host,new_port);
    Reconfig();
-   DontSleep();
    state=DISCONNECTED;
-   ClearPeer();
-   try_time=0;
 }
 
 DirList *Http::MakeDirList(ArgV *args)
@@ -1194,6 +1276,7 @@ ListInfo *Http::MakeListInfo()
 HFtp::HFtp()
 {
    hftp=true;
+   default_cwd="~";
    Reconfig(0);
 }
 HFtp::~HFtp()
@@ -1209,8 +1292,8 @@ void HFtp::Login(const char *u,const char *p)
    super::Login(u,p);
    if(u)
    {
-      home=(char*)xmalloc(strlen(u)+2);
-      sprintf(home,"~%s",u);
+      xfree(home);
+      home=xstrdup("~");
       xfree(cwd);
       cwd=xstrdup(home);
    }
