@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Http.h,v 1.14 1999/10/22 15:26:18 lav Exp $ */
+/* $Id: Http.h,v 1.21.2.1 2000/06/08 10:13:31 lav Exp $ */
 
 #ifndef HTTP_H
 #define HTTP_H
@@ -42,7 +42,6 @@ class Http : public NetAccess
 
    void Init();
 
-   int 	 max_send;
    void	 Send(const char *format,...) PRINTF_LIKE(2,3);
 
    FileOutputBuffer *send_buf;
@@ -61,6 +60,14 @@ class Http : public NetAccess
 
    int sock;
    void Disconnect();
+   int IsConnected()
+      {
+	 if(sock==-1)
+	    return 0;
+	 if(state==CONNECTING)
+	    return 1;
+	 return 2;
+      }
 
    char *status;
    int   status_consumed;
@@ -71,8 +78,6 @@ class Http : public NetAccess
    char *location;
    bool sent_eot;
 
-   bool CheckTimeout();
-
    bool ModeSupported();
 
    int  keep_alive_max;
@@ -84,9 +89,14 @@ class Http : public NetAccess
    long chunk_size;
    long chunk_pos;
 
+   bool no_ranges;
+   bool seen_ranges_bytes;
+
    bool no_cache;
    bool no_cache_this;
 
+   const char *user_agent;
+
 protected:
    bool hftp;  // ftp over http proxy.
    bool use_head;
@@ -101,7 +111,7 @@ public:
    const char *GetProto() { return "http"; }
 
    FileAccess *Clone() { return new Http(this); }
-   static FileAccess *New() { return new Http(); }
+   static FileAccess *New();
 
    int Do();
    int Done();
@@ -125,6 +135,8 @@ public:
    ListInfo *MakeListInfo();
 
    void UseCache(bool use) { no_cache_this=!use; }
+
+   bool NeedSizeDateBeforehand() { return true; }
 };
 
 class HFtp : public Http
@@ -137,7 +149,7 @@ public:
    const char *GetProto() { return "hftp"; }
 
    FileAccess *Clone() { return new HFtp(this); }
-   static FileAccess *New() { return new HFtp(); }
+   static FileAccess *New();
 
    virtual void Login(const char *,const char *);
    virtual void Reconfig(const char *);
