@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: HttpDir.cc,v 1.31.2.2 2000/01/13 14:08:11 lav Exp $ */
+/* $Id: HttpDir.cc,v 1.44.2.1 2000/06/08 23:08:37 lav Exp $ */
 
 #include <config.h>
 #include "ascii_ctype.h"
@@ -28,13 +28,9 @@
 #include "ArgV.h"
 #include "LsCache.h"
 #include "misc.h"
+#include "log.h"
 #include <ctype.h>
 
-static const char *find_char(const char *buf,int len,char ch)
-{
-   return (const char *)memchr(buf,ch,len);
-}
-
 static bool token_eq(const char *buf,int len,const char *token)
 {
    int token_len=strlen(token);
@@ -44,18 +40,6 @@ static bool token_eq(const char *buf,int len,const char *token)
 	    && (token_len==len || !is_ascii_alnum(buf[token_len]));
 }
 
-static int parse_month(const char *m)
-{
-   static const char *months[]={
-      "Jan","Feb","Mar","Apr","May","Jun",
-      "Jul","Aug","Sep","Oct","Nov","Dec",0
-   };
-   for(int i=0; months[i]; i++)
-      if(!strcasecmp(months[i],m))
-	 return(i%12);
-   return -1;
-}
-
 static bool find_value(const char *scan,const char *more,const char *name,char *store)
 {
    for(;;)
@@ -105,18 +89,75 @@ static bool find_value(const char *scan,const char *more,const char *name,char *
    return false;
 }
 
+static
+void remove_tags(char *buf)
+{
+   for(;;)
+   {
+      char *less=strchr(buf,'<');
+      if(!less)
+	 break;
+      if(token_eq(less+1,strlen(less+1),"a"))
+      {
+	 // don't allow anchors to be skipped.
+      	 *less=0;
+	 break;
+      }
+      char *more=strchr(less+1,'>');
+      if(!more)
+	 break;
+      memmove(less,more+1,strlen(more+1)+1);
+      buf=less;
+   }
+}
+
+#if 0 // unused
+static
+const char *strncasestr(const char *buf,int len,const char *str)
+{
+   int str_len=strlen(str);
+   while(len>=str_len)
+   {
+      if(!strncasecmp(buf,str,str_len))
+	 return buf;
+      buf++;
+      len--;
+   }
+   return 0;
+}
+#endif
+
 static
 const char *find_eol(const char *buf,int len,bool eof,int *eol_size)
 {
    const char *real_eol=find_char(buf,len,'\n');
-   const char *less=find_char(buf,len,'<');
+   // check if the tag after eol is <TD> or </TD>
+   while(real_eol)
+   {
+      const char *scan=real_eol+1;
+      while(scan<buf+len && is_ascii_space(*scan))
+	 scan++;  // skip space
+      if(scan<buf+len && *scan!='<')
+	 break;
+      if(scan+5>buf+len)
+      {
+	 if(!eof)
+	    real_eol=0;
+	 break;
+      }
+      if(strncasecmp(scan,"<td>",4) && strncasecmp(scan,"</td>",5))
+	 break;
+      real_eol=find_char(scan,len-(scan-buf),'\n');
+   }
+   const char *less=find_char(buf,len,'<');;
    const char *more=0;
    if(less)
    {
       more=find_char(less+1,len-(less+1-buf),'>');
-      if(more && !token_eq(less+1,len-(less+1-buf),"br"))
+      if(more && !token_eq(less+1,len-(less+1-buf),"br")
+      && !token_eq(less+1,len-(less+1-buf),"/tr"))
       {
-	 // if the tag is finished and not BR, ignore it.
+	 // if the tag is finished and not BR nor /TR, ignore it.
 	 less=0;
 	 more=0;
       }
@@ -127,7 +168,7 @@ const char *find_eol(const char *buf,int len,bool eof,int *eol_size)
    // real_eol not found?
    if(!real_eol)
    {
-      // BR found?
+      // BR or /TR found?
       if(less && more)
       {
 	 *eol_size=more-less+1;
@@ -143,7 +184,7 @@ const char *find_eol(const char *buf,int len,bool eof,int *eol_size)
 }
 
 /* This function replaces &amp; &lt; &gt; &quot; to appropriate characters */
-void decode_amps(char *s)
+static void decode_amps(char *s)
 {
    static const struct pair
       { char str[7]; char ch; }
@@ -172,6 +213,8 @@ void decode_amps(char *s)
    }
 }
 
+#define debug(str) Log::global->Format(10,"* %s\n",str)
+
 static int parse_html(const char *buf,int len,bool eof,Buffer *list,
       FileSet *set,FileSet *all_links,ParsedURL *prefix,char **base_href,
       LsOptions *lsopt=0)
@@ -253,6 +296,8 @@ static int parse_html(const char *buf,int len,bool eof,Buffer *list,
    if(tag_scan->tag==0)
       return tag_len;	// not interesting
 
+   bool hftp=(prefix && !xstrcmp(prefix->proto,"hftp"));
+
    // ok, found the target.
 
    decode_amps(link_target);  // decode all &amp; and similar
@@ -262,7 +307,20 @@ static int parse_html(const char *buf,int len,bool eof,Buffer *list,
       if(base_href)
       {
 	 xfree(*base_href);
-	 *base_href=xstrdup(link_target);
+	 *base_href=xstrdup(link_target,+2);
+	 if(hftp)
+	 {
+	    // workaround apache proxy bugs.
+	    char *t=strstr(*base_href,";type=");
+	    if(t && t[6] && t[7]=='/' && t[8]==0)
+	       *t=0;
+	    char *p=*base_href+url::path_index(*base_href);
+	    if(p[0]=='/' && p[1]=='/')
+      	    {
+	       memmove(p+4,p+2,strlen(p+2)+1);
+	       memcpy(p+1,"%2F",3);
+	    }
+	 }
       }
       return tag_len;
    }
@@ -279,6 +337,14 @@ static int parse_html(const char *buf,int len,bool eof,Buffer *list,
       if(strncasecmp(link_target,"URL=",4))
 	 return tag_len;
       link_target+=4;
+      if(link_target[0]=='\'')
+      {
+	 // FIXME: maybe a more complex value parser is required.
+	 link_target++;
+	 int len=strlen(link_target);
+	 if(len>0 && link_target[len-1]=='\'')
+	    link_target[len-1]=0;
+      }
    }
 
    bool icon=false;
@@ -316,11 +382,7 @@ parse_url_again:
       if(!prefix)
 	 return tag_len;	// no way
 
-      const char *pproto=prefix->proto;
-      if(!xstrcmp(pproto,"hftp"))
-	 pproto++;
-
-      if(xstrcmp(link_url.proto,pproto)
+      if(xstrcmp(link_url.proto,prefix->proto+hftp)
       || xstrcmp(link_url.host,prefix->host)
       || xstrcmp(link_url.user,prefix->user)
       || xstrcmp(link_url.port,prefix->port))
@@ -357,16 +419,23 @@ parse_url_again:
    else
       strcpy(link_target,link_url.path);
 
+   if(link_target[0]=='/' && link_target[1]=='/' && hftp)
+   {
+      // workaround for apache proxy.
+      link_target++;
+   }
+
    int link_len=strlen(link_target);
    bool is_directory=(link_len>0 && link_target[link_len-1]=='/');
    if(is_directory && link_len>1)
       link_target[--link_len]=0;
 
-   if(prefix && prefix->path)
+   if(prefix)
    {
       const char *p_path=prefix->path;
+      if(p_path==0)
+	 p_path="~";
       int p_len=strlen(p_path);
-
       if(p_len==1 && p_path[0]=='/' && link_target[0]=='/')
       {
 	 if(link_len>1)
@@ -416,7 +485,7 @@ parse_url_again:
    }
 
    bool show_in_list=true;
-   if(icon && link_target[0]=='/')
+   if(icon && (link_target[0]=='/' || link_target[0]=='~'))
       show_in_list=false;  // makes apache listings look better.
 
    skip_len=tag_len;
@@ -428,11 +497,14 @@ parse_url_again:
       int year=-1,month=-1,day=0,hour=0,minute=0;
       char month_name[32]="";
       char size_str[32]="";
+      char perms[10]="";
       const char *more1;
-      char *str;
+      char *str,*str_with_tags;
       int n;
       char *line_add=(char*)alloca(link_len+128+2*1024);
       bool data_available=false;
+      const char *info_string=0;
+      int         info_string_len=0;
 
       if(!a_href)
 	 goto add_file;	// only <a href> tags can have useful info.
@@ -481,19 +553,27 @@ parse_url_again:
       str=string_alloca(eol-more1);
       memcpy(str,more1+1,eol-more1-1);
       str[eol-more1-1]=0;
+      str_with_tags=alloca_strdup(str);
+      remove_tags(str);
 
       // usual apache listing: DD-Mon-YYYY hh:mm size
       n=sscanf(str,"%2d-%3s-%4d %2d:%2d %30s",
 		    &day,month_name,&year,&hour,&minute,size_str);
       if(n==6)
+      {
+	 debug("apache listing matched");
 	 goto got_info;
+      }
 
       hour=0;
       minute=0;
       // unusual apache listing: size DD-Mon-YYYY
-      n=sscanf(str,"%30s %2d-%3s-%4d",size_str,&day,month_name,&year);
-      if(n==4)
+      n=sscanf(str,"%30s %2d-%3s-%d",size_str,&day,month_name,&year);
+      if(n==4 && (size_str[0]=='-' || is_ascii_digit(size_str[0])))
+      {
+	 debug("unusual apache listing matched");
 	 goto got_info;
+      }
 
       char size_unit[7];
       long size;
@@ -508,14 +588,25 @@ parse_url_again:
 	    sprintf(size_str,"%ld",size);
 	 else
 	    sprintf(size_str,"%ld%s",size,size_unit);
+	 debug("Netscape-Proxy 2.53 listing matched");
 	 goto got_info;
       }
-      if(7==sscanf(str,"%3s %3s %d %2d:%2d:%2d %4d",
-	       week_day,month_name,&day,&hour,&minute,&second,&year))
+      n=sscanf(str,"%3s %3s %d %2d:%2d:%2d %4d %s",
+	       week_day,month_name,&day,&hour,&minute,&second,&year,size_str);
+      if(n==7 || (n==8 && !is_ascii_digit(size_str[0])))
       {
 	 strcpy(size_str,"-");
 	 if(!is_directory)
 	    is_sym_link=true;
+	 debug("Netscape-Proxy 2.53 listing matched (dir/symlink)");
+	 goto got_info;
+      }
+      if(n==8) // maybe squid's EPLF listing.
+      {
+	 // skip rest of line, because there may be href to link target.
+	 skip_len=eol-buf+eol_len;
+	 // no symlinks here.
+	 debug("squid EPLF listing matched");
 	 goto got_info;
       }
 
@@ -537,34 +628,61 @@ parse_url_again:
 	    strcpy(size_str,"-");
 	 }
 	 month--;
+	 debug("Mini-Proxy web server listing matched");
 	 goto got_info;
       }
 
-      strcpy(size_str,"-");
+      // Apache Unix-like listing (from apache proxy):
+      //   Perms Nlnk user group size Mon DD (YYYY or hh:mm)
+      int perms_code;
+      int n_links;
+      char user[32];
+      char group[32];
       char year_or_time[6];
+      int consumed;
+
+      n=sscanf(buf,"%10s %d %31s %31s %ld %3s %2d %5s%n",perms,&n_links,
+	    user,group,&size,month_name,&day,year_or_time,&consumed);
+      if(n==8 && -1!=(perms_code=parse_perms(perms+1))
+      && -1!=(month=parse_month(month_name))
+      && -1!=parse_year_or_time(year_or_time,&year,&hour,&minute))
+      {
+	 sprintf(size_str,"%ld",size);
+	 if(perms[0]=='d')
+	    is_directory=true;
+	 else if(perms[0]=='l')
+	 {
+	    is_sym_link=true;
+	    str=string_alloca(more1-more);
+	    memcpy(str,more+1,more1-more-4);
+	    str[more1-more-4]=0;
+	    sym_link=strstr(str," -> ");
+	    if(sym_link)
+	       sym_link+=4;
+	 }
+	 info_string=buf;
+	 info_string_len=consumed;
+	 debug("apache ftp over http proxy listing matched");
+	 goto got_info;
+      }
+      perms[0]=0;
+      size=-1;
+      strcpy(size_str,"-");
+
       // squid's ftp listing: Mon DD (YYYY or hh:mm) [size]
       n=sscanf(str,"%3s %2d %5s %30s",month_name,&day,year_or_time,size_str);
       if(n<3)
 	 goto add_file;
       if(!is_ascii_digit(size_str[0]))
 	 strcpy(size_str,"-");
-      if(year_or_time[2]==':')
-      {
-	 if(2!=sscanf(year_or_time,"%2d:%2d",&hour,&minute))
-	    goto add_file;
-	 year=-1;
-      }
-      else
-      {
-	 if(1!=sscanf(year_or_time,"%d",&year))
-	    goto add_file;
-	 hour=minute=0;
-      }
+      if(-1==parse_year_or_time(year_or_time,&year,&hour,&minute))
+	 goto add_file;
+
       // skip rest of line, because there may be href to link target.
       skip_len=eol-buf+eol_len;
 
       char *ptr;
-      ptr=strstr(str," -> <A HREF=\"");
+      ptr=strstr(str_with_tags," -> <A HREF=\"");
       if(ptr)
       {
 	 is_sym_link=true;
@@ -578,6 +696,7 @@ parse_url_again:
 	    url::decode_string(sym_link);
 	 }
       }
+      debug("squid ftp listing matched");
 
    got_info:
       if(year!=-1)
@@ -598,29 +717,37 @@ parse_url_again:
    add_file:
       if(data_available)
       {
+	 if(info_string)
+	 {
+	    sprintf(line_add,"%.*s %s",info_string_len,info_string,link_target);
+	    goto append_symlink_maybe;
+	 }
 	 if(month==-1)
 	    month=parse_month(month_name);
 	 if(month>=0)
 	 {
 	    sprintf(month_name,"%02d",month+1);
 	    if(year==-1)
-	    {
-	       time_t curr=time(0);
-	       struct tm &now=*localtime(&curr);
-	       year=now.tm_year+1900;
-	       if(month*64+day>now.tm_mon*64+now.tm_mday)
-		  year--;
-	    }
+	       year=guess_year(month,day);
+	 }
+	 if(perms[0]==0)
+	 {
+	    if(is_directory)
+	       strcpy(perms,"drwxr-xr-x");
+	    else if(is_sym_link)
+	       strcpy(perms,"lrwxrwxrwx");
+	    else
+	       strcpy(perms,"-rw-r--r--");
 	 }
 	 sprintf(line_add,"%s  %10s  %04d-%s-%02d %02d:%02d  %s",
-	    is_directory?"drwxr-xr-x":(is_sym_link?"lrwxrwxrwx":"-rw-r--r--"),
-	    size_str,year,month_name,day,hour,minute,link_target);
+	    perms,size_str,year,month_name,day,hour,minute,link_target);
+      append_symlink_maybe:
 	 if(sym_link)
 	    sprintf(line_add+strlen(line_add)," -> %s",sym_link);
       }
       else
       {
-	 sprintf(line_add,"%s    %s",
+	 sprintf(line_add,"%s  --  %s",
 	    is_directory?"drwxr-xr-x":"-rw-r--r--",link_target);
       }
       if(lsopt && lsopt->append_type)
@@ -695,7 +822,7 @@ int HttpDirList::Do()
 	 buf->Put(":\n");
       }
 
-      char *cache_buffer=0;
+      const char *cache_buffer=0;
       int cache_buffer_size=0;
       if(use_cache && LsCache::Find(session,curr,mode,
 				    &cache_buffer,&cache_buffer_size))
@@ -703,7 +830,6 @@ int HttpDirList::Do()
 	 ubuf=new Buffer();
 	 ubuf->Put(cache_buffer,cache_buffer_size);
 	 ubuf->PutEOF();
-	 xfree(cache_buffer);
       }
       else
       {
@@ -732,7 +858,7 @@ int HttpDirList::Do()
    {
       LsCache::Add(session,curr,mode,ubuf);
 
-      delete ubuf;
+      Delete(ubuf);
       ubuf=0;
       return MOVED;
    }
@@ -792,7 +918,7 @@ HttpDirList::HttpDirList(ArgV *a,FileAccess *fa)
 
 HttpDirList::~HttpDirList()
 {
-   delete ubuf;
+   Delete(ubuf);
    if(curr_url)
       delete curr_url;
    xfree(base_href);
@@ -866,10 +992,8 @@ HttpGlob::HttpGlob(FileAccess *s,const char *n_pattern)
 
 HttpGlob::~HttpGlob()
 {
-   if(updir_glob)
-      delete updir_glob;
-   if(ubuf)
-      delete ubuf;
+   Delete(updir_glob);
+   Delete(ubuf);
    xfree(dir);
 }
 
@@ -907,7 +1031,7 @@ int HttpGlob::Do()
 	 return MOVED;
       }
 
-      char *cache_buffer=0;
+      const char *cache_buffer=0;
       int cache_buffer_size=0;
       if(use_cache && LsCache::Find(session,curr_dir,FA::LONG_LIST,
 				    &cache_buffer,&cache_buffer_size))
@@ -915,12 +1039,12 @@ int HttpGlob::Do()
 	 ubuf=new Buffer();
 	 ubuf->Put(cache_buffer,cache_buffer_size);
 	 ubuf->PutEOF();
-	 xfree(cache_buffer);
       }
       else
       {
 	 session->Open(curr_dir,FA::LONG_LIST);
 	 session->UseCache(use_cache);
+	 session->RereadManual();
 	 ubuf=new FileInputBuffer(session);
 	 if(LsCache::IsEnabled())
 	    ubuf->Save(LsCache::SizeLimit());
@@ -931,13 +1055,21 @@ int HttpGlob::Do()
    if(ubuf->Error())
    {
       SetError(ubuf->ErrorText());
-      delete ubuf;
+      Delete(ubuf);
       ubuf=0;
       return MOVED;
    }
 
    if(!ubuf->Eof())
+   {
+      if(session->GetRealPos()==0 && session->GetPos()>0)
+      {
+	 session->SeekReal();
+	 ubuf->Empty();
+	 return MOVED;
+      }
       return m;
+   }
 
    // now we have all the index in ubuf; parse it.
    const char *b;
@@ -959,7 +1091,7 @@ int HttpGlob::Do()
 
    LsCache::Add(session,curr_dir,FA::LONG_LIST,ubuf);
 
-   delete ubuf;
+   Delete(ubuf);
    ubuf=0;
 
    set.rewind();
@@ -1008,7 +1140,7 @@ int HttpListInfo::Do()
 
    if(!ubuf && !get_info)
    {
-      char *cache_buffer=0;
+      const char *cache_buffer=0;
       int cache_buffer_size=0;
       if(use_cache && LsCache::Find(session,"",FA::LONG_LIST,
 				    &cache_buffer,&cache_buffer_size))
@@ -1016,7 +1148,6 @@ int HttpListInfo::Do()
 	 ubuf=new Buffer();
 	 ubuf->Put(cache_buffer,cache_buffer_size);
 	 ubuf->PutEOF();
-	 xfree(cache_buffer);
       }
       else
       {
@@ -1033,7 +1164,7 @@ int HttpListInfo::Do()
       if(ubuf->Error())
       {
 	 SetError(ubuf->ErrorText());
-	 delete ubuf;
+	 Delete(ubuf);
 	 ubuf=0;
 	 return MOVED;
       }
@@ -1061,7 +1192,7 @@ int HttpListInfo::Do()
 
       LsCache::Add(session,"",FA::LONG_LIST,ubuf);
 
-      delete ubuf;
+      Delete(ubuf);
       ubuf=0;
       m=MOVED;
 
@@ -1170,8 +1301,7 @@ HttpListInfo::~HttpListInfo()
 {
    session->Close();
    xfree(get_info);
-   if(ubuf)
-      delete ubuf;
+   Delete(ubuf);
 }
 
 const char *HttpListInfo::Status()
