@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Job.cc,v 1.7 1999/09/15 15:10:40 lav Exp $ */
+/* $Id: Job.cc,v 1.13 2000/03/01 23:28:31 lav Exp $ */
 
 #include <config.h>
 #include <stdio.h>
@@ -64,7 +64,7 @@ Job::~Job()
 	    }
 	    else
 	    {
-	       delete scan;
+	       Delete(scan);
 	       scan=chain;
 	    }
 	 }
@@ -84,8 +84,7 @@ Job::~Job()
       }
    }
 
-   if(cmdline)
-      free(cmdline);
+   xfree(cmdline);
    if(fg_data)
       delete fg_data;
 }
@@ -123,7 +122,7 @@ void Job::Kill(int n)
 	 j->cmdline=0;
 	 j->parent->waiting=r;
       }
-      delete j;
+      Delete(j);
    }
 }
 
@@ -132,7 +131,7 @@ void Job::KillAll()
    for(Job **scan=&chain; *scan; )
    {
       if((*scan)->jobno>=0)
-	 delete *scan;
+	 Delete(*scan);
       else
 	 scan=&(*scan)->next;
    }
@@ -190,22 +189,47 @@ void  Job::SortJobs()
    }
 }
 
+void Job::PrintJobTitle(int indent,const char *suffix)
+{
+   if(jobno<0 && cmdline==0)
+      return;
+   printf("%*s",indent,"");
+   if(jobno>=0)
+      printf("[%d] ",jobno);
+   printf("%s",cmdline?cmdline:"?");
+   if(suffix)
+      printf(" %s",suffix);
+   printf("\n");
+}
+
+void Job::ListOneJob(int verbose,int indent,const char *suffix)
+{
+   PrintJobTitle(indent,suffix);
+   PrintStatus(verbose);
+   if(waiting && waiting->jobno<0 && waiting->cmdline==0)
+      waiting->ListOneJob(verbose,indent+1);
+}
+
+void Job::ListOneJobRecursively(int verbose,int indent)
+{
+   PrintJobTitle(indent);
+   PrintStatus(verbose);
+   ListJobs(verbose,indent+1);
+}
+
 void  Job::ListJobs(int verbose,int indent)
 {
    if(indent==0)
       SortJobs();
 
+   // list the foreground job first.
+   if(waiting && waiting!=this && waiting->parent==this)
+      waiting->ListOneJobRecursively(verbose,indent);
+
    for(Job *scan=chain; scan; scan=scan->next)
    {
-      if(!scan->Done() && scan->parent==this)
-      {
-	 printf("%*s",indent,"");
-	 if(scan->jobno>=0)
-	    printf("[%d] ",scan->jobno);
-	 printf("%s\n",scan->cmdline?scan->cmdline:"?");
-	 scan->PrintStatus(verbose);
-	 scan->ListJobs(verbose,indent+1);
-      }
+      if(!scan->Done() && scan->parent==this && waiting!=scan)
+	 scan->ListOneJobRecursively(verbose,indent);
    }
 }
 
@@ -238,7 +262,7 @@ void  Job::BuryDoneJobs()
       if((scan->parent==this || scan->parent==0) && scan->jobno>=0
 		  && scan->Done())
       {
-	 delete scan;
+	 Delete(scan);
       	 scan=chain;
       }
       else
@@ -299,6 +323,19 @@ void SessionJob::PrintStatus(int v)
       printf("\t%s\n",url);
 }
 
+void SessionJob::Fg()
+{
+   if(session)
+      session->SetPriority(1);
+   Job::Fg();
+}
+void SessionJob::Bg()
+{
+   Job::Bg();
+   if(session)
+      session->SetPriority(0);
+}
+
 void Job::vfprintf(FILE *file,const char *fmt,va_list v)
 {
    if(file!=stdout && file!=stderr)
@@ -358,7 +395,7 @@ int Job::AcceptSig(int s)
       {
 	 Job *new_waiting=waiting->waiting;
 	 waiting->waiting=0;
-	 delete waiting;
+	 Delete(waiting);
 	 waiting=new_waiting;
       }
    }
