@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Job.h,v 1.5 1999/08/16 21:23:33 lav Exp $ */
+/* $Id: Job.h,v 1.10 2000/03/24 18:53:24 lav Exp $ */
 
 #ifndef JOB_H
 #define JOB_H
@@ -40,15 +40,24 @@ class Job : public SMTask
    static Job *chain;
 
 protected:
-   void SetParent(Job *j) { parent=j; }
    bool fg;
    FgData *fg_data;
 
+   virtual ~Job();
+
 public:
    int	 jobno;
    Job	 *parent;
    Job	 *waiting;
 
+   void SetParent(Job *j) { parent=j; }
+   void SetParentFg(Job *j, bool f=true)
+      {
+	 SetParent(j);
+	 if(j->fg && f)
+	    Fg();
+      }
+
    void	 AllocJobno();
 
    virtual void	  PrintStatus(int) {};
@@ -63,6 +72,9 @@ public:
 
    char	 *cmdline;
    virtual void ListJobs(int verbose_level,int indent=0);
+   void PrintJobTitle(int indent=0,const char *suffix=0);
+   void ListOneJob(int verbose,int indent=0,const char *suffix=0);
+   void ListOneJobRecursively(int verbose,int indent);
    void ListDoneJobs();
    void BuryDoneJobs();
 
@@ -89,7 +101,6 @@ public:
    void puts(const char *);
 
    Job();
-   virtual ~Job();
 
    virtual const char *GetConnectURL() { return 0; }
 };
@@ -112,6 +123,8 @@ public:
 	    return 0;
 	 return session->GetConnectURL();
       }
+   void Fg();
+   void Bg();
 };
 
 #endif /* JOB_H */
