@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: LocalAccess.cc,v 1.12.2.1 1999/12/16 20:02:39 lav Exp $ */
+/* $Id: LocalAccess.cc,v 1.24.2.1 2000/06/08 10:13:31 lav Exp $ */
 
 /* this is not very useful, just a proof of concept */
 /* this is used in mirror command to build local files list */
@@ -38,6 +38,9 @@
 #include "xalloca.h"
 #include "xstring.h"
 #include "misc.h"
+#include "log.h"
+
+FileAccess *LocalAccess::New() { return new LocalAccess(); }
 
 void LocalAccess::ClassInit()
 {
@@ -47,18 +50,22 @@ void LocalAccess::ClassInit()
 
 void LocalAccess::Init()
 {
-   done=true;
+   done=false;
    error_code=OK;
    stream=0;
    xfree(home);
    home=xstrdup(getenv("HOME"));
+   xfree(hostname);
+   hostname=xstrdup("localhost");
 }
 
 LocalAccess::LocalAccess() : FileAccess()
 {
    Init();
+   xfree(cwd);
+   cwd=xgetcwd();
 }
-LocalAccess::LocalAccess(const LocalAccess *o) : FileAccess((const FileAccess *)o)
+LocalAccess::LocalAccess(const LocalAccess *o) : FileAccess(o)
 {
    Init();
 }
@@ -77,6 +84,7 @@ void LocalAccess::errno_handle()
       sprintf(error,"rename(%s, %s): %s",file,file1,err);
    else
       sprintf(error,"%s: %s",file,err);
+   Log::global->Format(0,"**** %s\n",error);
 }
 
 int LocalAccess::Done()
@@ -100,6 +108,8 @@ int LocalAccess::Do()
    if(Error() || done)
       return STALL;
    int m=STALL;
+   if(mode!=CLOSED)
+      ExpandTildeInCWD();
    switch((open_mode)mode)
    {
    case(CLOSED):
@@ -110,6 +120,7 @@ int LocalAccess::Do()
       if(stream==0)
       {
 	 char *cmd=(char*)alloca(10+xstrlen(file));
+	 // FIXME: shell-quote file name
 	 if(mode==LIST)
 	 {
 	    if(file && file[0])
@@ -126,7 +137,7 @@ int LocalAccess::Do()
 	 }
 	 else// if(mode==QUOTE_CMD)
 	    strcpy(cmd,file);
-	 DebugPrint("---- ",cmd);
+	 DebugPrint("---- ",cmd,5);
 	 InputFilter *f_stream=new InputFilter(cmd);
 	 f_stream->SetCwd(cwd);
 	 stream=f_stream;
@@ -140,11 +151,11 @@ int LocalAccess::Do()
 	    Fatal(stream->error_text);
 	    return MOVED;
 	 }
-	 block+=TimeOut(1000);
+	 TimeoutS(1);
 	 return m;
       }
       stream->Kill(SIGCONT);
-      block+=PollVec(stream->getfd(),POLLIN);
+      Block(stream->getfd(),POLLIN);
       return m;
    case(CHANGE_DIR):
       if(access(file,X_OK)==-1)
@@ -159,14 +170,17 @@ int LocalAccess::Do()
       }
       done=true;
       return MOVED;
-   case(REMOVE):
-      if(remove(dir_file(cwd,file))==-1)
+   case(REMOVE): {
+      const char *f=dir_file(cwd,file);
+      Log::global->Format(5,"---- remove(%s)\n",f);
+      if(remove(f)==-1)
       {
 	 errno_handle();
 	 error_code=NO_FILE;
       }
       done=true;
       return MOVED;
+   }
    case(REMOVE_DIR):
       if(rmdir(dir_file(cwd,file))==-1)
       {
@@ -226,7 +240,7 @@ int LocalAccess::Do()
 	    SetError(NO_FILE,stream->error_text);
 	    return MOVED;
 	 }
-	 block+=TimeOut(1000);
+	 TimeoutS(1);
 	 return m;
       }
       stream->Kill(SIGCONT);
@@ -246,7 +260,7 @@ int LocalAccess::Do()
 	 opt_size=0;
 	 opt_date=0;
       }
-      block+=PollVec(stream->getfd(),(mode==STORE?POLLOUT:POLLIN));
+      Block(stream->getfd(),(mode==STORE?POLLOUT:POLLIN));
       return m;
 
    case(CONNECT_VERIFY):
@@ -291,14 +305,22 @@ int LocalAccess::Read(void *buf,int size)
       return DO_AGAIN;
    if(real_pos==-1)
    {
-      if(lseek(fd,pos,SEEK_SET)==-1)
+      if(ascii || lseek(fd,pos,SEEK_SET)==-1)
 	 real_pos=0;
       else
 	 real_pos=pos;
    }
    stream->Kill(SIGCONT);
 read_again:
-   int res=read(fd,buf,size);
+   int res;
+
+#ifndef NATIVE_CRLF
+   if(ascii)
+      res=read(fd,buf,size/2);
+   else
+#endif
+      res=read(fd,buf,size);
+
    if(res<0)
    {
       if(errno==EINTR || errno==EAGAIN)
@@ -308,10 +330,28 @@ read_again:
    }
    if(res==0)
       return res; // eof
+
+#ifndef NATIVE_CRLF
+   if(ascii)
+   {
+      char *p=(char*)buf;
+      for(int i=res; i>0; i--)
+      {
+	 if(*p=='\n')
+	 {
+	    memmove(p+1,p,i);
+	    *p++='\r';
+	    res++;
+	 }
+	 p++;
+      }
+   }
+#endif
+
    real_pos+=res;
    if(real_pos<=pos)
       goto read_again;
-   int shift;
+   long shift;
    if((shift=pos+res-real_pos)>0)
    {
       memmove(buf,(char*)buf+shift,size-shift);
@@ -321,8 +361,9 @@ read_again:
    return(res);
 }
 
-int LocalAccess::Write(const void *buf,int size)
+int LocalAccess::Write(const void *vbuf,int len)
 {
+   const char *buf=(const char *)vbuf;
    if(error_code<0)
       return error_code;
    if(stream==0)
@@ -332,7 +373,7 @@ int LocalAccess::Write(const void *buf,int size)
       return DO_AGAIN;
    if(real_pos==-1)
    {
-      if(lseek(fd,pos,SEEK_SET)==-1)
+      if(ascii || lseek(fd,pos,SEEK_SET)==-1)
 	 real_pos=0;
       else
 	 real_pos=pos;
@@ -344,7 +385,43 @@ int LocalAccess::Write(const void *buf,int size)
    }
    stream->Kill(SIGCONT);
 
-   int res=write(fd,buf,size);
+   int skip_cr=0;
+
+#ifndef NATIVE_CRLF
+   if(ascii)
+   {
+      // find where line ends.
+      const char *cr=buf;
+      for(;;)
+      {
+	 cr=(const char *)memchr(cr,'\r',len-(cr-buf));
+	 if(!cr)
+	    break;
+	 if(cr-buf<len-1 && cr[1]=='\n')
+	 {
+	    skip_cr=1;
+	    len=cr-buf;
+	    break;
+	 }
+	 if(cr-buf==len-1)
+	 {
+	    if(len==1)	   // last CR in stream will be lost. (FIX?)
+	       skip_cr=1;
+	    len--;
+	    break;
+	 }
+	 cr++;
+      }
+   }
+#endif	 // NATIVE_CRLF
+
+   if(len==0)
+   {
+      pos=(real_pos+=skip_cr);
+      return skip_cr;
+   }
+
+   int res=write(fd,buf,len);
    if(res<0)
    {
       if(errno==EINTR || errno==EAGAIN)
@@ -353,6 +430,8 @@ int LocalAccess::Write(const void *buf,int size)
       return SEE_ERRNO;
    }
 
+   if(res==len)
+      res+=skip_cr;
    pos=(real_pos+=res);
    return res;
 }
@@ -497,18 +576,28 @@ check_again:
 
 Glob *LocalAccess::MakeGlob(const char *pattern)
 {
-   return new LocalGlob(cwd,pattern);
+   xfree(file);
+   file=xstrdup(pattern);
+   ExpandTildeInCWD();
+   return new LocalGlob(cwd,file);
 }
 
-LocalGlob::LocalGlob(const char *cwd,const char *pattern)
+LocalGlob::LocalGlob(const char *c,const char *pattern)
    : Glob(pattern)
 {
+   cwd=c;
+}
+int LocalGlob::Do()
+{
+   if(done)
+      return STALL;
+
    glob_t g;
    char *oldcwd=xgetcwd();
    if(!oldcwd)
    {
       SetError("cannot get current directory");
-      return;
+      return MOVED;
    }
    if(chdir(cwd)==-1)
    {
@@ -517,27 +606,74 @@ LocalGlob::LocalGlob(const char *cwd,const char *pattern)
       sprintf(err,"chdir(%s): %s",cwd,se);
       SetError(err);
       xfree(oldcwd);
-      return;
+      return MOVED;
    }
 
    glob(pattern, 0, NULL, &g);
 
+   for(unsigned i=0; i<g.gl_pathc; i++)
+   {
+      struct stat st;
+      if(stat(g.gl_pathv[i],&st)!=-1)
+      {
+	 if(dirs_only && !S_ISDIR(st.st_mode))
+	    continue;
+	 if(files_only && !S_ISREG(st.st_mode))
+	    continue;
+      }
+      add(g.gl_pathv[i]);
+   }
+   globfree(&g);
+
    if(chdir(oldcwd)==-1)
       fprintf(stderr,"chdir(%s): %s",oldcwd,strerror(errno));
    xfree(oldcwd);
-   for(unsigned i=0; i<g.gl_pathc; i++)
-      add(g.gl_pathv[i],strlen(g.gl_pathv[i]));
-   globfree(&g);
+
+   done=true;
+   return MOVED;
 }
-int LocalGlob::Do()
+
+DirList *LocalAccess::MakeDirList(ArgV *a)
 {
-   if(!done)
+   return new LocalDirList(a,cwd);
+}
+
+#include "ArgV.h"
+LocalDirList::LocalDirList(ArgV *a,const char *cwd)
+   : DirList(a)
+{
+   fg_data=0;
+   a->insarg(1,"-l");
+   InputFilter *f=new InputFilter(a);
+   f->SetCwd(cwd);
+   Delete(buf);
+   buf=new FileInputBuffer(f);
+}
+int LocalDirList::Do()
+{
+   if(done)
+      return STALL;
+
+   if(buf->Eof())
    {
       done=true;
       return MOVED;
    }
+
+   if(buf->Error())
+   {
+      SetError(buf->ErrorText());
+      return MOVED;
+   }
+   if(!fg_data)
+      fg_data=buf->GetFgData(false);
    return STALL;
 }
+LocalDirList::~LocalDirList()
+{
+   if(fg_data)
+      delete fg_data;
+}
 
 #ifdef MODULE
 CDECL void module_init()
