@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: MirrorJob.cc,v 1.28.2.1 2000/01/21 19:41:12 lav Exp $ */
+/* $Id: MirrorJob.cc,v 1.47.2.1 2000/07/12 12:00:06 lav Exp $ */
 
 #include <config.h>
 
@@ -31,12 +31,12 @@
 #include "CmdExec.h"
 #include "rmJob.h"
 #include "mkdirJob.h"
-#include "PutJob.h"
 #include "ChmodJob.h"
 #include "misc.h"
 #include "xalloca.h"
 #include "plural.h"
 #include "getopt.h"
+#include "FindJob.h"
 
 void  MirrorJob::PrintStatus(int v)
 {
@@ -47,12 +47,43 @@ void  MirrorJob::PrintStatus(int v)
    else
       tab="";
 
-   if(!Done())
-      return;
+   if(Done())
+      goto final;
 
-   printf(plural(N_("%sTotal: %d director$y|ies$, %d file$|s$, %d symlink$|s$\n"),
+   switch(state)
+   {
+   case(INITIAL_STATE):
+   case(DONE):
+   case(WAITING_FOR_SUBGET):
+   case(WAITING_FOR_SUBMIRROR):
+   case(WAITING_FOR_RM_BEFORE_PUT):
+   case(WAITING_FOR_MKDIR_BEFORE_SUBMIRROR):
+   case(REMOTE_REMOVE_OLD):
+   case(REMOTE_CHMOD):
+      break;
+
+   case(MAKE_REMOTE_DIR):
+      printf("\tmkdir `%s' [%s]\n",remote_dir,session->CurrentStatus());
+      break;
+
+   case(CHANGING_REMOTE_DIR):
+      printf("\tcd `%s' [%s]\n",remote_dir,session->CurrentStatus());
+      break;
+
+   case(GETTING_LIST_INFO):
+      if(remote_relative_dir)
+	 printf("\t%s: %s\n",remote_relative_dir,list_info->Status());
+      else
+	 printf("\t%s\n",list_info->Status());
+      break;
+   }
+   return;
+
+final:
+   if(dirs>0)
+      printf(plural(N_("%sTotal: %d director$y|ies$, %d file$|s$, %d symlink$|s$\n"),
 		     dirs,tot_files,tot_symlinks),
-      tab,dirs,tot_files,tot_symlinks);
+	 tab,dirs,tot_files,tot_symlinks);
    if(new_files || new_symlinks)
       printf(plural(N_("%sNew: %d file$|s$, %d symlink$|s$\n"),
 		     new_files,new_symlinks),
@@ -67,12 +98,11 @@ void  MirrorJob::PrintStatus(int v)
 	      :N_("%sTo be removed: %d director$y|ies$, %d file$|s$, %d symlink$|s$\n"),
 	      del_dirs,del_files,del_symlinks),
 	 tab,del_dirs,del_files,del_symlinks);
+   return;
 }
 
 void  MirrorJob::ShowRunStatus(StatusLine *s)
 {
-   const char *status;
-
    switch(state)
    {
    case(INITIAL_STATE):
@@ -99,9 +129,11 @@ void  MirrorJob::ShowRunStatus(StatusLine *s)
       break;
 
    case(GETTING_LIST_INFO):
-      status=list_info->Status();
-      if(status[0])
-	 s->Show("%s [%s]",status,session->CurrentStatus());
+      if(remote_relative_dir)
+	 s->Show("%s: %s",squeeze_file_name(remote_relative_dir,20),
+	    list_info->Status());
+      else
+	 s->Show("%s",list_info->Status());
       break;
    }
 }
@@ -111,7 +143,6 @@ void  MirrorJob::HandleFile(int how)
    ArgV *args;
    int	 res;
    mode_t mode;
-   GetJob   *gj;
    struct stat st;
 
    // dir_name returns pointer to static data - need to dup it.
@@ -150,7 +181,7 @@ void  MirrorJob::HandleFile(int how)
 		  }
 	       }
 	       waiting=new rmJob(Clone(),args);
-	       waiting->parent=this;
+	       waiting->SetParentFg(this);
 	       waiting->cmdline=args->Combine();
 	       mod_files++;
 	    }
@@ -174,7 +205,12 @@ void  MirrorJob::HandleFile(int how)
 	    && (file->defined&(file->DATE|file->DATE_UNPREC))
 	    && file->date + prec.Seconds() < st.st_mtime
 	    && (file->defined&file->SIZE) && file->size >= st.st_size)
+	    {
 	       cont_this=true;
+	       // make it writable
+	       if((st.st_mode&0200)==0)
+		  chmod(local_name,st.st_mode|0200);
+	    }
 	    else
 	    {
 	       Report(_("Removing old local file `%s'"),
@@ -197,29 +233,39 @@ void  MirrorJob::HandleFile(int how)
 	 // launch get job
 	 Report(_("Retrieving remote file `%s'"),
 		  dir_file(remote_relative_dir,file->name));
-	 args=new ArgV("get");
-	 args->Append(file->name);
 	 if(script)
 	 {
+	    args=new ArgV("get1");
+	    if(cont_this)
+	       args->Append("-c");
+	    args->Append(file->name);
+	    args->Append("-o");
+	    args->Append(local_name);
 	    char *cmd=args->CombineQuoted();
 	    fprintf(script,"%s",cmd);
 	    xfree(cmd);
+	    delete args; args=0;
 	    if(script_only)
-	    {
-	       delete args;
 	       goto skip;
-	    }
 	 }
-	 args->Append(local_name);
-	 gj=new GetJob(Clone(),args,cont_this);
+
+	 FileCopyPeerFA *src_peer=
+	    new FileCopyPeerFA(session,file->name,FA::RETRIEVE);
+	 src_peer->DontReuseSession(); // mirror will need this session
+	 FileCopyPeer *dst_peer=
+	    FileCopyPeerFDStream::NewPut(local_name,cont_this);
+
+	 FileCopy *c=FileCopy::New(src_peer,dst_peer,cont_this);
+	 if(remove_source_files)
+	    c->RemoveSourceLater();
+	 CopyJob *cp=
+	    new CopyJob(c,file->name,"mirror");
 	 if(file->defined&(file->DATE|file->DATE_UNPREC))
-	    gj->SetTime(file->date);
-#if 0
+	    cp->SetDate(file->date);
 	 if(file->defined&file->SIZE)
-	    gj->SetSize(file->size);
-#endif
-	 waiting=gj;
-	 waiting->parent=this;
+	    cp->SetSize(file->size);
+	 waiting=cp;
+	 waiting->SetParentFg(this);
 	 waiting->cmdline=(char*)xmalloc(10+strlen(file->name));
 	 sprintf(waiting->cmdline,"\\get %s",file->name);
 	 break;
@@ -244,7 +290,7 @@ void  MirrorJob::HandleFile(int how)
 		  args->Append("--");
 		  args->Append(file->name);
 		  waiting=new mkdirJob(Clone(),args);
-		  waiting->parent=this;
+		  waiting->SetParentFg(this);
 		  waiting->cmdline=args->Combine();
 	       }
 	       else
@@ -256,41 +302,47 @@ void  MirrorJob::HandleFile(int how)
 	 else // !REVERSE
 	 {
 	    mode=((file->defined&file->MODE)&&!(flags&NO_PERMS))?file->mode:0755;
-	    res=mkdir(local_name,mode|0700);
-	    if(res==-1 && errno==EEXIST)
+	    struct stat st;
+	    if(lstat(local_name,&st)!=-1)
 	    {
-	       struct stat st;
-	       if(stat(local_name,&st)!=-1)
+	       if(S_ISDIR(st.st_mode))
+		  chmod(local_name,st.st_mode|0700);
+	       else
 	       {
-		  if(S_ISDIR(st.st_mode))
-		     chmod(local_name,st.st_mode|0700);
-	    	  else
+		  Report(_("Removing old local file `%s'"),
+			   dir_file(local_relative_dir,file->name));
+		  if(remove(local_name)==-1)
 		  {
-		     Report(_("Removing old local file `%s'"),
-			      dir_file(local_relative_dir,file->name));
-		     remove(local_name);
-		     res=mkdir(local_name,mode|0700);
+		     eprintf("mirror: remove(%s): %s\n",local_name,strerror(errno));
+		     goto skip;
 		  }
+		  goto do_mkdir;
 	       }
 	    }
-	    else if(res==-1) // other error
+	    else // no such directory
 	    {
-	       eprintf("mirror: mkdir(%s): %s\n",local_name,strerror(errno));
-	       goto skip;
+	    do_mkdir:
+	       res=mkdir(local_name,mode|0700);
+	       if(res==-1)
+	       {
+		  eprintf("mirror: mkdir(%s): %s\n",local_name,strerror(errno));
+		  goto skip;
+	       }
 	    }
 	 }
 	 // launch sub-mirror
 	 MirrorJob *mj=new MirrorJob(Clone(),local_name,remote_name);
 	 mj->parent_mirror=this;
 	 waiting=mj;
-	 waiting->parent=this;
+	 waiting->SetParentFg(this);
 	 waiting->cmdline=(char*)xmalloc(strlen("\\mirror")+1+
 					 strlen(file->name)+1);
 	 sprintf(waiting->cmdline,"\\mirror %s",file->name);
 
 	 // inherit flags and other things
 	 mj->SetFlags(flags,1);
-	 mj->SetPrec(prec);
+	 mj->SetPrec(prec,loose_prec);
+	 mj->UseCache(use_cache);
 
 	 if(rx_include)	mj->SetInclude(rx_include);
 	 if(rx_exclude)	mj->SetExclude(rx_exclude);
@@ -303,6 +355,10 @@ void  MirrorJob::HandleFile(int how)
 	 mj->verbose_report=verbose_report;
 	 mj->newer_than=newer_than;
 
+	 if(verbose_report>=3)
+	    Report(_("Mirroring directory `%s'"),
+	       (flags&REVERSE)?mj->local_relative_dir:mj->remote_relative_dir);
+
 	 dir_made=false;   // for next directory
 
 	 break;
@@ -325,7 +381,11 @@ void  MirrorJob::HandleFile(int how)
 	       Report(_("Removing old local file `%s'"),
 			dir_file(local_relative_dir,file->name));
 	       mod_symlinks++;
-	       remove(local_name);
+	       if(remove(local_name)==-1)
+	       {
+		  eprintf("mirror: remove(%s): %s\n",local_name,strerror(errno));
+		  goto skip;
+	       }
 	    }
 	    else
 	    {
@@ -335,7 +395,7 @@ void  MirrorJob::HandleFile(int how)
 		     dir_file(local_relative_dir,file->name),file->symlink);
 	    res=symlink(file->symlink,local_name);
 	    if(res==-1)
-	       perror(local_name);
+	       eprintf("mirror: symlink(%s): %s\n",local_name,strerror(errno));
 	 }
 #endif /* LSTAT */
 	 goto skip;
@@ -369,10 +429,7 @@ void  MirrorJob::InitSets(FileSet *source,FileSet *dest)
 
    same=new FileSet(source);
    to_transfer=new FileSet(source);
-   int ignore=0;
-   if(prec.IsInfty())
-      ignore|=FileInfo::DATE;
-   to_transfer->SubtractSame(dest,flags&ONLY_NEWER,prec.Seconds(),ignore);
+   to_transfer->SubtractSame(dest,flags&ONLY_NEWER,&prec,&loose_prec,0);
 
    same->SubtractAny(to_transfer);
 
@@ -408,22 +465,22 @@ int   MirrorJob::Do()
 	 }
 	 local_session->Chdir(local_dir,false);
 	 list_info=local_session->MakeListInfo();
-	 list_info->UseCache(false);
+	 list_info->UseCache(use_cache);
 	 if(flags&RETR_SYMLINKS)
 	    list_info->FollowSymlinks();
 	 list_info->SetExclude(local_relative_dir,
 			rx_exclude?&rxc_exclude:0,rx_include?&rxc_include:0);
 
 	 while(!list_info->Done())
-	    list_info->Do();  // this should be fast
+	    Roll(list_info);  // this should be fast
 
 	 if(list_info->Error())
 	    goto list_info_error;
 
 	 local_set=list_info->GetResult();
-	 delete list_info;
+	 Delete(list_info);
 	 list_info=0;
-	 delete local_session;
+	 Delete(local_session);
 	 local_session=0;
 
 	 local_set->ExcludeDots();  // don't need .. and .
@@ -436,7 +493,7 @@ int   MirrorJob::Do()
 	 return MOVED;
       }
       session->Chdir(remote_dir);
-      while(session->Do()==MOVED);
+      Roll(session);
       state=CHANGING_REMOTE_DIR;
       return MOVED;
 
@@ -472,16 +529,14 @@ int   MirrorJob::Do()
       	 state=DONE;
       	 return MOVED;
       }
-      list_info->UseCache(false);
+      list_info->UseCache(use_cache);
       list_info->Need(FileInfo::ALL_INFO);
       if(flags&RETR_SYMLINKS)
 	 list_info->FollowSymlinks();
 
       list_info->SetExclude(remote_relative_dir,
 		     rx_exclude?&rxc_exclude:0,rx_include?&rxc_include:0);
-      while(list_info->Do()==MOVED)
-	 ;
-
+      Roll(list_info);
       state=GETTING_LIST_INFO;
       return MOVED;
 
@@ -494,17 +549,18 @@ int   MirrorJob::Do()
       list_info_error:
 	 eprintf("mirror: %s\n",list_info->ErrorText());
 	 state=DONE;
-      	 delete list_info;
+      	 Delete(list_info);
 	 list_info=0;
 	 return MOVED;
       }
       remote_set=list_info->GetResult();
-      delete list_info;
+      Delete(list_info);
       list_info=0;
 
       remote_set->ExcludeDots(); // don't need .. and .
 
       // now we have both local and remote file sets.
+      dirs++;
 
       if(flags&REVERSE)
 	 InitSets(local_set,remote_set);
@@ -523,18 +579,43 @@ int   MirrorJob::Do()
       {
 	 if(!waiting->Done())
 	    return STALL;
-	 delete waiting;
+	 Delete(waiting);
 	 waiting=0;
       }
       Report(_("Sending local file `%s'"),
 	       dir_file(local_relative_dir,file->name));
       const char *local_name=dir_file(local_dir,file->name);
-      ArgV *args=new ArgV("put");
-      args->Append(local_name);
-      args->Append(file->name);
-      waiting=new PutJob(Clone(),args);
-      waiting->parent=this;
-      waiting->cmdline=(char*)xmalloc(6+strlen(file->name));
+
+#if 0 // unfinished
+      if(script)
+      {
+	 args=new ArgV("put1");
+	 args->Append(local_name);
+	 args->Append("-o");
+	 args->Append(file->name);
+	 char *cmd=args->CombineQuoted();
+	 fprintf(script,"%s",cmd);
+	 xfree(cmd);
+	 delete args; args=0;
+	 if(script_only)
+	    goto skip;
+      }
+#endif
+
+      FileCopyPeerFA *dst_peer=
+	 new FileCopyPeerFA(session,file->name,FA::STORE);
+      dst_peer->DontReuseSession(); // mirror won't need session
+      FileCopyPeer *src_peer=
+	 FileCopyPeerFDStream::NewGet(local_name);
+
+      FileCopy *c=FileCopy::New(src_peer,dst_peer,false);
+      if(remove_source_files)
+	 c->RemoveSourceLater();
+      CopyJob *cp=
+	 new CopyJob(c,file->name,"mirror");
+      waiting=cp;
+      waiting->SetParentFg(this);
+      waiting->cmdline=(char*)xmalloc(10+strlen(file->name));
       sprintf(waiting->cmdline,"\\put %s",file->name);
       state=WAITING_FOR_SUBGET;
       return MOVED;
@@ -544,7 +625,7 @@ int   MirrorJob::Do()
       if(waiting && waiting->Done())
       {
 	 to_transfer->next();
-	 delete waiting;
+	 Delete(waiting);
 	 waiting=0;
       }
       while(!waiting && state==WAITING_FOR_SUBGET)
@@ -568,7 +649,7 @@ int   MirrorJob::Do()
 	    return STALL;
 	 if(waiting->ExitCode()==0)
 	    dir_made=true;
-	 delete waiting;
+	 Delete(waiting);
 	 waiting=0;
 	 if(!dir_made)
 	    to_transfer->next();
@@ -592,7 +673,7 @@ int   MirrorJob::Do()
 	 del_dirs+=mj.del_dirs;
 
 	 to_transfer->next();
-	 delete waiting;
+	 Delete(waiting);
 	 waiting=0;
       }
       while(!waiting && state==WAITING_FOR_SUBMIRROR)
@@ -604,28 +685,6 @@ int   MirrorJob::Do()
 	    to_rm->rewind();
 	    if(flags&REVERSE)
 	    {
-	       if(flags&DELETE)
-	       {
-		  ArgV *args=new ArgV("rm");
-		  while((file=to_rm->curr())!=0)
-		  {
-		     Report(_("Removing old remote file `%s'"),
-			      dir_file(remote_relative_dir,file->name));
-		     args->Append(file->name);
-		     to_rm->next();
-		  }
-		  waiting=new rmJob(Clone(),args);
-		  waiting->parent=this;
-		  waiting->cmdline=args->Combine();
-	       }
-	       else if(flags&REPORT_NOT_DELETED)
-	       {
-		  for(file=to_rm->curr(); file; file=to_rm->next())
-		  {
-		     Report(_("Old remote file `%s' is not removed"),
-			      dir_file(remote_relative_dir,file->name));
-		  }
-	       }
 	       state=REMOTE_REMOVE_OLD;
 	       return MOVED;
 	    }
@@ -657,15 +716,7 @@ int   MirrorJob::Do()
 			   dir_file(local_relative_dir,file->name));
 	       }
 	    }
-	    mode_t mode_mask=0;
-	    if(!(flags&ALLOW_SUID))
-	       mode_mask|=S_ISUID|S_ISGID;
-	    if(!(flags&NO_UMASK))
-	    {
-	       mode_t u=umask(022); // get+set
-	       umask(u);	    // retore
-	       mode_mask|=u;
-	    }
+	    mode_t mode_mask=get_mode_mask();
 	    if(!(flags&NO_PERMS))
 	       to_transfer->LocalChmod(local_dir,mode_mask);
 	    to_transfer->LocalUtime(local_dir,/*only_dirs=*/true);
@@ -684,24 +735,65 @@ int   MirrorJob::Do()
 
    case(REMOTE_REMOVE_OLD):
       if(!waiting)
-	 goto pre_REMOTE_CHMOD;
+      {
+	 if(flags&DELETE)
+	 {
+	    ArgV *args=new ArgV("rm");
+	    file=to_rm->curr();
+	    if(file)
+	    {
+	       args->Append(file->name);
+	       to_rm->next();
+	       if(file->defined&file->TYPE && file->filetype==file->DIRECTORY)
+	       {
+		  Report(_("Removing old remote directory `%s'"),
+			   dir_file(remote_relative_dir,file->name));
+		  args->getnext(); // prepare args position.
+	       	  waiting=new FinderJob_Cmd(Clone(),args,FinderJob_Cmd::RM);
+	       }
+	       else
+	       {
+		  Report(_("Removing old remote file `%s'"),
+			   dir_file(remote_relative_dir,file->name));
+		  waiting=new rmJob(Clone(),args);
+	       }
+	       waiting->SetParentFg(this);
+	       waiting->cmdline=args->Combine();
+	    }
+	 }
+	 else if(flags&REPORT_NOT_DELETED)
+	 {
+	    for(file=to_rm->curr(); file; file=to_rm->next())
+	    {
+	       Report(_("Old remote file `%s' is not removed"),
+			dir_file(remote_relative_dir,file->name));
+	    }
+	 }
+	 if(!waiting)
+	    goto pre_REMOTE_CHMOD;
+      }
       if(waiting->Done())
       {
-	 delete waiting;
+	 Delete(waiting);
 	 waiting=0;
-	 state=DONE;
-	 m=MOVED;
-	 goto pre_REMOTE_CHMOD;
+	 return MOVED;
       }
       return m;
 
    pre_REMOTE_CHMOD:
+      if(flags&NO_PERMS)
+      {
+	 state=DONE;
+	 return MOVED;
+      }
       to_transfer->rewind();
+      state=REMOTE_CHMOD;
+      m=MOVED;
       goto remote_chmod_next;
    case(REMOTE_CHMOD):
       if(waiting->Done())
       {
-	 delete waiting;
+	 Delete(waiting);
 	 waiting=0;
 
       remote_chmod_next:
@@ -716,9 +808,9 @@ int   MirrorJob::Do()
 	    goto remote_chmod_next;
 	 ArgV *a=new ArgV("chmod");
 	 a->Append(fi->name);
-	 ChmodJob *cj=new ChmodJob(Clone(),fi->mode,a);
+	 ChmodJob *cj=new ChmodJob(Clone(),fi->mode&~get_mode_mask(),a);
 	 waiting=cj;
-	 waiting->parent=this;
+	 waiting->SetParentFg(this);
 	 waiting->cmdline=a->Combine();
 	 cj->BeQuiet();   // chmod is not supported on all servers; be quiet.
 
@@ -731,7 +823,7 @@ int   MirrorJob::Do()
 }
 
 MirrorJob::MirrorJob(FileAccess *f,const char *new_local_dir,const char *new_remote_dir)
-   : SessionJob(f), prec("0")
+   : SessionJob(f), prec("0"), loose_prec("0")
 {
    verbose_report=0;
    parent_mirror=0;
@@ -750,7 +842,7 @@ MirrorJob::MirrorJob(FileAccess *f,const char *new_local_dir,const char *new_rem
 
    tot_files=new_files=mod_files=del_files=
    tot_symlinks=new_symlinks=mod_symlinks=del_symlinks=0;
-   dirs=1; del_dirs=0;
+   dirs=0; del_dirs=0;
 
    flags=0;
 
@@ -767,6 +859,9 @@ MirrorJob::MirrorJob(FileAccess *f,const char *new_local_dir,const char *new_rem
    script=0;
    script_only=false;
    script_needs_closing=false;
+
+   use_cache=false;
+   remove_source_files=false;
 }
 
 MirrorJob::~MirrorJob()
@@ -790,18 +885,16 @@ MirrorJob::~MirrorJob()
    if(old_files_set)
       delete old_files_set;
    // don't delete this->file -- it is a reference
-   if(list_info)
-      delete list_info;
-   if(local_session)
-      delete local_session;
+   Delete(list_info);
+   Delete(local_session);
    if(rx_include)
    {
-      free(rx_include);
+      xfree(rx_include);
       regfree(&rxc_include);
    }
    if(rx_exclude)
    {
-      free(rx_exclude);
+      xfree(rx_exclude);
       regfree(&rxc_exclude);
    }
    if(script && script_needs_closing)
@@ -888,6 +981,7 @@ CMD(mirror)
       {"include",required_argument,0,'i'},
       {"exclude",required_argument,0,'x'},
       {"time-prec",required_argument,0,'t'},
+      {"loose-time-prec",required_argument,0,'T'},
       {"only-newer",no_argument,0,'n'},
       {"no-recursion",no_argument,0,'r'},
       {"no-perms",no_argument,0,'p'},
@@ -897,6 +991,8 @@ CMD(mirror)
       {"verbose",optional_argument,0,'v'},
       {"newer-than",required_argument,0,'N'},
       {"dereference",no_argument,0,'L'},
+      {"use-cache",no_argument,0,256+'C'},
+      {"Remove-source-files",no_argument,0,256+'R'},
       {0}
    };
 
@@ -909,6 +1005,7 @@ CMD(mirror)
    static int include_alloc=0;
    static char *exclude=0;
    static int exclude_alloc=0;
+   bool use_cache=false;
 #define APPEND_STRING(s,a,s1) \
    {			                  \
       int len,len1=strlen(s1);            \
@@ -932,13 +1029,15 @@ CMD(mirror)
    if(exclude)
       exclude[0]=0;
 
-   TimeInterval prec("12h");
+   TimeInterval prec      ((const char*)ResMgr::Query("mirror:time-precision",0));
+   TimeInterval loose_prec((const char*)ResMgr::Query("mirror:loose-time-precision",0));
    bool	 create_remote_dir=false;
    int	 verbose=0;
    const char *newer_than=0;
+   bool  remove_source_files=false;
 
    args->rewind();
-   while((opt=args->getopt_long("esi:x:t:nrpcRvN:L",mirror_opts,0))!=EOF)
+   while((opt=args->getopt_long("esi:x:t:T:nrpcRvN:L",mirror_opts,0))!=EOF)
    {
       switch(opt)
       {
@@ -961,14 +1060,21 @@ CMD(mirror)
 	 flags|=MirrorJob::CONTINUE;
 	 break;
       case('t'):
-	 prec=TimeInterval(optarg);
-	 if(prec.Error())
+      case('T'):
+      {
+	 TimeInterval p(optarg);
+	 if(p.Error())
 	 {
-	    parent->eprintf("%s: %s: %s\n",args->a0(),optarg,prec.ErrorText());
+	    parent->eprintf("%s: %s: %s\n",args->a0(),optarg,p.ErrorText());
 	    parent->eprintf(_("Try `help %s' for more information.\n"),args->a0());
 	    return 0;
 	 }
+	 if(opt=='t')
+	    prec=p;
+	 else
+	    loose_prec=p;
 	 break;
+      }
       case('x'):
 	 APPEND_STRING(exclude,exclude_alloc,optarg);
 	 break;
@@ -995,7 +1101,14 @@ CMD(mirror)
       case(256+'u'):
 	 flags|=MirrorJob::NO_UMASK;
 	 break;
+      case(256+'C'):
+	 use_cache=true;
+	 break;
+      case(256+'R'):
+	 remove_source_files=true;
+	 break;
       case('?'):
+	 parent->eprintf(_("Try `help %s' for more information.\n"),args->a0());
 	 return 0;
       }
    }
@@ -1081,18 +1194,34 @@ CMD(mirror)
    if(exclude && exclude[0] && (err=j->SetExclude(exclude)))
       goto err_out;
 
-   j->SetPrec(prec);
+   j->SetPrec(prec,loose_prec);
    if(newer_than)
       j->SetNewerThan(newer_than);
+   j->UseCache(use_cache);
+   if(remove_source_files)
+      j->RemoveSourceFiles();
    return j;
 
 err_out:
    parent->eprintf("%s: %s: %s\n",args->a0(),err_tag,err);
-   delete j;
+   SMTask::Delete(j);
    return 0;
 #undef args
 }
 
+mode_t MirrorJob::get_mode_mask()
+{
+   mode_t mode_mask=0;
+   if(!(flags&ALLOW_SUID))
+      mode_mask|=S_ISUID|S_ISGID;
+   if(!(flags&NO_UMASK))
+   {
+      mode_t u=umask(022); // get+set
+      umask(u);	    // retore
+      mode_mask|=u;
+   }
+   return mode_mask;
+}
 
 #ifdef MODULE
 CDECL void module_init()
