@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: MirrorJob.h,v 1.9 1999/09/20 09:29:00 lav Exp $ */
+/* $Id: MirrorJob.h,v 1.13 2000/04/15 18:57:30 lav Exp $ */
 
 #ifndef MIRRORJOB_H
 #define MIRRORJOB_H
@@ -92,6 +92,7 @@ class MirrorJob : public SessionJob
    regex_t rxc_include,rxc_exclude;
 
    TimeInterval prec;
+   TimeInterval loose_prec;
 
    bool	 dir_made;
    bool	 create_remote_dir;
@@ -109,6 +110,12 @@ class MirrorJob : public SessionJob
    bool script_only;
    bool script_needs_closing;
 
+   bool use_cache;
+
+   bool remove_source_files;
+
+   mode_t get_mode_mask();
+
 public:
    enum
    {
@@ -152,10 +159,14 @@ public:
 
    void	 SetVerbose(int v) { verbose_report=v; }
 
-   void	 SetPrec(const TimeInterval &p) { prec=p; }
+   void	 SetPrec(const TimeInterval &p,const TimeInterval &lp)
+      { prec=p; loose_prec=lp; }
    void	 CreateRemoteDir() { create_remote_dir=true; }
 
    void	 SetNewerThan(const char *file);
+
+   void  UseCache(bool u) { use_cache=u; }
+   void	 RemoveSourceFiles() { remove_source_files=true; }
 };
 
 #endif//MIRRORJOB_H
