@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1999 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1999-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,12 +18,13 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: NetAccess.cc,v 1.6.2.2 2000/01/11 13:15:12 lav Exp $ */
+/* $Id: NetAccess.cc,v 1.28 2000/03/15 16:44:22 lav Exp $ */
 
 #include <config.h>
 
 #include <errno.h>
 #include <assert.h>
+#include <math.h>
 #include "NetAccess.h"
 #include "log.h"
 #include "url.h"
@@ -45,6 +46,8 @@ void NetAccess::Init()
    peer_curr=0;
 
    reconnect_interval=30;  // retry with 30 second interval
+   reconnect_interval_multiplier=1.2;
+   reconnect_interval_max=300;
    timeout=600;		   // 10 minutes with no events = reconnect
 
    proxy=0;
@@ -53,6 +56,10 @@ void NetAccess::Init()
 
    rate_limit=0;
 
+   connection_limit=0;	// no limit.
+   connection_takeover=false;
+   takeover_time=NO_DATE;
+
    Reconfig(0);
 }
 
@@ -66,8 +73,7 @@ NetAccess::NetAccess(const NetAccess *o) : super(o)
 }
 NetAccess::~NetAccess()
 {
-   if(resolver)
-      delete resolver;
+   Delete(resolver);
    if(rate_limit)
       delete rate_limit;
    ClearPeer();
@@ -85,22 +91,33 @@ void NetAccess::Reconfig(const char *name)
    const char *c=hostname;
 
    timeout = ResMgr::Query("net:timeout",c);
-   reconnect_interval = ResMgr::Query("net:reconnect-interval",c);
+   reconnect_interval = ResMgr::Query("net:reconnect-interval-base",c);
+   reconnect_interval_multiplier = ResMgr::Query("net:reconnect-interval-multiplier",c);
+   if(reconnect_interval_multiplier<1)
+      reconnect_interval_multiplier=1;
+   reconnect_interval_max = ResMgr::Query("net:reconnect-interval-max",c);
+   if(reconnect_interval_max<reconnect_interval)
+      reconnect_interval_max=reconnect_interval;
    idle = ResMgr::Query("net:idle",c);
    max_retries = ResMgr::Query("net:max-retries",c);
-   relookup_always = ResMgr::Query("net:relookup-always",c);
    socket_buffer = ResMgr::Query("net:socket-buffer",c);
    socket_maxseg = ResMgr::Query("net:socket-maxseg",c);
+   connection_limit = ResMgr::Query("net:connection-limit",c);
+   connection_takeover = ResMgr::Query("net:connection-takeover",c);
 
    if(rate_limit)
       rate_limit->Reconfig(name,c);
 }
 
+static int one=1;
 void NetAccess::KeepAlive(int sock)
 {
-   static int one=1;
    setsockopt(sock,SOL_SOCKET,SO_KEEPALIVE,(char*)&one,sizeof(one));
 }
+void NetAccess::ReuseAddress(int sock)
+{
+   setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char*)&one,sizeof(one));
+}
 void NetAccess::SetSocketBuffer(int sock,int socket_buffer)
 {
    if(socket_buffer==0)
@@ -173,6 +190,8 @@ socklen_t NetAccess::SocketAddrLen(const sockaddr_u *u)
 
 int NetAccess::SocketConnect(int fd,const sockaddr_u *u)
 {
+   // some systems have wrong connect() prototype, so we have to cast off const.
+   // in any case, connect does not alter the address.
    int res=connect(fd,(sockaddr*)&u->sa,SocketAddrLen(u));
    if(res!=-1)
       UpdateNow(); // if non-blocking doesn't work
@@ -186,7 +205,7 @@ void NetAccess::SayConnectingTo()
    char *str=string_alloca(256+strlen(h));
    sprintf(str,_("Connecting to %s%s (%s) port %u"),proxy?"proxy ":"",
       h,SocketNumericAddress(&peer[peer_curr]),SocketPort(&peer[peer_curr]));
-   DebugPrint("---- ",str,0);
+   DebugPrint("---- ",str,1);
 }
 
 void NetAccess::SetProxy(const char *px)
@@ -199,17 +218,16 @@ void NetAccess::SetProxy(const char *px)
    xfree(proxy_pass); proxy_pass=0;
 
    if(!px)
+      px="";
+
+   ParsedURL url(px);
+   if(!url.host || url.host[0]==0)
    {
-   no_proxy:
       if(was_proxied)
 	 ClearPeer();
       return;
    }
 
-   ParsedURL url(px);
-   if(!url.host || url.host[0]==0)
-      goto no_proxy;
-
    proxy=xstrdup(url.host);
    proxy_port=xstrdup(url.port);
    proxy_user=xstrdup(url.user);
@@ -217,17 +235,38 @@ void NetAccess::SetProxy(const char *px)
    ClearPeer();
 }
 
-int NetAccess::CheckTimeout()
+bool NetAccess::NoProxy()
+{
+   // match hostname against no-proxy var.
+   if(!hostname)
+      return false;
+   const char *no_proxy_c=ResMgr::Query("net:no-proxy",0);
+   if(!no_proxy_c)
+      return false;
+   char *no_proxy=alloca_strdup(no_proxy_c);
+   int h_len=strlen(hostname);
+   for(char *p=strtok(no_proxy," ,"); p; p=strtok(0," ,"))
+   {
+      int p_len=strlen(p);
+      if(p_len>h_len || p_len==0)
+	 continue;
+      if(!strcasecmp(hostname+h_len-p_len,p))
+	 return true;
+   }
+   return false;
+}
+
+bool NetAccess::CheckTimeout()
 {
    if(now-event_time>=timeout)
    {
-      DebugPrint("**** ",_("Timeout - reconnecting"));
+      DebugPrint("**** ",_("Timeout - reconnecting"),0);
       Disconnect();
       event_time=now;
-      return(1);
+      return(true);
    }
-   block+=TimeOut((timeout-(now-event_time))*1000);
-   return(0);
+   TimeoutS(timeout-(now-event_time));
+   return(false);
 }
 
 void NetAccess::ClearPeer()
@@ -252,30 +291,29 @@ void NetAccess::Connect(const char *h,const char *p)
    ClearPeer();
 }
 
+void NetAccess::ConnectVerify()
+{
+   if(peer)
+      return;
+   mode=CONNECT_VERIFY;
+}
+
 int NetAccess::Resolve(const char *defp,const char *ser,const char *pr)
 {
    int m=STALL;
 
-   if(peer)
-   {
-      if(relookup_always)
-	 ClearPeer();
-      else
-	 return m;
-   }
-
    if(!resolver)
    {
+      ClearPeer();
       if(proxy)
 	 resolver=new Resolver(proxy,proxy_port,defp);
       else
 	 resolver=new Resolver(hostname,portname,defp,ser,pr);
+      Roll(resolver);
+      if(!resolver)
+	 return MOVED;
       m=MOVED;
    }
-   if(relookup_always)
-      resolver->NoCache();
-
-   resolver->Do();
 
    if(!resolver->Done())
       return m;
@@ -287,6 +325,8 @@ int NetAccess::Resolve(const char *defp,const char *ser,const char *pr)
       hostname=0;
       xfree(portname);
       portname=0;
+      xfree(cwd);
+      cwd=0;
       return(MOVED);
    }
 
@@ -296,7 +336,7 @@ int NetAccess::Resolve(const char *defp,const char *ser,const char *pr)
    resolver->GetResult(peer);
    peer_curr=0;
 
-   delete resolver;
+   Delete(resolver);
    resolver=0;
    return MOVED;
 }
@@ -395,3 +435,89 @@ void RateLimit::ReconfigTotal()
    total.Reset();
    total_reconfig_needed = false;
 }
+
+long NetAccess::ReconnectInterval()
+{
+   // cyclic exponential growth.
+   float interval = reconnect_interval;
+   if(reconnect_interval_multiplier>1
+   && reconnect_interval_max>=reconnect_interval*reconnect_interval_multiplier
+   && retries>0)
+   {
+      int modval = (int)(log((float)reconnect_interval_max/reconnect_interval)
+                              / log(reconnect_interval_multiplier) + 1.999);
+
+      interval *= pow(reconnect_interval_multiplier, (retries-1)%modval);
+
+      if( interval > reconnect_interval_max )
+         interval = reconnect_interval_max;
+   }
+   return long(interval+.5);
+}
+
+bool NetAccess::ReconnectAllowed()
+{
+   if(max_retries>0 && retries>=max_retries)
+      return true; // it will fault later - no need to wait.
+   if(connection_limit>0 && connection_limit<=CountConnections())
+      return false;
+   if(try_time==0)
+      return true;
+   long interval = ReconnectInterval();
+   if(now-try_time >= interval)
+      return true;
+   TimeoutS(interval-(now-try_time));
+   return false;
+}
+
+const char *NetAccess::DelayingMessage()
+{
+   static char buf[80];
+   if(connection_limit>0 && connection_limit<=CountConnections())
+      return _("Connections limit reached");
+   long remains=ReconnectInterval()-(now-try_time);
+   if(remains<=0)
+      return "";
+   sprintf(buf,"%s: %ld",_("Delaying before reconnect"),remains);
+   current->TimeoutS(1);
+   return buf;
+}
+
+bool NetAccess::NextTry()
+{
+   try_time=now;
+
+   if(max_retries>0 && retries>=max_retries)
+   {
+      Fatal(_("max-retries exceeded"));
+      return false;
+   }
+   retries++;
+
+   assert(peer!=0);
+   assert(peer_curr<peer_num);
+
+   return true;
+}
+
+void NetAccess::Close()
+{
+   retries=0;
+   takeover_time=NO_DATE;
+
+   Delete(resolver);
+   resolver=0;
+
+   super::Close();
+}
+
+int NetAccess::CountConnections()
+{
+   int count=0;
+   for(FileAccess *o=FirstSameSite(); o!=0; o=NextSameSite(o))
+   {
+      if(o->IsConnected())
+	 count++;
+   }
+   return count;
+}
