@@ -18,12 +18,14 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: ResMgr.cc,v 1.14 1999/10/12 22:44:09 lav Exp $ */
+/* $Id: ResMgr.cc,v 1.21 2000/04/03 18:29:38 lav Exp $ */
 
 #include <config.h>
 
 #include <fnmatch.h>
 #include <ctype.h>
+#include <unistd.h>
+#include <math.h>
 #include "ResMgr.h"
 #include "SMTask.h"
 #include "xmalloc.h"
@@ -104,14 +106,13 @@ const char *ResMgr::Set(const char *name,const char *cclosure,const char *cvalue
       return msg;
 
    char *value=xstrdup(cvalue);
-   char *closure=xstrdup(cclosure);
-
    if(value && type->val_valid && (msg=(*type->val_valid)(&value))!=0)
    {
       xfree(value);
       return msg;
    }
 
+   char *closure=xstrdup(cclosure);
    if(closure && type->closure_valid && (msg=(*type->closure_valid)(&closure))!=0)
    {
       xfree(closure);
@@ -133,16 +134,16 @@ const char *ResMgr::Set(const char *name,const char *cclosure,const char *cvalue
    {
       if(value)
       {
-	 free((*scan)->value);
+	 xfree((*scan)->value);
 	 (*scan)->value=value;
-	 free(closure);
+	 xfree(closure);
       }
       else
       {
 	 Resource *to_free=*scan;
 	 *scan=(*scan)->next;
-	 free(to_free);
-	 free(closure);
+	 delete to_free;
+	 xfree(closure);
       }
       SMTask::ReconfigAll(type->name);
    }
@@ -154,7 +155,7 @@ const char *ResMgr::Set(const char *name,const char *cclosure,const char *cvalue
 	 SMTask::ReconfigAll(type->name);
       }
       else
-	 free(closure);
+	 xfree(closure);
    }
    return 0;
 }
@@ -311,7 +312,7 @@ const char *ResMgr::BoolValidate(char **value)
    }
    if(strcmp(v,newval))
    {
-      free(v);
+      xfree(v);
       *value=xstrdup(newval);
    }
 
@@ -337,6 +338,21 @@ const char *ResMgr::NumberValidate(char **value)
    return 0;
 }
 
+const char *ResMgr::FloatValidate(char **value)
+{
+   char *v=*value;
+
+   int n=0;
+
+   double f;
+   if(1>sscanf(v,"%lf%n",&f,&n))
+      return _("invalid floating point number");
+
+   v[n]=0;
+
+   return 0;
+}
+
 const char *ResMgr::UNumberValidate(char **value)
 {
    char *v=*value;
@@ -509,3 +525,62 @@ const char *ResMgr::TimeIntervalValidate(char **s)
       return t.ErrorText();
    return 0;
 }
+
+Range::Range(const char *s)
+{
+   start=0;
+   end=0;
+   full=true;
+   error_text=0;
+
+   if(!strcasecmp(s,"full") || !strcasecmp(s,"any"))
+      return;
+
+   if(sscanf(s,"%d-%d",&start,&end)!=2)
+   {
+      error_text=_("Invalid range format. Format is min-max, e.g. 10-20.");
+      return;
+   }
+   if(start<end)
+   {
+      int tmp=start;
+      start=end;
+      end=tmp;
+   }
+   full=false;
+}
+
+bool Range::Match(int n)
+{
+   return full || (n>=start && n<=end);
+}
+
+#ifndef HAVE_RANDOM_DECL
+CDECL long random();
+CDECL void srandom(unsigned seed);
+#endif
+
+int Range::Random()
+{
+   static bool init=false;
+   if(!init)
+   {
+      srandom(time(NULL)+getpid());
+      init=true;
+   }
+
+   if(full)
+      return random();
+
+   // interval [0;1)
+   double mult=random()/2147483648.0;
+   return start + (int)((end-start+1)*mult);
+}
+
+const char *ResMgr::RangeValidate(char **s)
+{
+   Range r(*s);
+   if(r.Error())
+      return r.ErrorText();
+   return 0;
+}
