@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: ResMgr.h,v 1.12 1999/09/29 13:52:22 lav Exp $ */
+/* $Id: ResMgr.h,v 1.15 2000/04/15 18:57:30 lav Exp $ */
 
 #ifndef RESMGR_H
 #define RESMGR_H
@@ -82,7 +82,9 @@ public:
    static const char *BoolValidate(char **value);
    static const char *NumberValidate(char **value);
    static const char *UNumberValidate(char **value);
+   static const char *FloatValidate(char **value);
    static const char *TimeIntervalValidate(char **value);
+   static const char *RangeValidate(char **value);
    static bool str2bool(const char *value);
 
    static void ClassInit();
@@ -124,6 +126,14 @@ public:
       {
 	 return atoi(s);
       }
+   operator double()
+      {
+	 return atof(s);
+      }
+   operator float()
+      {
+	 return atof(s);
+      }
    operator const char*()
       {
 	 return s;
@@ -145,11 +155,27 @@ public:
 	 error_text=0;
       }
    ~TimeInterval();
+   bool Error() const { return error_text!=0; };
+   const char *ErrorText() const { return error_text; }
+
+   bool IsInfty() const { return infty; }
+   time_t Seconds() const { return interval; }
+};
+
+class Range
+{
+   int start;
+   int end;
+   bool full;
+   const char *error_text;
+
+public:
+   Range(const char *s);
+   bool Match(int n);
+   bool IsFull() { return full; }
+   int Random();
    bool Error() { return error_text!=0; };
    const char *ErrorText() { return error_text; }
-
-   bool IsInfty() { return infty; }
-   time_t Seconds() { return interval; }
 };
 
 #endif //RESMGR_H
