@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Resolver.cc,v 1.22 1999/10/01 16:20:35 lav Exp $ */
+/* $Id: Resolver.cc,v 1.30.2.1 2000/09/07 15:20:59 lav Exp $ */
 
 #include <config.h>
 
@@ -34,8 +34,14 @@
 #include <fcntl.h>
 
 #include <netinet/in.h>
-#include <arpa/nameser.h>
-#include <resolv.h>
+#ifdef HAVE_ARPA_NAMESER_H
+# define class _class // workaround for FreeBSD 3.2.
+# include <arpa/nameser.h>
+# undef class
+#endif
+#ifdef HAVE_RESOLV_H
+# include <resolv.h>
+#endif
 
 #include "xstring.h"
 #include "xmalloc.h"
@@ -50,9 +56,11 @@
 extern "C" { const char *hstrerror(int); }
 #endif
 
+#ifdef HAVE_H_ERRNO
 #ifndef HAVE_H_ERRNO_DECL
 CDECL int h_errno;
 #endif
+#endif
 
 #if defined(HAVE_RES_SEARCH) && !defined(HAVE_RES_SEARCH_DECL)
 CDECL int res_search(const char*,int,int,unsigned char*,int);
@@ -70,7 +78,8 @@ static ResDecl
    res_cache_size  ("dns:cache-size",   "256", ResMgr::UNumberValidate,0),
    res_timeout	   ("dns:fatal-timeout","0",   ResMgr::UNumberValidate,0),
    res_order	   ("dns:order",	DEFAULT_ORDER, Resolver::OrderValidate,0),
-   res_query_srv   ("dns:SRV-query",    "no",  ResMgr::BoolValidate,0);
+   res_query_srv   ("dns:SRV-query",    "no",  ResMgr::BoolValidate,0),
+   res_use_fork	   ("dns:use-fork",     "yes", ResMgr::BoolValidate,0);
 
 
 struct address_family
@@ -110,6 +119,7 @@ Resolver::Resolver(const char *h,const char *p,const char *defp,
    addr=0;
    addr_num=0;
    buf=0;
+   use_fork=res_use_fork.Query(0);
 
    error=0;
 
@@ -136,8 +146,7 @@ Resolver::~Resolver()
       w->Kill(SIGKILL);
       w->Auto();
    }
-   if(buf)
-      delete buf;
+   Delete(buf);
 }
 
 int   Resolver::Do()
@@ -171,7 +180,7 @@ int   Resolver::Do()
       {
 	 if(errno==ENFILE || errno==EMFILE)
 	 {
-	    block+=TimeOut(1000);
+	    TimeoutS(1);
 	    return m;
 	 }
 	 MakeErrMsg("pipe()");
@@ -184,41 +193,51 @@ int   Resolver::Do()
       Log::global->Format(4,"---- %s\n",_("Resolving host address..."));
    }
 
-   if(!w)
+   if(use_fork)
    {
-      pid_t proc=fork();
-      if(proc==-1)
+      if(!w && !buf)
       {
-	 block+=TimeOut(1000);
-	 return m;
-      }
-      if(proc==0)
-      {	 // child
-	 SignalHook::Ignore(SIGINT);
-	 SignalHook::Ignore(SIGTSTP);
-	 SignalHook::Ignore(SIGQUIT);
-	 SignalHook::Ignore(SIGHUP);
-	 close(0);
-// 	 close(2);
-	 dup2(pipe_to_child[1],1);
+	 pid_t proc=fork();
+	 if(proc==-1)
+	 {
+	    TimeoutS(1);
+	    return m;
+	 }
+	 if(proc==0)
+	 {	 // child
+	    SignalHook::Ignore(SIGINT);
+	    SignalHook::Ignore(SIGTSTP);
+	    SignalHook::Ignore(SIGQUIT);
+	    SignalHook::Ignore(SIGHUP);
+	    close(0);	// no input will be needed.
+	    close(pipe_to_child[0]);
+	    pipe_to_child[0]=-1;
+	    DoGethostbyname();
+	    _exit(0);
+	 }
+	 // parent
 	 close(pipe_to_child[1]);
-	 close(pipe_to_child[0]);
-	 DoGethostbyname();
-	 _exit(0);
+	 pipe_to_child[1]=-1;
+
+	 w=new ProcWait(proc);
+	 m=MOVED;
+      }
+      if(!buf)
+      {
+	 buf=new FileInputBuffer(new FDStream(pipe_to_child[0],"<pipe-in>"));
+// 	 Roll(buf);
+	 m=MOVED;
       }
-      // parent
-      close(pipe_to_child[1]);
-      pipe_to_child[1]=-1;
-
-      w=new ProcWait(proc);
-      m=MOVED;
    }
-
-   if(!buf)
+   else
    {
-      buf=new FileInputBuffer(new FDStream(pipe_to_child[0],"<pipe>"));
-      while(buf->Do()==MOVED);
-      m=MOVED;
+      if(!buf)
+      {
+	 buf=new Buffer();
+	 DoGethostbyname();
+	 if(deleting)
+	    return MOVED;
+      }
    }
 
    if(buf->Error())
@@ -238,7 +257,7 @@ int   Resolver::Do()
 	    done=true;
 	    return MOVED;
 	 }
-	 block+=TimeOut((timeout-(now-start_time))*1000);
+	 TimeoutS(timeout-(now-start_time));
       }
       return m;
    }
@@ -355,6 +374,7 @@ const char *Resolver::OrderValidate(char **s)
    return ParseOrder(*s,0);
 }
 
+#ifdef HAVE_RES_SEARCH
 static
 int extract_domain(const unsigned char *answer,const unsigned char *scan,int len,
 		     char *store,int store_len)
@@ -426,6 +446,7 @@ struct SRV
    int order;
 };
 
+static
 int SRV_compare(const void *a,const void *b)
 {
    struct SRV *sa=(struct SRV*)a;
@@ -444,6 +465,7 @@ int SRV_compare(const void *a,const void *b)
       return 1;
    return 0;
 }
+#endif // RES_SEARCH
 
 void Resolver::LookupSRV_RR()
 {
@@ -459,6 +481,12 @@ void Resolver::LookupSRV_RR()
    int len;
    for(;;)
    {
+      if(!use_fork)
+      {
+	 Schedule();
+	 if(deleting)
+	    return;
+      }
       time(&try_time);
       len=res_search(srv_name, C_IN, T_SRV, (u_char*)answer, sizeof(answer));
       if(len>=0)
@@ -626,6 +654,13 @@ void Resolver::LookupOne(const char *name)
 
    for(;;)
    {
+      if(!use_fork)
+      {
+	 Schedule();
+	 if(deleting)
+	    return;
+      }
+
       time(&try_time);
 
       int af=af_order[af_index];
@@ -668,8 +703,8 @@ void Resolver::LookupOne(const char *name)
 	 af_index++;
 	 continue; // try other address families
       }
-      time_t t=time(0);
-      if(t-try_time<5)
+      time_t t;
+      if((t=time(0))-try_time<5)
 	 sleep(5-(t-try_time));
    }
 }
@@ -690,11 +725,11 @@ void Resolver::DoGethostbyname()
 	    port_number=se->s_port;
 	 else
 	 {
-	    write(1,"P",1);
+	    buf->Put("P");
 	    char *msg=string_alloca(64+strlen(tproto));
 	    sprintf(msg,"no such %s service",tproto);
-	    write(1,msg,strlen(msg));
-	    return;
+	    buf->Put(msg);
+	    goto flush;
 	 }
       }
    }
@@ -702,18 +737,35 @@ void Resolver::DoGethostbyname()
    if(service && !portname && !isdigit((unsigned char)hostname[0]))
       LookupSRV_RR();
 
+   if(!use_fork && deleting)
+      return;
+
    LookupOne(hostname);
 
+   if(!use_fork && deleting)
+      return;
+
+   if(!buf)
+      buf=new FileOutputBuffer(new FDStream(pipe_to_child[1],"<pipe-out>"));
+
    if(addr_num==0)
    {
-      write(1,"E",1);
+      buf->Put("E");
       if(error==0)
 	 error="No address found";
-      write(1,error,strlen(error));
-      return;
+      buf->Put(error);
+      goto flush;
+   }
+   buf->Put("O");
+   buf->Put((char*)addr,addr_num*sizeof(*addr));
+
+flush:
+   buf->PutEOF();
+   if(use_fork)
+   {
+      while(buf->Size()>0)
+	 Roll(buf);  // should flush quickly.
    }
-   write(1,"O",1);
-   write(1,addr,addr_num*sizeof(*addr));
    return;
 }
 
