@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -33,23 +33,26 @@ class SMTask
    static SMTask *sched_scan;
    static PollVec sched_total;
 
-
 protected:
    bool	 running;
    bool	 suspended;
+   bool	 deleting;
 
    enum
    {
-      STALL,
-      MOVED,
-      WANTDIE
+      STALL=0,
+      MOVED=1,	  // STALL|MOVED==MOVED.
+      WANTDIE=2	  // for AcceptSig
    };
 
+   virtual ~SMTask();
+
 public:
    PollVec  block;
 
-   void Block(int fd,int mask) { block+=PollVec(fd,mask); }
-   void Timeout(int ms) { block+=TimeOut(ms); }
+   void Block(int fd,int mask) { block.AddFD(fd,mask); }
+   void Timeout(int ms) { block.AddTimeout(ms); }
+   void TimeoutS(int s) { Timeout(1000*s); }
 
    static time_t now;
    static int now_ms;
@@ -65,7 +68,19 @@ public:
    static void ReconfigAll(const char *name);
 
    SMTask();
-   virtual ~SMTask();
+
+   static void Delete(SMTask *);
+   static int Roll(SMTask *);
+
+   static SMTask *current;
+};
+
+class SMTaskInit : public SMTask
+{
+   int Do();
+public:
+   SMTaskInit();
+   ~SMTaskInit();
 };
 
 #endif /* SMTASK_H */
