@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1998 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: bookmark.cc,v 1.3 1999/05/20 10:06:22 lav Exp $ */
+/* $Id: bookmark.cc,v 1.5 2000/01/09 15:06:03 lav Exp $ */
 
 #include <config.h>
 #include "bookmark.h"
@@ -33,7 +33,7 @@ Bookmark::Bookmark()
 {
    const char *home=getenv("HOME");
    if(home==0)
-      home=".";
+      home="";
    int home_len=strlen(home);
    const char *add="/.lftp/bookmarks";
    bm_file=(char*)xmalloc(home_len+strlen(add)+1);
