@@ -52,6 +52,22 @@ void Buffer::GetSaved(const char **buf,int *size) const
    *size=buffer_ptr+in_buffer;
 }
 
+void Buffer::SaveRollback(long p)
+{
+   if(buffer_ptr<p)
+      save=false;
+   if(!save)
+   {
+      buffer_ptr=0;
+      in_buffer=0;
+   }
+   else
+   {
+      buffer_ptr=p;
+      in_buffer=0;
+   }
+}
+
 void Buffer::Allocate(int size)
 {
    if(in_buffer==0 && !save)
@@ -95,6 +111,7 @@ void Buffer::Put(const char *buf,int size)
 	 {
 	    buf+=res;
 	    size-=res;
+	    pos+=res;
 	 }
       }
    }
@@ -106,6 +123,7 @@ void Buffer::Put(const char *buf,int size)
 
    memcpy(buffer+buffer_ptr+in_buffer,buf,size);
    in_buffer+=size;
+   pos+=size;
 }
 
 void Buffer::Skip(int len)
@@ -114,6 +132,23 @@ void Buffer::Skip(int len)
       len=in_buffer;
    in_buffer-=len;
    buffer_ptr+=len;
+   pos+=len;
+}
+void Buffer::UnSkip(int len)
+{
+   if(len>buffer_ptr)
+      len=buffer_ptr;
+   in_buffer+=len;
+   buffer_ptr-=len;
+   pos-=len;
+}
+
+void Buffer::Empty()
+{
+   in_buffer=0;
+   buffer_ptr=0;
+   if(save_max>0)
+      save=true;
 }
 
 int Buffer::Do()
@@ -133,6 +168,7 @@ Buffer::Buffer()
    broken=false;
    save=false;
    save_max=0;
+   pos=0;
 }
 Buffer::~Buffer()
 {
@@ -140,6 +176,12 @@ Buffer::~Buffer()
    xfree(buffer);
 }
 
+void Buffer::SetError(const char *e)
+{
+   xfree(error_text);
+   error_text=xstrdup(e);
+}
+
 // FileOutputBuffer implementation
 #include <fcntl.h>
 #include <unistd.h>
@@ -174,9 +216,9 @@ int FileOutputBuffer::Do()
    }
    int fd=out->getfd();
    if(fd>=0)
-      block+=PollVec(fd,POLLOUT);
+      Block(fd,POLLOUT);
    else
-      block+=TimeOut(1000);
+      TimeoutS(1);
    return STALL;
 }
 int FileOutputBuffer::Put_LL(const char *buf,int size)
@@ -187,23 +229,20 @@ int FileOutputBuffer::Put_LL(const char *buf,int size)
       return -1;
    }
 
+   int res;
    int fd=out->getfd();
    if(fd==-1)
    {
       if(out->error())
-      {
-      err:
-	 error_text=xstrdup(out->error_text);
-	 return -1;
-      }
+	 goto out_err;
       event_time=now;
       return 0;
    }
 
-   int res=write(fd,buf,size);
+   res=write(fd,buf,size);
    if(res==-1)
    {
-      if(errno==EAGAIN)
+      if(errno==EAGAIN || errno==EINTR)
 	 return 0;
       if(errno==EPIPE)
       {
@@ -212,9 +251,13 @@ int FileOutputBuffer::Put_LL(const char *buf,int size)
       }
       saved_errno=errno;
       out->MakeErrorText();
-      goto err;
+      goto out_err;
    }
    return res;
+
+out_err:
+   SetError(out->error_text);
+   return -1;
 }
 
 FgData *FileOutputBuffer::GetFgData(bool fg)
@@ -264,7 +307,7 @@ int FileInputBuffer::Do()
    if(Done() || Error())
       return STALL;
 
-   int res=Get_LL(0x4000);
+   int res=Get_LL(GET_BUFSIZE);
    if(res>0)
    {
       in_buffer+=res;
@@ -305,7 +348,7 @@ int FileInputBuffer::Get_LL(int size)
       {
 	 if(res==FA::DO_AGAIN)
 	    return 0;
-	 error_text=xstrdup(in_FA->StrError(res));
+	 SetError(in_FA->StrError(res));
 	 return -1;
       }
    }
@@ -315,11 +358,7 @@ int FileInputBuffer::Get_LL(int size)
       if(fd==-1)
       {
 	 if(in->error())
-	 {
-	 err:
-	    error_text=xstrdup(in->error_text);
-	    return -1;
-	 }
+	    goto in_err;
 	 return 0;
       }
 
@@ -332,12 +371,16 @@ int FileInputBuffer::Get_LL(int size)
 	    return 0;
 	 saved_errno=errno;
 	 in->MakeErrorText();
-	 goto err;
+	 goto in_err;
       }
    }
    if(res==0)
       eof=true;
    return res;
+
+in_err:
+   SetError(in->error_text);
+   return -1;
 }
 
 FgData *FileInputBuffer::GetFgData(bool fg)
