@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: buffer.h,v 1.7 1999/08/26 18:18:50 lav Exp $ */
+/* $Id: buffer.h,v 1.15 2000/01/27 17:07:54 lav Exp $ */
 
 #ifndef BUFFER_H
 #define BUFFER_H
@@ -28,6 +28,8 @@
 #include "fg.h"
 #include "xstring.h"
 
+#define GET_BUFSIZE 0x4000
+
 class Buffer : public SMTask
 {
 protected:
@@ -44,6 +46,8 @@ protected:
    bool save;  // save skipped data
    int save_max;
 
+   long pos;
+
    // low-level for derived classes
    virtual int Put_LL(const char *buf,int size) { return 0; }
    virtual int PutEOF_LL() { return 0; }
@@ -52,10 +56,13 @@ protected:
 
    void SaveMaxCheck(int addsize);
 
+   virtual ~Buffer();
+
 public:
    virtual int Do();
    virtual bool Done() { return in_buffer==0; }
    bool Error() { return error_text!=0; }
+   void SetError(const char *e);
    int  Errno() { return saved_errno; }
    const char *ErrorText() { return error_text; }
    int Size() { return in_buffer; }
@@ -64,6 +71,7 @@ public:
 
    void Get(const char **buf,int *size);
    void Skip(int len); // Get(); consume; Skip()
+   void UnSkip(int len); // this only works if there were no Put's.
    void Put(const char *buf,int size);
    void Put(const char *buf) { Put(buf,strlen(buf)); }
    void PutEOF() { eof=true; PutEOF_LL(); }
@@ -71,13 +79,17 @@ public:
    // useful for cache.
    void Save(long m) { save=true; save_max=m; }
    void GetSaved(const char **buf,int *size) const;
+   void SaveRollback(long p);
+
+   void SetPos(long p) { pos=p; }
+
+   void Empty();
 
    virtual FgData *GetFgData(bool) { return 0; }
 
    virtual time_t EventTime() { return now; }
 
    Buffer();
-   virtual ~Buffer();
 };
 
 class FileOutputBuffer : public Buffer
@@ -88,9 +100,11 @@ class FileOutputBuffer : public Buffer
 
    time_t event_time; // used to detect timeouts
 
+protected:
+   ~FileOutputBuffer();
+
 public:
    FileOutputBuffer(FDStream *o);
-   ~FileOutputBuffer();
    int Do();
    bool Done();
    FgData *GetFgData(bool fg);
@@ -109,10 +123,12 @@ class FileInputBuffer : public Buffer
 
    time_t event_time; // used to detect timeouts
 
+protected:
+   ~FileInputBuffer();
+
 public:
    FileInputBuffer(FDStream *i);
    FileInputBuffer(FileAccess *i);
-   ~FileInputBuffer();
    int Do();
    bool Done();
    FgData *GetFgData(bool fg);
