@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1999 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: commands.cc,v 1.63 1999/10/20 12:12:34 lav Exp $ */
+/* $Id: commands.cc,v 1.107.2.1 2000/09/07 16:22:12 lav Exp $ */
 
 #include <config.h>
 
@@ -36,22 +36,18 @@
 
 #include "CmdExec.h"
 #include "GetJob.h"
-#include "PutJob.h"
 #include "CatJob.h"
-#include "LsJob.h"
 #include "LsCache.h"
 #include "mgetJob.h"
-#include "mputJob.h"
 #include "mkdirJob.h"
 #include "rmJob.h"
 #include "SysCmdJob.h"
-#include "QuoteJob.h"
 #include "mvJob.h"
 #include "pgetJob.h"
-#include "FtpCopy.h"
 #include "SleepJob.h"
 #include "FindJob.h"
 #include "ChmodJob.h"
+#include "CopyJob.h"
 
 #include "misc.h"
 #include "alias.h"
@@ -65,6 +61,7 @@
 #include "log.h"
 #include "module.h"
 #include "getopt.h"
+#include "FileCopy.h"
 
 #include "confpaths.h"
 
@@ -77,17 +74,17 @@ History	 cwd_history;
 
 CMD(alias); CMD(anon);   CMD(cd);      CMD(debug);
 CMD(exit);  CMD(get);    CMD(help);    CMD(jobs);
-CMD(kill);  CMD(lcd);    CMD(ls);      CMD(mget);
-CMD(open);  CMD(pwd);    CMD(put);     CMD(set);
+CMD(kill);  CMD(lcd);    CMD(ls);
+CMD(open);  CMD(pwd);    CMD(set);
 CMD(shell); CMD(source); CMD(user);    CMD(rm);
-CMD(wait);  CMD(site);   CMD(subsh);   CMD(mirror);
-CMD(mput);  CMD(mv);	 CMD(cat);     CMD(cache);
-CMD(mkdir); CMD(quote);  CMD(scache);  CMD(mrm);
+CMD(wait);  CMD(subsh);  CMD(mirror);
+CMD(mv);    CMD(cat);    CMD(cache);
+CMD(mkdir); CMD(scache); CMD(mrm);
 CMD(ver);   CMD(close);  CMD(bookmark);CMD(lftp);
-CMD(echo);  CMD(suspend);CMD(ftpcopy); CMD(sleep);
+CMD(echo);  CMD(suspend);CMD(sleep);
 CMD(at);    CMD(find);   CMD(command); CMD(module);
 CMD(lpwd);  CMD(glob);	 CMD(chmod);   CMD(queue);
-CMD(repeat);
+CMD(repeat);CMD(get1);
 
 #ifdef MODULE_CMD_MIRROR
 # define cmd_mirror 0
@@ -98,6 +95,8 @@ CMD(repeat);
 # define cmd_repeat 0
 #endif
 
+#define S "\001"
+
 const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 {
    {"!",       cmd_shell,  N_("!<shell_command>"),
@@ -132,9 +131,9 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 "  size <lim>  - set memory limit, -1 means unlimited\n"
 	 "  expire <Nx> - set cache expiration time to N seconds (x=s)\n"
 	 "                minutes (x=m) hours (x=h) or days (x=d)\n")},
-   {"cat",     cmd_cat,    N_("cat [-u] <files>"),
-	 N_("cat - output remote files to stdout\n"
-	 " -u  try to recognize URLs\n")},
+   {"cat",     cmd_cat,    N_("cat [-b] <files>"),
+	 N_("cat - output remote files to stdout (can be redirected)\n"
+	 " -b  use binary mode (ascii is the default)\n")},
    {"cd",      cmd_cd,     N_("cd <rdir>"),
 	 N_("Change current remote directory to <rdir>. The previous remote directory\n"
 	 "is stored as `-'. You can do `cd -' to change the directory back.\n"
@@ -155,15 +154,28 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 "If no jobs active, the code is passed to operating system as lftp\n"
 	 "termination status. If omitted, exit code of last command is used.\n")},
    {"fg",      cmd_wait,   0,"wait"},
-   {"find",    cmd_find},
-   {"ftpcopy", cmd_ftpcopy},
+   {"find",    cmd_find,0,
+	 N_("Usage: find [directory]\n"
+	 "Print contents of specified directory of current directory recursively.\n"
+	 "Directories in the list are marked with trailing slash.\n"
+	 "You can redirect output of this command.\n")},
    {"get",     cmd_get,    N_("get [OPTS] <rfile> [-o <lfile>]"),
 	 N_("Retrieve remote file <rfile> and store it to local file <lfile>.\n"
 	 " -o <lfile> specifies local file name (default - basename of rfile)\n"
 	 " -c  continue, reget\n"
-	 " -e  delete remote files after successful transfer\n"
-	 " -u  try to recognize URLs\n")},
-   {"glob",    cmd_glob,   0,0},
+	 " -E  delete remote files after successful transfer\n"
+	 " -a  use ascii mode (binary is the default)\n"
+	 " -O <base> specifies base directory or URL where files should be placed\n")},
+   {"get1",    cmd_get1,   0,0},
+   {"glob",    cmd_glob,   0,
+	 N_("Usage: glob [OPTS] command args...\n"
+	 "Expand wildcards and run specified command.\n"
+	 "Options can be used to expand wildcards to list of files, directories,\n"
+	 "or both types. Type selection is not very reliable and depends on server.\n"
+	 "If entry type cannot be determined, it will be included in the list.\n"
+	 " -f  plain files (default)\n"
+	 " -d  directories\n"
+	 " -a  all types\n")},
    {"help",    cmd_help,   N_("help [<cmd>]"),
 	 N_("Print help for command <cmd>, or list of available commands\n")},
    {"jobs",    cmd_jobs,   "jobs [-v]",
@@ -191,12 +203,14 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 "or via pipe to external command.\n"
 	 "By default, ls output is cached, to see new listing use `rels' or\n"
 	 "`cache flush'.\n")},
-   {"mget",    cmd_mget,   N_("mget [-c] [-d] [-e] <files>"),
+   {"mget",    cmd_get,	   N_("mget [OPTS] <files>"),
 	 N_("Gets selected files with expanded wildcards\n"
 	 " -c  continue, reget\n"
 	 " -d  create directories the same as in file names and get the\n"
 	 "     files into them instead of current directory\n"
-	 " -e  delete remote files after successful transfer\n")},
+	 " -E  delete remote files after successful transfer\n"
+	 " -a  use ascii mode (binary is the default)\n"
+	 " -O <base> specifies base directory or URL where files should be placed\n")},
    {"mirror",  cmd_mirror, N_("mirror [OPTS] [remote [local]]"),
 	 N_("\nMirror specified remote directory to local directory\n\n"
 	 " -c, --continue         continue a mirror job if possible\n"
@@ -209,13 +223,16 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 " -R, --reverse          reverse mirror (put files)\n"
 	 " -L, --dereference      download symbolic links as files\n"
 	 " -N, --newer-than FILE  download only files newer than the file\n"
-	 " -i RX, --include RX    include matching files (only one allowed)\n"
-	 " -x RX, --exclude RX    exclude matching files (only one allowed)\n"
+	 " -i RX, --include RX    include matching files\n"
+	 " -x RX, --exclude RX    exclude matching files\n"
 	 "                        RX is extended regular expression\n"
 	 " -t Nx, --time-prec Nx  set time precision to N seconds (x=s)\n"
 	 "                        minutes (x=m) hours (x=h) or days (x=d)\n"
-	 "                        default - 12 hours\n"
-	 " -v, --verbose          verbose operation\n"
+	 "                        default - mirror:time-precision setting\n"
+	 " -T Nx, --loose-time-prec  set time precision for imprecise time\n"
+	 "                        default - mirror:loose-time-precision\n"
+	 " -v, --verbose[=N]      verbose operation\n"
+	 "     --use-cache        use cached directory listings\n"
 	 "\n"
 	 "When using -R, the first directory is local and the second is remote.\n"
 	 "If the second directory is omitted, basename of first directory is used.\n"
@@ -229,19 +246,21 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 "   void module_init(int argc,const char *const *argv)\n"
 	 "If name contains a slash, then the module is searched in current\n"
 	 "directory, otherwise in PKGLIBDIR.\n")},
-   {"more",    cmd_cat,    N_("more [-u] <files>"),
-	 N_("Same as `cat <files> | more'. if PAGER is set, it is used as filter\n"
-	 " -u  try to recognize URLs\n")},
-   {"mput",    cmd_mput,   N_("mput [-c] [-d] <files>"),
+   {"more",    cmd_cat,    N_("more <files>"),
+	 N_("Same as `cat <files> | more'. if PAGER is set, it is used as filter\n")},
+   {"mput",    cmd_get,	   N_("mput [OPTS] <files>"),
 	 N_("Upload files with wildcard expansion\n"
 	 " -c  continue, reput\n"
 	 " -d  create directories the same as in file names and put the\n"
-	 "     files into them instead of current directory\n")},
+	 "     files into them instead of current directory\n"
+	 " -E  delete local files after successful transfer (dangerous)\n"
+	 " -a  use ascii mode (binary is the default)\n"
+	 " -O <base> specifies base directory or URL where files should be placed\n")},
    {"mrm",     cmd_mrm,    N_("mrm <files>"),
 	 N_("Removes specified files with wildcard expansion\n")},
    {"mv",      cmd_mv,	    N_("mv <file1> <file2>"),
 	 N_("Rename <file1> to <file2>\n")},
-   {"nlist",   cmd_ls,	    N_("nlist [<args>]"),
+   {"nlist",   cmd_ls,     N_("nlist [<args>]"),
 	 N_("List remote file names\n")},
    {"open",    cmd_open,   N_("open [OPTS] <site>"),
 	 N_("Select a server, URL or bookmark\n"
@@ -254,18 +273,25 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 "but loads the net heavily impacting other users. Use only if you really\n"
 	 "have to transfer the file ASAP, or some other user may go mad :)\n"
 	 "\nOptions:\n"
-	 " -n <maxconn>  set maximum number of connections (default 5)\n"
-	 " -u            try to recognize URLs\n")},
-   {"put",     cmd_put,    N_("put [-c] <lfile> [-o <rfile>]"),
+	 " -n <maxconn>  set maximum number of connections (default 5)\n")},
+   {"put",     cmd_get,    N_("put [OPTS] <lfile> [-o <rfile>]"),
 	 N_("Upload <lfile> with remote name <rfile>.\n"
 	 " -o <rfile> specifies remote file name (default - basename of lfile)\n"
 	 " -c  continue, reput\n"
-	 "     it requires permission to overwrite remote files\n")},
-   {"pwd",     cmd_pwd,    "pwd",
-	 N_("Print current remote directory\n")},
-   {"queue",   cmd_queue,  0,0},
+	 "     it requires permission to overwrite remote files\n"
+	 " -E  delete local files after successful transfer (dangerous)\n"
+	 " -a  use ascii mode (binary is the default)\n"
+	 " -O <base> specifies base directory or URL where files should be placed\n")},
+   {"pwd",     cmd_pwd,    "pwd [-u]",
+	 N_("Print current remote URL.\n"
+	 " -p  show password\n")},
+   {"queue",   cmd_queue,  0,
+	 N_("Usage: queue <command>\n"
+	 "Add the command to queue for current site. Each site has its own\n"
+	 "command queue. It is possible to queue up a running job by using\n")
+	 "command `queue wait <jobno>'.\n"},
    {"quit",    cmd_exit,   0,"exit"},
-   {"quote",   cmd_quote,  N_("quote <cmd>"),
+   {"quote",   cmd_ls,	   N_("quote <cmd>"),
 	 N_("Send the command uninterpreted. Use with caution - it can lead to\n"
 	 "unknown remote state and thus will cause reconnect. You cannot\n"
 	 "be sure that any change of remote state because of quoted command\n"
@@ -276,13 +302,17 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
 	 N_("Same as `ls', but don't look in cache\n")},
    {"renlist", cmd_ls,	    N_("renlist [<args>]"),
 	 N_("Same as `nlist', but don't look in cache\n")},
-   {"repeat",  cmd_repeat},
-   {"reput",   cmd_put,    N_("reput <lfile> [-o <rfile>]"),
+   {"repeat",  cmd_repeat, 0,
+	 N_("Usage: repeat [delay] [command]\n"
+	 "Repeat specified command with a delay between iterations.\n"
+	 "Default delay is one second, default command is empty.\n")},
+   {"reput",   cmd_get,    N_("reput <lfile> [-o <rfile>]"),
 	 N_("Same as `put -c'\n")},
-   {"rm",      cmd_rm,	    N_("rm [-r] <files>"),
+   {"rm",      cmd_rm,	    N_("rm [-r] [-f] <files>"),
 	 N_("Remove remote files\n"
-	    " -r  recursive directory remove, be careful\n")},
-   {"rmdir",   cmd_rm,	    N_("rmdir <dirs>"),
+	    " -r  recursive directory removal, be careful\n"
+	    " -f  work quietly\n")},
+   {"rmdir",   cmd_rm,	    N_("rmdir [-f] <dirs>"),
 	 N_("Remove remote directories\n")},
    {"scache",  cmd_scache, N_("scache [<session_no>]"),
 	 N_("List cached sessions or switch to specified session number\n")},
@@ -295,7 +325,7 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
    	 " -a  list all settings, including default values\n"
 	 " -d  list only default values, not necessary current ones\n")},
    {"shell",   cmd_shell,  0,"!"},
-   {"site",    cmd_site,   N_("site <site_cmd>"),
+   {"site",    cmd_ls,	   N_("site <site_cmd>"),
 	 N_("Execute site command <site_cmd> and output the result\n"
 	 "You can redirect its output\n")},
    {"sleep",   cmd_sleep, 0,
@@ -306,15 +336,17 @@ const struct CmdExec::cmd_rec CmdExec::static_cmd_table[]=
    {"source",  cmd_source, N_("source <file>"),
 	 N_("Execute commands recorded in file <file>\n")},
    {"suspend", cmd_suspend},
-   {"user",    cmd_user,   N_("user <user> [<pass>]"),
-	 N_("Use specified info for remote login\n")},
+   {"user",    cmd_user,   N_("user <user|URL> [<pass>]"),
+	 N_("Use specified info for remote login. If you specify URL, the password\n"
+	 "will be cached for future usage.")},
    {"version", cmd_ver,    "version",
 	 N_("Shows lftp version\n")},
-   {"wait",    cmd_wait,   N_("wait <jobno>"),
-	 N_("Wait for specified job to terminate.\n")},
-   {"zcat",    cmd_cat,    N_("zcat [-u] <files>"),
+   {"wait",    cmd_wait,   N_("wait [<jobno>]"),
+	 N_("Wait for specified job to terminate. If jobno is omitted, wait\n"
+	 "for last backgrounded job.\n")},
+   {"zcat",    cmd_cat,    N_("zcat <files>"),
 	 N_("Same as cat, but filter each file through zcat\n")},
-   {"zmore",   cmd_cat,    N_("zmore [-u] <files>"),
+   {"zmore",   cmd_cat,    N_("zmore <files>"),
 	 N_("Same as more, but filter each file through zcat\n")},
 
    {NULL,NULL}
@@ -414,8 +446,7 @@ Job *CmdExec::builtin_cd()
 
    const char *dir=args->getarg(1);
 
-   char c;
-   if(sscanf(dir,"%*[a-z]://%*[^/]%c",&c)==1)
+   if(url::is_url(dir))
       return builtin_open();
 
    if(!strcmp(dir,"-"))
@@ -439,7 +470,7 @@ Job *CmdExec::builtin_cd()
       return 0;
    }
    session->Chdir(dir);
-   while(session->Do()==MOVED);
+   Roll(session);
    builtin=BUILTIN_CD;
    return this;
 }
@@ -522,6 +553,8 @@ Job *CmdExec::builtin_lftp()
 
 Job *CmdExec::builtin_open()
 {
+   ReuseSavedSession();
+
    bool	 debug=false;
    char	 *port=NULL;
    char	 *host=NULL;
@@ -532,9 +565,11 @@ Job *CmdExec::builtin_open()
    NetRC::Entry *nrc=0;
    char  *cmd_to_exec=0;
    char  *op=args->a0();
+   bool insecure=false;
+   bool no_bm=false;
 
    args->rewind();
-   while((c=args->getopt("u:p:e:d"))!=EOF)
+   while((c=args->getopt("u:p:e:dB"))!=EOF)
    {
       switch(c)
       {
@@ -550,6 +585,7 @@ Job *CmdExec::builtin_open()
    	    break;
 	 *pass=0;
 	 pass++;
+	 insecure=true;
          break;
       case('d'):
 	 debug=true;
@@ -557,6 +593,9 @@ Job *CmdExec::builtin_open()
       case('e'):
 	 cmd_to_exec=optarg;
 	 break;
+      case('B'):
+	 no_bm=true;
+	 break;
       case('?'):
 	 if(!strcmp(op,"lftp"))
 	    eprintf(_("Try `%s --help' for more information\n"),op);
@@ -577,10 +616,10 @@ Job *CmdExec::builtin_open()
    if(cmd_to_exec)
       PrependCmd(cmd_to_exec);
 
-   if(host && (bm=lftp_bookmarks.Lookup(host))!=0)
+   if(!no_bm && host && (bm=lftp_bookmarks.Lookup(host))!=0)
    {
-      char *cmd=string_alloca(5+strlen(bm)+2+1);
-      sprintf(cmd,"open %s%s\n",bm,background?"&":"");
+      char *cmd=string_alloca(5+3+strlen(bm)+2+1);
+      sprintf(cmd,"open -B %s%s\n",bm,background?"&":"");
       PrependCmd(cmd);
    }
    else
@@ -590,23 +629,26 @@ Job *CmdExec::builtin_open()
 	 url=new ParsedURL(host);
 
 	 const ParsedURL &uc=*url;
-	 if(uc.host)
+	 if(uc.host && uc.host[0])
 	 {
 	    cwd_history.Set(session,session->GetCwd());
 
 	    FileAccess *new_session=0;
 	    // get session with specified protocol if protocol differs
 	    if(uc.proto && strcmp(uc.proto,session->GetProto())
-	    || session->GetProto()[0]==0) // or if current session is dummy
+	    || session->GetProto()[0]==0 // or if current session is dummy
+	    || !strcmp(session->GetProto(),"file"))
 	    {
 	       const char *p=uc.proto;
+	       if(!p)
+		  p=ResMgr::Query("cmd:default-protocol",0);
 	       if(!p)
 		  p="ftp";
 	       new_session=FileAccess::New(p);
 	       if(!new_session)
 	       {
-		  eprintf(N_("%s: %s - not supported protocol\n"),
-			   args->getarg(0),p);
+		  eprintf("%s: %s%s\n",args->a0(),p,
+			   _(" - not supported protocol"));
 		  return 0;
 	       }
 	    }
@@ -615,13 +657,17 @@ Job *CmdExec::builtin_open()
 	       new_session=session->Clone();
 	       new_session->AnonymousLogin();
 	    }
-	    Reuse(session);
-	    session=new_session;
+	    saved_session=session;
+	    session=0;
+	    ChangeSession(new_session);
 
 	    if(uc.user && !user)
 	       user=uc.user;
 	    if(uc.pass && !pass)
+	    {
 	       pass=uc.pass;
+	       insecure=true;
+	    }
 	    host=uc.host;
 	    if(uc.port!=0 && port==0)
 	       port=uc.port;
@@ -629,12 +675,12 @@ Job *CmdExec::builtin_open()
 	       path=uc.path;
 	 }
 
-	 if(!strcmp(session->GetProto(),"ftp"))
+	 if(!pass)
 	 {
 	    nrc=NetRC::LookupHost(host);
 	    if(nrc)
 	    {
-	       if(nrc->user && !user
+	       if((!uc.proto && nrc->user && !user)
 	       || (nrc->user && user && !strcmp(nrc->user,user) && !pass))
 	       {
 		  user=nrc->user;
@@ -654,7 +700,12 @@ Job *CmdExec::builtin_open()
 	    eprintf(_("%s: GetPass() failed -- assume anonymous login\n"),
 	       args->getarg(0));
 	 else
-	    session->Login(user,pass);
+	 {
+	    session->Login(user,0);
+	    // assume the new password is the correct one.
+	    session->SetPasswordGlobal(pass);
+	    session->InsecurePassword(insecure && !no_bm);
+	 }
       }
       if(host)
       {
@@ -670,6 +721,10 @@ Job *CmdExec::builtin_open()
 
    if(path)
    {
+      const char *old=cwd_history.Lookup(session);
+      if(old)
+	 session->Chdir(old,false);
+
       char *s=string_alloca(strlen(path)*4+40);
       strcpy(s,"&& cd \"");
       unquote(s+strlen(s),path);
@@ -677,18 +732,6 @@ Job *CmdExec::builtin_open()
       if(background)
 	 strcat(s,"&");
       strcat(s,"\n");
-#if 0
-      char *slash=strrchr(path,'/');
-      if(slash && slash[1])
-      {
-	 *slash=0;
-	 strcat(s,"|| (cd \"");
-	 unquote(s+strlen(s),path);
-	 strcat(s,"\" && get -- \"");
-	 unquote(s+strlen(s),slash+1);
-	 strcat(s,"\")\n");
-      }
-#endif
       PrependCmd(s);
    }
 
@@ -703,6 +746,8 @@ Job *CmdExec::builtin_open()
    if(builtin==BUILTIN_OPEN)
       return this;
 
+   ReuseSavedSession();
+
    exit_code=0;
    return 0;
 }
@@ -715,9 +760,36 @@ Job *CmdExec::builtin_restart()
 
 Job *CmdExec::builtin_glob()
 {
+   const char *op=args->a0();
+   bool files_only=true;
+   bool dirs_only=false;
+   int opt;
+
+   while((opt=args->getopt("+adf"))!=EOF)
+   {
+      switch(opt)
+      {
+      case('a'):
+	 files_only=dirs_only=false;
+	 break;
+      case('d'):
+	 dirs_only=true;
+	 files_only=false;
+	 break;
+      case('f'):
+	 files_only=true;
+	 dirs_only=false;
+	 break;
+      case('?'):
+	 eprintf(_("Try `help %s' for more information.\n"),op);
+	 return 0;
+      }
+   }
+   while(args->getindex()>1)
+      args->delarg(1);	   // remove options.
    if(args->count()<2)
    {
-      eprintf(_("Usage: %s command args...\n"),args->a0());
+      eprintf(_("Usage: %s [OPTS] command args...\n"),op);
       return 0;
    }
    assert(args_glob==0 && glob==0);
@@ -732,9 +804,11 @@ Job *CmdExec::builtin_glob()
       args->rewind();
       return cmd_command(this);
    }
-   glob=session->MakeGlob(pat);
-   if(!glob)
-      glob=new NoGlob(pat);
+   glob=new GlobURL(session,pat);
+   if(dirs_only)
+      glob->glob->DirectoriesOnly();
+   if(files_only)
+      glob->glob->FilesOnly();
    builtin=BUILTIN_GLOB;
    return this;
 }
@@ -751,7 +825,7 @@ Job *CmdExec::builtin_queue()
    if(queue==0)
    {
       queue=new CmdExec(session->Clone());
-      queue->parent=this;
+      queue->SetParentFg(this,false);
       queue->AllocJobno();
       const char *url=session->GetConnectURL(FA::NO_PATH);
       queue->cmdline=(char*)xmalloc(9+strlen(url));
@@ -806,77 +880,138 @@ CMD(lcd)
 
 CMD(ls)
 {
-   int mode=FA::LONG_LIST;
-   if(strstr(args->a0(),"nlist"))
-      mode=FA::LIST;
-   if(mode==FA::LONG_LIST && args->count()==1)
+   bool nlist=false;
+   bool re=false;
+   int mode=FA::LIST;
+   const char *op=args->a0();
+   if(strstr(op,"nlist"))
+      nlist=true;
+   if(!strncmp(op,"re",2))
+      re=true;
+   if(!strcmp(op,"quote") || !strcmp(op,"site"))
+   {
+      if(args->count()<=1)
+      {
+	 eprintf(_("Usage: %s <cmd>\n"),op);
+	 return 0;
+      }
+      nlist=true;
+      mode=FA::QUOTE_CMD;
+      if(!strcmp(op,"site"))
+	 args->insarg(1,"SITE");
+   }
+
+   char *a=args->Combine(1);
+
+   if(!nlist && args->count()==1 && parent->var_ls && parent->var_ls[0])
       args->Append(parent->var_ls);
-   LsJob *j=new LsJob(Clone(),output,args,mode);
-   if(!strncmp(args->a0(),"re",2))
-      j->NoCache();
+
+   FileCopyPeer *src_peer=0;
+   if(!nlist)
+      src_peer=new FileCopyPeerDirList(Clone(),args);
+   else
+      src_peer=new FileCopyPeerFA(Clone(),a,mode);
+
+   if(re)
+      src_peer->NoCache();
+   src_peer->SetDate(NO_DATE);
+   src_peer->SetSize(NO_SIZE);
+   FileCopyPeer *dst_peer=new FileCopyPeerFDStream(output,FileCopyPeer::PUT);
+
+   FileCopy *c=FileCopy::New(src_peer,dst_peer,false);
+   c->DontCopyDate();
+   c->LineBuffered();
+   c->Ascii();
+
+   CopyJob *j=new CopyJob(c,a,op);
+   if(!output || output->usesfd(1))
+      j->NoStatus();
+   xfree(a);
    output=0;
-   args=0;
+   if(!nlist)
+      args=0;  // `ls' consumes args itself.
+
    return j;
 }
 
 CMD(cat)
 {
-   char *op=args->a0();
-   ArgV *n_args=new ArgV(op);
-   bool use_urls=false;
+   const char *op=args->a0();
    int opt;
+   bool ascii=true;
 
-   while((opt=args->getopt("+u"))!=EOF)
+   while((opt=args->getopt("+bau"))!=EOF)
    {
       switch(opt)
       {
-      case('u'):
-	 use_urls=true;
+      case('a'):
+	 ascii=true;
+	 break;
+      case('b'):
+	 ascii=false;
 	 break;
       case('?'):
-      err:
 	 eprintf(_("Try `help %s' for more information.\n"),op);
 	 return 0;
       }
    }
-   args->back();
-   char *a;
-   while((a=args->getnext())!=0)
-      n_args->Append(a);
-   if(n_args->count()<=1)
+   while(args->getindex()>1)
+      args->delarg(1);
+   args->rewind();
+   if(args->count()<=1)
    {
-      eprintf(_("Usage: %s files...\n"),args->a0());
-      goto err;
+      eprintf(_("Usage: %s [OPTS] files...\n"),op);
+      return 0;
    }
-   CatJob *j=new CatJob(Clone(),output,n_args);
+   CatJob *j=new CatJob(Clone(),output,args);
+   if(ascii && args->a0()[0]!='z')
+      j->Ascii();
    output=0;
-   if(use_urls)
-      j->UseURLs();
+   args=0;
    return j;
 }
 
 CMD(get)
 {
    int opt;
-   bool use_urls=false;
    bool cont=false;
-   const char *opts="+ceu";
+   const char *opts="+cEuaO:";
    const char *op=args->a0();
    ArgV	 *get_args=new ArgV(op);
    int n_conn=0;
    bool del=false;
+   bool ascii=false;
+   bool glob=false;
+   bool make_dirs=false;
+   bool reverse=false;
+   char *output_dir=0;
 
    args->rewind();
    if(!strncmp(op,"re",2))
    {
       cont=true;
-      opts="+eu";
+      opts="+EuaO:";
    }
    if(!strcmp(op,"pget"))
    {
-      opts="+n:u";
+      opts="+n:uO:";
       n_conn=-1;
    }
+   else if(!strcmp(op,"put") || !strcmp(op,"reput"))
+   {
+      reverse=true;
+   }
+   else if(!strcmp(op,"mget"))
+   {
+      glob=true;
+      opts="cEadO:";
+   }
+   else if(!strcmp(op,"mput"))
+   {
+      glob=true;
+      opts="cEadO:";
+      reverse=true;
+   }
    while((opt=args->getopt(opts))!=EOF)
    {
       switch(opt)
@@ -892,67 +1027,70 @@ CMD(get)
 	 }
 	 n_conn=atoi(optarg);
 	 break;
-      case('e'):
+      case('E'):
 	 del=true;
 	 break;
-      case('u'):
-	 use_urls=true;
+      case('a'):
+	 ascii=true;
+	 break;
+      case('d'):
+	 make_dirs=true;
+	 break;
+      case('O'):
+	 output_dir=optarg;
 	 break;
       case('?'):
       err:
 	 eprintf(_("Try `help %s' for more information.\n"),op);
+	 delete get_args;
 	 return 0;
       }
    }
+   if(glob)
+   {
+      if(args->getcurr()==0)
+      {
+      file_name_missed:
+	 // xgettext:c-format
+	 eprintf(_("File name missed. "));
+	 goto err;
+      }
+      delete get_args;
+      output_dir=xstrdup(output_dir);  // save it from deleting.
+      // remove options
+      while(args->getindex()>1)
+	 args->delarg(1);
+      mgetJob *j=new mgetJob(Clone(),args,cont,make_dirs);
+      if(reverse)
+	 j->Reverse();
+      if(del)
+	 j->DeleteFiles();
+      if(ascii)
+	 j->Ascii();
+      if(output_dir)
+	 j->OutputDir(output_dir);
+      args=0;
+      return j;
+   }
    args->back();
    const char *a=args->getnext();
-   const char *a1;
    if(a==0)
-   {
-      // xgettext:c-format
-      eprintf(_("File name missed. "));
-      goto err;
-   }
-   if(!use_urls)
-      use_urls=ResMgr::Query("xfer:use-urls",0);
+      goto file_name_missed;
    while(a)
    {
-      get_args->Append(a);
-      ParsedURL url(a);
-      a1=0;
-      if(use_urls)
-      {
-	 if(url.proto && url.path)
-	    a1=basename_ptr(url.path);
-      }
-      if(a1==0)
-	 a1=basename_ptr(a);
+      const char *src=a;
+      const char *dst=0;
       a=args->getnext();
       if(a && !strcmp(a,"-o"))
       {
-	 a=args->getnext();
-	 if(a)
-	 {
-	    a=expand_home_relative(a);
-	    struct stat st;
-	    int res=stat(a,&st);
-	    if(res!=-1 && S_ISDIR(st.st_mode))
-	    {
-	       char *comb=string_alloca(strlen(a)+strlen(a1)+2);
-	       sprintf(comb,"%s/%s",a,a1);
-	       get_args->Append(comb);
-	    }
-	    else
-	    {
-	       get_args->Append(a);
-	    }
-	 }
-	 else
-	    get_args->Append(a1);
+	 dst=args->getnext();
 	 a=args->getnext();
       }
-      else
-	 get_args->Append(a1);
+      if(reverse)
+	 src=expand_home_relative(src);
+      dst=output_file_name(src,dst,!reverse,output_dir,false);
+      get_args->Append(src);
+      get_args->Append(dst);
    }
 
    if(n_conn==0)
@@ -960,8 +1098,10 @@ CMD(get)
       GetJob *j=new GetJob(Clone(),get_args,cont);
       if(del)
 	 j->DeleteFiles();
-      if(use_urls)
-	 j->UseURLs();
+      if(ascii)
+	 j->Ascii();
+      if(reverse)
+	 j->Reverse();
       return j;
    }
    else
@@ -969,85 +1109,10 @@ CMD(get)
       pgetJob *j=new pgetJob(Clone(),get_args);
       if(n_conn!=-1)
 	 j->SetMaxConn(n_conn);
-      if(use_urls)
-	 j->UseURLs();
       return j;
    }
 }
 
-CMD(put)
-{
-   int opt;
-   bool cont=false;
-   const char *opts="+c";
-   const char *op=args->a0();
-   ArgV	 *get_args=new ArgV(op);
-
-   args->rewind();
-   if(!strncmp(op,"re",2))
-   {
-      cont=true;
-      opts="+";
-   }
-   while((opt=args->getopt(opts))!=EOF)
-   {
-      switch(opt)
-      {
-      case('c'):
-	 cont=true;
-	 break;
-      case('?'):
-      err:
-	 eprintf(_("Try `help %s' for more information.\n"),op);
-	 return 0;
-      }
-   }
-   args->back();
-   const char *a=args->getnext();
-   const char *a1;
-   if(a==0)
-   {
-      // xgettext:c-format
-      eprintf(_("File name missed. "));
-      goto err;
-   }
-   while(a)
-   {
-      a=expand_home_relative(a);
-      get_args->Append(a);
-      a1=basename_ptr(a);
-      a=args->getnext();
-      if(a && !strcmp(a,"-o"))
-      {
-	 a=args->getnext();
-	 if(a)
-	    get_args->Append(a);
-	 else
-	    get_args->Append(a1);
-	 a=args->getnext();
-      }
-      else
-	 get_args->Append(a1);
-   }
-
-   PutJob *j=new PutJob(Clone(),get_args,cont);
-   return j;
-}
-
-CMD(mget)
-{
-   Job *j=new mgetJob(Clone(),args);
-   args=0;
-   return j;
-}
-
-CMD(mput)
-{
-   Job *j=new mputJob(Clone(),args);
-   args=0;
-   return j;
-}
-
 CMD(shell)
 {
    Job *j;
@@ -1057,7 +1122,7 @@ CMD(shell)
    {
       char *a=args->Combine(1);
       j=new SysCmdJob(a);
-      free(a);
+      xfree(a);
    }
    return j;
 }
@@ -1072,10 +1137,11 @@ CMD(rm)
 {
    int opt;
    bool recursive=false;
-   const char *opts="+r";
+   bool silent=false;
+   const char *opts="+rf";
 
    if(!strcmp(args->a0(),"rmdir"))
-      opts="+";
+      opts="+f";
 
    while((opt=args->getopt(opts))!=EOF)
    {
@@ -1084,9 +1150,12 @@ CMD(rm)
       case('r'):
 	 recursive=true;
 	 break;
+      case('f'):
+	 silent=true;
+	 break;
       case('?'):
       print_usage:
-	 eprintf(_("Usage: %s files...\n"),args->a0());
+	 eprintf(_("Usage: %s [-r] [-f] files...\n"),args->a0());
 	 return 0;
       }
    }
@@ -1097,8 +1166,10 @@ CMD(rm)
 
    if(recursive)
    {
-      Job *j=new FinderJob_Cmd(Clone(),args,FinderJob_Cmd::RM);
+      FinderJob *j=new FinderJob_Cmd(Clone(),args,FinderJob_Cmd::RM);
       args=0;
+      if(silent)
+	 j->BeQuiet();
       return j;
    }
 
@@ -1112,6 +1183,9 @@ CMD(rm)
       curr=args->getnext();
    }
 
+   if(silent)
+      j->BeQuiet();
+
    return j;
 }
 CMD(mkdir)
@@ -1164,10 +1238,27 @@ CMD(cd)
 
 CMD(pwd)
 {
-   char *url=xstrdup(session->GetConnectURL());
-   int len=strlen(url);
+   int opt;
+   args->rewind();
+   int flags=0;
+   while((opt=args->getopt("p"))!=EOF)
+   {
+      switch(opt)
+      {
+      case('p'):
+	 flags|=FA::WITH_PASSWORD;
+	 break;
+      case('?'):
+         // xgettext:c-format
+	 eprintf(_("Usage: %s [-p]\n"),args->a0());
+	 return 0;
+      }
+   }
+   const char *url_c=session->GetConnectURL(flags);
+   char *url=alloca_strdup(url_c);
+   int len=strlen(url_c);
    url[len++]='\n';  // replaces \0
-   Job *j=new CatJob(output,url,len);
+   Job *j=CopyJob::NewEcho(url,len,output,args->a0());
    output=0;
    return j;
 }
@@ -1255,18 +1346,53 @@ CMD(debug)
 
 CMD(user)
 {
-   char	 *pass;
    if(args->count()<2 || args->count()>3)
    {
       eprintf(_("Usage: %s userid [pass]\n"),args->getarg(0));
       return 0;
    }
-   if(args->count()==2)
+   char *user=args->getarg(1);
+   char	*pass=args->getarg(2);
+   bool insecure=(pass!=0);
+
+   ParsedURL u(user,true);
+   if(u.proto && !u.user)
+   {
+      exit_code=0;
+      return 0;
+   }
+   if(u.proto && u.user && u.pass)
+   {
+      pass=u.pass;
+      insecure=true;
+   }
+   if(!pass)
       pass=GetPass(_("Password: "));
+   if(!pass)
+      return 0;
+
+   if(u.proto && u.user)
+   {
+      FA *s=FA::New(&u,false);
+      if(s)
+      {
+	 s->SetPasswordGlobal(pass);
+	 s->InsecurePassword(insecure);
+	 SessionPool::Reuse(s);
+      }
+      else
+      {
+	 eprintf("%s: %s%s\n",args->a0(),u.proto,
+		  _(" - not supported protocol"));
+	 return 0;
+      }
+   }
    else
-      pass=args->getarg(2);
-   if(pass)
-      session->Login(args->getarg(1),pass);
+   {
+      session->Login(args->getarg(1),0);
+      session->SetPasswordGlobal(pass);
+      session->InsecurePassword(insecure);
+   }
    exit_code=0;
    return 0;
 }
@@ -1373,7 +1499,8 @@ CMD(set)
    if(a==0)
    {
       char *s=ResMgr::Format(with_defaults,only_defaults);
-      Job *j=new CatJob(output,s,strlen(s));
+      Job *j=CopyJob::NewEcho(s,output,args->a0());
+      xfree(s);
       output=0;
       return j;
    }
@@ -1416,7 +1543,8 @@ CMD(alias)
    if(args->count()<2)
    {
       char *list=Alias::Format();
-      Job *j=new CatJob(output,list,strlen(list));
+      Job *j=CopyJob::NewEcho(list,output,args->a0());
+      xfree(list);
       output=0;
       return j;
    }
@@ -1428,7 +1556,7 @@ CMD(alias)
    {
       char *val=args->Combine(2);
       Alias::Add(args->getarg(1),val);
-      free(val);
+      xfree(val);
    }
    exit_code=0;
    return 0;
@@ -1476,37 +1604,7 @@ CMD(wait)
       return 0;
    }
    j->parent=0;
-   return j;
-}
-
-CMD(quote)
-{
-   if(args->count()<=1)
-   {
-      eprintf(_("Usage: %s <raw_cmd>\n"),args->a0());
-      return 0;
-   }
-   Job *j=new QuoteJob(Clone(),args->a0(),args->Combine(1),
-      output?output:new FDStream(1,"<stdout>"));
-   output=0;
-   return j;
-}
-
-CMD(site)
-{
-   if(args->count()<=1)
-   {
-      // xgettext:c-format
-      eprintf(_("Usage: site <site_cmd>\n"));
-      return 0;
-   }
-   char *cmd=args->Combine(1);
-   cmd=(char*)xrealloc(cmd,strlen(cmd)+6);
-   memmove(cmd+5,cmd,strlen(cmd)+1);
-   memcpy(cmd,"SITE ",5);
-   Job *j=new QuoteJob(Clone(),args->a0(),cmd,
-      output?output:new FDStream(1,"<stdout>"));
-   output=0;
+   j->Bg();
    return j;
 }
 
@@ -1719,7 +1817,7 @@ CMD(help)
 CMD(ver)
 {
    printf(
-      _("Lftp | Version %s | Copyright (c) 1996-1999 Alexander V. Lukyanov\n"),VERSION);
+      _("Lftp | Version %s | Copyright (c) 1996-2000 Alexander V. Lukyanov\n"),VERSION);
    printf(
       _("This is free software with ABSOLUTELY NO WARRANTY. See COPYING for details.\n"));
    printf(
@@ -1784,7 +1882,8 @@ CMD(bookmark)
    if(!strcmp(op,"list"))
    {
       char *list=lftp_bookmarks.Format();
-      Job *j=new CatJob(output,list,strlen(list));
+      Job *j=CopyJob::NewEcho(list,output,args->a0());
+      xfree(list);
       output=0;
       return j;
    }
@@ -1802,8 +1901,9 @@ CMD(bookmark)
 	 if(value==0)
 	 {
 	    value=session->GetConnectURL(flags);
-	    char *a=string_alloca(strlen(value)+2);
-	    strcpy(a,value);
+	    char *a=string_alloca(strlen(value)*3+2);
+	    // encode some more characters, special to CmdExec parser.
+	    url::encode_string(value,a,"&;|\"\\");
 	    if(value[0] && value[strlen(value)-1]!='/')
 	       strcat(a,"/");
 	    value=a;
@@ -1813,6 +1913,11 @@ CMD(bookmark)
 	    eprintf(_("%s: cannot add empty bookmark\n"),args->a0());
 	    return 0;
 	 }
+	 if(strchr(key,' ') || strchr(key,'\t'))
+	 {
+	    eprintf(_("%s: spaces in bookmark name are not allowed\n"),args->a0());
+	    return 0;
+	 }
 	 lftp_bookmarks.Add(key,value);
    	 exit_code=0;
       }
@@ -1869,7 +1974,8 @@ CMD(echo)
    {
       s[len++]='\n'; // replaces \0 char
    }
-   Job *j=new CatJob(output,s,len);
+   Job *j=CopyJob::NewEcho(s,len,output,args->a0());
+   xfree(s);
    output=0;
    return j;
 }
@@ -1880,13 +1986,6 @@ CMD(suspend)
    return 0;
 }
 
-CMD(ftpcopy)
-{
-   Job *j=new FtpCopy(args,session);
-   args=0;
-   return j;
-}
-
 CMD(find)
 {
    const char *path=".";
@@ -1900,6 +1999,11 @@ CMD(find)
 
 CMD(command)
 {
+   if(args->count()<2)
+   {
+      eprintf(_("Usage: %s command args...\n"),args->a0());
+      return 0;
+   }
    args->delarg(0);
    return parent->builtin_restart();
 }
@@ -1963,3 +2067,111 @@ CMD(queue)
 {
    return parent->builtin_queue();
 }
+
+CMD(get1)
+{
+   static struct option get1_options[]=
+   {
+      {"ascii",no_argument,0,'a'},
+      {"region",required_argument,0,'r'},
+      {"continue",no_argument,0,'c'},
+      {"output",required_argument,0,'o'},
+      {0,0,0,0}
+   };
+   int opt;
+   const char *src=0;
+   const char *dst=0;
+   bool cont=false;
+   bool ascii=false;
+
+   args->rewind();
+   while((opt=args->getopt_long("arco:",get1_options,0))!=EOF)
+   {
+      switch(opt)
+      {
+      case 'c':
+	 cont=true;
+	 break;
+      case 'a':
+	 ascii=true;
+	 break;
+      case 'o':
+	 dst=optarg;
+	 break;
+      case 'r':
+	 // FIXME
+	 break;
+      case '?':
+      usage:
+	 eprintf(_("Usage: %s [OPTS] file\n"),args->a0());
+	 return 0;
+      }
+   }
+   src=args->getcurr();
+   if(src==0)
+      goto usage;
+   if(args->getnext()!=0)
+      goto usage;
+
+   if(dst==0 || dst[0]==0)
+   {
+      dst=basename_ptr(src);
+   }
+   else
+   {
+      if(dst[strlen(dst)-1]=='/')
+      {
+	 const char *slash=strrchr(src,'/');
+	 if(slash)
+	    slash++;
+	 else
+	    slash=src;
+	 char *dst1=string_alloca(strlen(dst)+strlen(slash)+1);
+	 strcpy(dst1,dst);
+	 strcat(dst1,slash);
+	 dst=dst1;
+      }
+   }
+
+   ParsedURL dst_url(dst,true);
+
+   if(dst_url.proto==0)
+   {
+      dst=expand_home_relative(dst);
+      // check if dst is a directory.
+      struct stat st;
+      if(stat(dst,&st)!=-1)
+      {
+	 if(S_ISDIR(st.st_mode))
+	 {
+	    const char *slash=strrchr(src,'/');
+	    if(slash)
+	       slash++;
+	    else
+	       slash=src;
+	    char *dst1=string_alloca(strlen(dst)+strlen(slash)+2);
+	    strcpy(dst1,dst);
+	    strcat(dst1,"/");
+	    strcat(dst1,slash);
+	    dst=dst1;
+	 }
+      }
+   }
+
+   FileCopyPeer *src_peer=0;
+   FileCopyPeer *dst_peer=0;
+
+   src_peer=FileCopyPeerFA::New(Clone(),src,FA::RETRIEVE,true);
+
+   if(dst_url.proto==0)
+      dst_peer=FileCopyPeerFDStream::NewPut(dst,cont);
+   else
+      dst_peer=new FileCopyPeerFA(&dst_url,FA::STORE);
+
+   FileCopy *c=FileCopy::New(src_peer,dst_peer,cont);
+
+   if(ascii)
+      c->Ascii();
+
+   return new CopyJob(c,src,args->a0());
+}
