@@ -18,12 +18,13 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: complete.cc,v 1.23 1999/10/13 17:44:44 lav Exp $ */
+/* $Id: complete.cc,v 1.32.2.2 2000/07/31 16:42:20 lav Exp $ */
 
 #include <config.h>
 
 #include <stdio.h>
 #include <sys/types.h>
+#include <sys/stat.h>
 #include <time.h>
 #include "xalloca.h"
 #include "xmalloc.h"
@@ -35,6 +36,7 @@
 #include "LsCache.h"
 #include "complete.h"
 #include "lftp_rl.h"
+#include "url.h"
 
 CDECL_BEGIN
 #include "readline/readline.h"
@@ -163,6 +165,42 @@ static char *array_generator(char *text,int state)
    return NULL;
 }
 
+static bool not_dir(char *f)
+{
+   struct stat st;
+   f=tilde_expand(f);
+   bool res=(stat(f,&st)==-1 || !S_ISDIR(st.st_mode));
+   free(f);
+   return res;
+}
+
+void ignore_non_dirs(char **matches)
+{
+   // filter out non-dirs.
+   int out=1;
+   for(int i=1; matches[i]; i++)
+   {
+      if(!not_dir(matches[i]))
+	 matches[out++]=matches[i];
+      else
+	 free(matches[i]);
+   }
+   matches[out]=0;
+   if(out==1)
+   {
+      // we have only the LCD prefix. Handle it carefully.
+      char *f=matches[0];
+      int len=strlen(f);
+      if((len>2 && f[len-1]=='/') // all files, no dirs.
+      || not_dir(f))		 // or single non dir.
+      {
+	 // all files, no dirs.
+	 free(f);
+	 matches[0]=0;
+      }
+   }
+}
+
 static const char *find_word(const char *p)
 {
    while(*p && isspace(*p))
@@ -185,15 +223,15 @@ static bool copy_word(char *buf,const char *p,int n)
 
 enum completion_type
 {
-   LOCAL, REMOTE_FILE, REMOTE_DIR, BOOKMARK, COMMAND,
+   LOCAL, LOCAL_DIR, REMOTE_FILE, REMOTE_DIR, BOOKMARK, COMMAND,
    STRING_ARRAY, NO_COMPLETION
 };
 
-static completion_type cmd_completion_type(int start)
+static completion_type cmd_completion_type(const char *cmd,int start)
 {
-   const char *cmd=rl_line_buffer;
+   const char *w=find_word(cmd);
 
-   if(find_word(cmd)-cmd == start) // first word is command
+   if(w-cmd == start) // first word is command
       return COMMAND;
 
    // try to guess whether the completion word is remote
@@ -203,11 +241,18 @@ static completion_type cmd_completion_type(int start)
 
    for(;;)
    {
-      const char *w=find_word(cmd);
+      w=find_word(cmd);
       if(w[0]=='!')
 	 shell_cmd=true;
-      if(w[0]=='!' || w[0]=='#')
+      if(w[0]=='#')
+	 return NO_COMPLETION;
+      if(w[0]=='!')
+      {
+	 shell_cmd=quote_glob=true;
 	 return LOCAL;
+      }
+      if(w[0]=='?')  // help
+	 return COMMAND;
       if(w[0]=='(')
       {
 	 cmd=w+1;
@@ -244,7 +289,7 @@ static completion_type cmd_completion_type(int start)
    }
 
    if(!strcmp(buf,"shell"))
-      shell_cmd=true;
+      shell_cmd=quote_glob=true;
    if(!strcmp(buf,"glob")
    || !strcmp(buf,"mget")
    || !strcmp(buf,"mput")
@@ -257,7 +302,6 @@ static completion_type cmd_completion_type(int start)
 
    if(!strcmp(buf,"cat")
    || !strcmp(buf,"ls")
-   || !strcmp(buf,"mget")
    || !strcmp(buf,"more")
    || !strcmp(buf,"mrm")
    || !strcmp(buf,"mv")
@@ -272,9 +316,14 @@ static completion_type cmd_completion_type(int start)
    || !strcmp(buf,"lftp"))
       return BOOKMARK;
 
+   if(!strcmp(buf,"help"))
+      return COMMAND;
+
    bool was_o=false;
    bool was_N=false;
+   bool was_O=false;
    bool second=false;
+   int second_start=-1;
    for(int i=start; i>4; i--)
    {
       if(!isspace(rl_line_buffer[i-1]))
@@ -289,27 +338,48 @@ static completion_type cmd_completion_type(int start)
 	 was_N=true;
 	 break;
       }
+      if(!strncmp(rl_line_buffer+i-3,"-O",2) && isspace(rl_line_buffer[i-4]))
+      {
+	 was_O=true;
+	 break;
+      }
    }
-   const char *w=find_word(cmd);
+   w=find_word(cmd);
    while(*w && !isspace(*w))
-      w++;
+      w++;  // FIXME: handle quotations.
    if(*w)
    {
       w=find_word(w);
+      second_start=w-cmd;
       if(w-cmd==start)	// we complete second word
 	 second=true;
    }
 
    if(!strcmp(buf,"get")
-   || !strcmp(buf,"pget"))
+   || !strcmp(buf,"pget")
+   || !strcmp(buf,"get1"))
+   {
+      if(was_O)
+	 return LOCAL_DIR;
       if(!was_o)
 	 return REMOTE_FILE;
+   }
+   if(!strcmp(buf,"mget"))
+      if(!was_O)
+	 return REMOTE_FILE;
    if(!strcmp(buf,"put"))
       if(was_o)
 	 return REMOTE_FILE;
+   if(!strcmp(buf,"put")
+   || !strcmp(buf,"mput"))
+      if(was_O)
+	 return REMOTE_DIR;
    if(!strcmp(buf,"mirror"))
+   {
+      // FIXME: guess -R and take arg number into account.
       if(!was_N)
-	 return REMOTE_FILE;
+	 return REMOTE_DIR;
+   }
    if(!strcmp(buf,"bookmark"))
    {
       if(second)
@@ -327,12 +397,19 @@ static completion_type cmd_completion_type(int start)
       else
 	 return REMOTE_FILE;
    }
-   if(!strcmp(buf,"glob"))
+   if(!strcmp(buf,"glob")
+   || !strcmp(buf,"command")
+   || !strcmp(buf,"queue"))
    {
       if(second)
 	 return COMMAND;
       else
-	 return REMOTE_FILE;  // FIXME: it would be better to check argv[1].
+      {
+	 // FIXME: infinite alias expansion is possible.
+	 if(second_start>0 && start>second_start && (int)strlen(cmd)>second_start)
+	    return cmd_completion_type(cmd+second_start,start-second_start);
+	 return REMOTE_FILE;
+      }
    }
    if(!strcmp(buf,"cache"))
    {
@@ -345,6 +422,9 @@ static completion_type cmd_completion_type(int start)
 	 return NO_COMPLETION;
    }
 
+   if(!strcmp(buf,"lcd"))
+      return LOCAL_DIR;
+
    return LOCAL;
 }
 
@@ -388,16 +468,19 @@ static bool force_remote=false;
    array of matches, or NULL if there aren't any. */
 static char **lftp_completion (char *text,int start,int end)
 {
+   completion_shell->RestoreCWD();
+
    if(start>end)  // workaround for a bug in readline
       start=end;
 
-   Glob *rg=0;
+   GlobURL *rg=0;
 
    rl_completion_append_character=' ';
+   rl_ignore_some_completions_function=0;
    shell_cmd=false;
    quote_glob=false;
 
-   completion_type type=cmd_completion_type(start);
+   completion_type type=cmd_completion_type(rl_line_buffer,start);
 
    len=end-start;
 
@@ -418,8 +501,15 @@ static char **lftp_completion (char *text,int start,int end)
       generator = array_generator;
       break;
    case LOCAL:
-      if(force_remote)
+   case LOCAL_DIR:
+      if(force_remote || (url::is_url(text) && remote_completion))
+      {
+	 if(type==LOCAL_DIR)
+	    type=REMOTE_DIR;
 	 goto really_remote;
+      }
+      if(type==LOCAL_DIR)
+	 rl_ignore_some_completions_function = (Function*)ignore_non_dirs;
       break;
    case REMOTE_FILE:
    case REMOTE_DIR:
@@ -435,9 +525,12 @@ static char **lftp_completion (char *text,int start,int end)
 
       SignalHook::ResetCount(SIGINT);
       glob_res=NULL;
-      rg=completion_shell->session->MakeGlob(pat);
+      rg=new GlobURL(completion_shell->session,pat);
       if(rg)
       {
+	 rg->glob->NoMatchPeriod();
+	 if(type==REMOTE_DIR)
+	    rg->glob->DirectoriesOnly();
 	 for(;;)
 	 {
 	    SMTask::Schedule();
@@ -474,14 +567,17 @@ static char **lftp_completion (char *text,int start,int end)
    if(rg)
       delete rg;
 
-   free(text);
    if(!matches)
    {
       rl_attempted_completion_over = 1;
-      return 0;
+      goto leave;
    }
+
    if(type==REMOTE_DIR)
       rl_completion_append_character='/';
+
+leave:
+   xfree(text);
    return matches;
 }
 
@@ -591,9 +687,9 @@ backslash_quote (char *string)
 	  *r++ = c;
 	  break;
 	case '~':				/* tilde expansion */
+	case '#':				/* comment char */
 	  if(!shell_cmd)
 	    goto def;
-	case '#':				/* comment char */
 	  if (s == string)
 	    *r++ = '\\';
 	  /* FALLTHROUGH */
@@ -810,9 +906,9 @@ static int skip_double_quoted(char *s, int i)
 
 int lftp_char_is_quoted(char *string,int eindex)
 {
-  int i, pass_next, quoted;
+  int i, pass_next;
 
-  for (i = pass_next = quoted = 0; i <= eindex; i++)
+  for (i = pass_next = 0; i <= eindex; i++)
     {
       if (pass_next)
         {
@@ -867,8 +963,8 @@ int   lftp_rl_getc(FILE *file)
       SMTask::Block();
       if(SignalHook::GetCount(SIGINT)>0)
       {
-	 rl_kill_text(0,rl_end);
-	 rl_line_buffer[0]=0;
+	 if(rl_line_buffer && rl_end>0)
+	    rl_kill_full_line(0,0);
 	 return '\n';
       }
    }
