@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: ftpclass.h,v 1.32.2.1 2000/02/18 11:35:16 lav Exp $ */
+/* $Id: ftpclass.h,v 1.56 2000/03/21 14:27:02 lav Exp $ */
 
 #ifndef FTPCLASS_H
 #define FTPCLASS_H
@@ -27,7 +27,6 @@
 #include <time.h>
 
 #include "NetAccess.h"
-/*#include "ResMgr.h"*/
 
 class Ftp : public NetAccess
 {
@@ -93,6 +92,7 @@ class Ftp : public NetAccess
       CHECK_MDTM_OPT,	// check response for MDTM and save size to *opt_date
       CHECK_PASV,	// check response for PASV and save address
       CHECK_EPSV,	// check response for EPSV and save address
+      CHECK_PORT,	// check response for PORT or EPRT
       CHECK_FILE_ACCESS,// generic check for file access
       CHECK_PWD,	// check response for PWD and save it to home
       CHECK_RNFR,	// check RNFR and issue RNTO
@@ -119,13 +119,15 @@ class Ftp : public NetAccess
    int	 RQ_head;
    int	 RQ_tail;
 
-   int	 multiline_code;
+   int	 multiline_code;// the code of multiline response.
+   int	 sync_wait;	// number of commands in flight.
 
    void	 LogResp(const char *line);
 
    void  AddResp(int exp,int fail,check_case_t ck=CHECK_NONE,bool log=false);
    void  SetRespPath(const char *p);
    int   CheckResp(int resp);
+   int	 ReplyLogPriority(int code);
    void  PopResp();
    void	 EmptyRespQueue();
    void	 CloseRespQueue(); // treat responses on Close()
@@ -169,8 +171,11 @@ class Ftp : public NetAccess
    long	    nop_offset;
    int	    nop_count;
 
+   time_t   stat_time;
+
    char	 *result;
    int	 result_size;
+   void  FreeResult();
 
    void	 DataAbort();
    void  DataClose();
@@ -180,17 +185,16 @@ class Ftp : public NetAccess
 
    void	 SwitchToState(automate_state);
 
-   void  SendCmd(const char *cmd);
-   void	 FlushSendQueue(bool all=false);
+   void  SendCmd(const char *cmd,int len=-1);
+   void  SendCmd2(const char *cmd,const char *f);
+   void  SendUrgentCmd(const char *cmd);
+   int	 FlushSendQueue(bool all=false);
 
-   void	 ReceiveResp();
+   int	 ReceiveResp();
 	 // If a response is received, it checks it for accordance with
 	 // response_queue and switch to a state if necessary
 
-   int	 Poll(int fd,int ev) { return FileAccess::Poll(fd,ev); }
-   int	 Poll();
-
-   int	 AbsolutePath(const char *p);
+   bool	 AbsolutePath(const char *p);
 
    char	 *send_cmd_buffer;
    int   send_cmd_count;   // in buffer
@@ -199,6 +203,8 @@ class Ftp : public NetAccess
 
    void EmptySendQueue();
 
+   void MoveConnectionHere(Ftp *o);
+
    automate_state state;
 
    char	 *anon_user;
@@ -208,6 +214,13 @@ class Ftp : public NetAccess
 
    int	 flags;
 
+   bool  dos_path;
+   bool  vms_path;
+   bool  mdtm_supported;
+   bool  size_supported;
+   bool  site_chmod_supported;
+   long	 last_rest;
+
    int	 StateToError();
 
    int	 addr_received;	// state of PASV
@@ -217,17 +230,18 @@ class Ftp : public NetAccess
    bool  verify_data_address:1;
    bool  verify_data_port:1;
    bool	 rest_list:1;
+   char  *list_options;
 
-   void	 GetBetterConnection(int level);
+   void	 GetBetterConnection(int level,int count);
    bool  SameConnection(const Ftp *o);
 
    int	 nop_interval;
 
    void set_idle_start()
       {
-	 time(&idle_start);
+	 idle_start=now;
 	 if(control_sock!=-1 && idle>0)
-	    block+=TimeOut(idle*1000);
+	    TimeoutS(idle);
       }
 
    char *skey_pass;
@@ -237,50 +251,55 @@ class Ftp : public NetAccess
 
    bool data_address_ok(sockaddr_u *d=0,bool verify_this_data_port=true);
 
+public:
    enum copy_mode_t { COPY_NONE, COPY_SOURCE, COPY_DEST };
+private:
    copy_mode_t copy_mode;
    sockaddr_u copy_addr;
    bool copy_addr_valid;
    bool copy_passive;
-   friend class FtpCopy;
+   bool	copy_done;
+   bool	copy_connection_open;
+   bool copy_allow_store;
+
+   bool use_stat;
+   int  stat_interval;
 
    const char *encode_eprt(sockaddr_u *);
 
+protected:
+   ~Ftp();
+
 public:
    static void ClassInit();
 
    Ftp();
    Ftp(const Ftp *);
-   ~Ftp();
 
    const char *GetProto() { return "ftp"; }
 
    FileAccess *Clone() { return new Ftp(this); }
-   static FileAccess *New() { return new Ftp(); }
+   static FileAccess *New();
 
    bool	 SameLocationAs(FileAccess *);
    bool	 SameSiteAs(FileAccess *);
 
-   bool	 IsConnected()
+   int IsConnected()
    {
       if(control_sock==-1)
-	 return false;
-      Do();
-      return(control_sock!=-1 && state!=CONNECTING_STATE);
-   }
-   bool	 IsBetterThan(FileAccess *fa)
-   {
-      return(SameProtoAs(fa)
-	     && this->IsConnected() && !((Ftp*)fa)->IsConnected());
+	 return 0;
+      if(state==CONNECTING_STATE)
+	 return 1;
+      return 2;
    }
 
    void	 Connect(const char *h,const char *p);
-   void	 ConnectVerify();
 
    int   Read(void *buf,int size);
    int   Write(const void *buf,int size);
    int   Buffered();
    void  Close();
+   bool	 IOReady();
 
 	 // When you are putting a file, call SendEOT to terminate
 	 // transfer and then call StoreStatus until OK or error.
@@ -302,7 +321,6 @@ public:
    enum flag_mask
    {
       SYNC_MODE=1,
-      SYNC_WAIT=2,
       NOREST_MODE=4,
       IO_FLAG=8,
       DOSISH_PATH=16,
@@ -311,16 +329,6 @@ public:
       MODES_MASK=SYNC_MODE|PASSIVE_MODE|DOSISH_PATH
    };
 
-   void CopyOptions(FileAccess *fa)
-   {
-      if(SameProtoAs(fa))
-      {
-	 flags=(flags&~MODES_MASK)|(((Ftp*)fa)->flags&MODES_MASK);
-	 if(((Ftp*)fa)->try_time==0)
-	    try_time=0;
-      }
-   }
-
    void Reconfig(const char *name=0);
    void Cleanup();
    void CleanupThis();
@@ -328,6 +336,40 @@ public:
    ListInfo *MakeListInfo();
    Glob *MakeGlob(const char *pattern);
    DirList *MakeDirList(ArgV *args);
+
+   void SetCopyMode(copy_mode_t cm,bool rp,int rnum,time_t tt)
+      {
+	 copy_mode=cm;
+	 copy_passive=rp;
+	 retries=rnum;
+	 try_time=tt;
+      }
+   bool SetCopyAddress(Ftp *o)
+      {
+	 if(copy_addr_valid || !o->copy_addr_valid)
+	    return false;
+	 memcpy(&copy_addr,&o->copy_addr,sizeof(copy_addr));
+	 copy_addr_valid=true;
+	 return true;
+      }
+   bool CopyFailed() { return state==COPY_FAILED; }
+   bool RestartFailed() { return flags&NOREST_MODE; }
+   bool IsPassive() { return flags&PASSIVE_MODE; }
+   void SetPos(long p) { pos=real_pos=p; }
+   bool IsCopyPassive() { return copy_passive; }
+   void CopyAllowStore()
+      {
+	 SendCmd2("STOR",file);
+	 AddResp(RESP_TRANSFER_OK,STORE_FAILED_STATE,CHECK_TRANSFER);
+	 copy_allow_store=true;
+      }
+   bool CopyStoreAllowed() { return copy_allow_store; }
+   bool CopyIsReadyForStore()
+      {
+	 if(copy_mode==COPY_SOURCE)
+	    return copy_addr_valid && RespQueueSize()==1;
+	 return state==WAITING_STATE && RespQueueSize()==0;
+      }
 };
 
 #endif /* FTPCLASS_H */
