@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1998 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: history.cc,v 1.3 1999/05/22 12:37:51 lav Exp $ */
+/* $Id: history.cc,v 1.6 2000/01/31 11:12:44 lav Exp $ */
 
 #include <config.h>
 #include <sys/types.h>
@@ -28,6 +28,7 @@
 #include <stdio.h>
 #include "history.h"
 #include "xalloca.h"
+#include "url.h"
 
 #define super KeyValueDB
 
@@ -39,7 +40,7 @@ History::History()
 
    const char *home=getenv("HOME");
    if(home==0)
-      home=".";
+      home="";
    int home_len=strlen(home);
    const char *add="/.lftp/cwd_history";
    file=(char*)xmalloc(home_len+strlen(add)+1);
@@ -56,11 +57,23 @@ History::~History()
 
 const char *History::extract_url(const char *res)
 {
+   static char *buf;
+   static int buf_len;
+
    const char *url=strchr(res,':');
    if(url)
-      return url+1;
+      url++;
    else
-      return res;
+      url=res;
+
+   int len=strlen(url);
+   if(len>=buf_len);
+      buf=(char*)xrealloc(buf,buf_len=len+64);
+
+   strcpy(buf,url);
+   url::decode_string(buf);
+
+   return buf;
 }
 
 time_t History::extract_stamp(const char *res)
@@ -70,7 +83,7 @@ time_t History::extract_stamp(const char *res)
 
 const char *History::Lookup(FileAccess *s)
 {
-   const char *url=s->GetConnectURL(s->NO_PATH);
+   const char *url=s->GetConnectURL(s->NO_PATH|s->NO_PASSWORD);
    if(!url)
       return 0;
    const char *res=super::Lookup(url);
@@ -131,9 +144,10 @@ void History::Set(FileAccess *s,const char *cwd)
 {
    if(cwd==0 || !strcmp(cwd,"~") || s->GetHostName()==0)
       return;
-   char *res=(char*)alloca(32+strlen(cwd)+1);
-   sprintf(res,"%lu:%s",(unsigned long)time(0),cwd);
-   super::Add(s->GetConnectURL(s->NO_PATH),res);
+   char *res=(char*)alloca(32+strlen(cwd)*3+1);
+   sprintf(res,"%lu:",(unsigned long)time(0));
+   url::encode_string(cwd,res+strlen(res),URL_PATH_UNSAFE);
+   super::Add(s->GetConnectURL(s->NO_PATH|s->NO_PASSWORD),res);
 }
 
 void History::Save()
