@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1998 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: lftp.cc,v 1.28 1999/10/05 15:23:37 lav Exp $ */
+/* $Id: lftp.cc,v 1.35 2000/04/12 18:51:40 lav Exp $ */
 
 #include <config.h>
 
@@ -114,7 +114,7 @@ public:
       ::completion_shell=exec;
       ::remote_completion=exec->remote_completion;
 
-      const char *cmd_buf;
+      char *cmd_buf;
       if(tty)
       {
 	 if(ctty) // controlling terminal
@@ -123,7 +123,7 @@ public:
 	    if(term_pg!=(pid_t)-1 && getpgrp()!=term_pg)
 	    {
 	       // looks like we are in background. Can't read from tty
-	       exec->block+=TimeOut(500);
+	       exec->Timeout(500);
 	       return "";
 	    }
 	 }
@@ -145,7 +145,7 @@ public:
 		  /* If there was an error, return nothing. */
 		  if (expanded < 0 || expanded == 2)	/* 2 == print only */
 		  {
-		     exec->block+=NoWait();  // and retry immediately
+		     exec->Timeout(0);  // and retry immediately
 		     xfree(history_value);
 		     return "";
 		  }
@@ -158,19 +158,30 @@ public:
 	 }
 	 else if(cmd_buf==0 && exec->interactive)
 	    puts("exit");
+	 xmalloc_register_block(cmd_buf);
       }
       else
       {
-	 to_free=readline_from_file(stdin);
-	 cmd_buf=to_free;
+	 if(exec->interactive)
+	 {
+	    while(*prompt)
+	    {
+	       char ch=*prompt++;
+	       if(ch!=1 && ch!=2)
+		  putchar(ch);
+	    }
+	    fflush(stdout);
+	 }
+	 cmd_buf=readline_from_file(stdin);
       }
+      to_free=cmd_buf;
 
       ::completion_shell=0;
 
       if(cmd_buf &&
 	    (*cmd_buf==0 || cmd_buf[strlen(cmd_buf)-1]!='\n'))
       {
-	 exec->block+=NoWait();
+	 exec->Timeout(0);
 	 add_newline=true;
       }
       return cmd_buf;
@@ -205,22 +216,23 @@ static void move_to_background()
       SignalHook::Ignore(SIGTSTP);
 
       const char *home=getenv("HOME");
-      if(!home) home=".";
-
-      char *log=(char*)alloca(strlen(home)+1+9+1);
-      sprintf(log,"%s/.lftp",home);
-      if(access(log,F_OK)==-1)
-	 strcat(log,"_log");
-      else
-	 strcat(log,"/log");
-
-      int fd=open(log,O_WRONLY|O_APPEND|O_CREAT,0600);
-      if(fd>=0)
+      if(home)
       {
-	 dup2(fd,1); // stdout
-	 dup2(fd,2); // stderr
-	 if(fd!=1 && fd!=2)
-	    close(fd);
+	 char *log=(char*)alloca(strlen(home)+1+9+1);
+	 sprintf(log,"%s/.lftp",home);
+	 if(access(log,F_OK)==-1)
+	    strcat(log,"_log");
+	 else
+	    strcat(log,"/log");
+
+	 int fd=open(log,O_WRONLY|O_APPEND|O_CREAT,0600);
+	 if(fd>=0)
+	 {
+	    dup2(fd,1); // stdout
+	    dup2(fd,2); // stderr
+	    if(fd!=1 && fd!=2)
+	       close(fd);
+	 }
       }
       close(0);	  // close stdin.
       open("/dev/null",O_RDONLY); // reopen it, just in case.
@@ -293,34 +305,35 @@ int   main(int argc,char **argv)
    Ftp::ClassInit();
 #endif
 
-   const char *home=getenv("HOME");
-   if(!home) home=".";
-
-   top_exec=new CmdExec(new DummyProto());
-   top_exec->jobno=-1;
-   top_exec->status_line=new StatusLine(1);
-   Log::global=new Log();
-   Log::global->SetCB(tty_clear);
-
    lftp_readline_init();
 
    hook_signals();
 
+   top_exec=new CmdExec(new DummyProto());
+   top_exec->status_line=new StatusLine(1);
+   Log::global=new Log();
+   Log::global->SetCB(tty_clear);
+
    source_if_exist(top_exec,SYSCONFDIR"/lftp.conf");
 
-   char	 *rc=(char*)alloca(strlen(home)+9+1);
+   const char *home=getenv("HOME");
+   if(home)
+   {
+      char *rc=(char*)alloca(strlen(home)+9+1);
 
-   // create lftp own directory
-   sprintf(rc,"%s/.lftp",home);
-   mkdir(rc,0755);
+      // create lftp own directory
+      sprintf(rc,"%s/.lftp",home);
+      mkdir(rc,0755);
 
-   sprintf(rc,"%s/.lftprc",home);
-   source_if_exist(top_exec,rc);
-   sprintf(rc,"%s/.lftp/rc",home);
-   source_if_exist(top_exec,rc);
+      sprintf(rc,"%s/.lftprc",home);
+      source_if_exist(top_exec,rc);
+      sprintf(rc,"%s/.lftp/rc",home);
+      source_if_exist(top_exec,rc);
+   }
 
    WaitDone(top_exec);
 
+   top_exec->SetTopLevel();
    top_exec->Fg();
 
    ArgV *args=new ArgV(argc,argv);
@@ -329,7 +342,7 @@ int   main(int argc,char **argv)
    {
       char *line=args->CombineQuoted();
       lftp_add_history_nodups(line);
-      free(line);
+      xfree(line);
    }
    lftp_feeder=new ReadlineFeeder;
    top_exec->ExecParsed(args);
@@ -343,7 +356,7 @@ int   main(int argc,char **argv)
 
    if(Job::NumberOfJobs()>0)
    {
-      top_exec->interactive=false;
+      top_exec->SetInteractive(false);
       move_to_background();
    }
    return top_exec->ExitCode();
