@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: misc.h,v 1.5 1999/09/20 09:29:01 lav Exp $ */
+/* $Id: misc.h,v 1.12 2000/03/18 23:46:15 lav Exp $ */
 
 #ifndef MISC_H
 #define MISC_H
@@ -34,9 +34,17 @@ char *basename_ptr(char *f)
    return (char*)basename_ptr((const char *)f);
 }
 
-// clues file to dir; returns pointer to static storage
+// glues file to dir; returns pointer to static storage
 const char *dir_file(const char *dir,const char *file);
 
+// glues file to given url; returns pointer to static storage
+const char *url_file(const char *url,const char *file);
+
+const char *output_file_name(const char *src,const char *dst,bool dst_local,
+			     const char *dst_base,bool make_dirs);
+
+const char *squeeze_file_name(const char *name,int w);
+
 // mkdir -p
 int   create_directories(char *);
 
@@ -47,8 +55,18 @@ void  truncate_file_tree(const char *dir);
 // returns 0 on error.
 char *xgetcwd();
 
+int percent(long offset,long size);
+
+#define find_char(buf,len,ch) ((const char *)memchr(buf,ch,len))
+
 #define MINUTE (60)
 #define HOUR   (60*MINUTE)
 #define DAY    (24*HOUR)
 
+int parse_month(const char *);
+int parse_perms(const char *);
+const char *format_perms(int p);
+int parse_year_or_time(const char *year_or_time,int *year,int *hour,int *minute);
+int guess_year(int month,int day,int hour=0,int minute=0);
+
 #endif // MISC_H
