@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: pgetJob.h,v 1.2 1999/05/20 10:20:38 lav Exp $ */
+/* $Id: pgetJob.h,v 1.5 2000/04/20 16:45:50 lav Exp $ */
 
 #ifndef PGETJOB_H
 #define PGETJOB_H
@@ -27,25 +27,15 @@
 
 class pgetJob : public GetJob
 {
-   class ChunkXfer : public XferJob
+   class ChunkXfer : public CopyJob
    {
       friend class pgetJob;
 
       long start;
       long limit;
-      char *error_text;
-      bool error:1;
-      bool done:1;
-      bool no_parallel:1;
-      FDStream *local;
 
-      ChunkXfer(FileAccess *session,const char *remote,FDStream *local,
-	        long start,long limit);
+      ChunkXfer(FileCopy *c,const char *remote,long start,long limit);
       ~ChunkXfer();
-
-      int Do();
-      int Done() { return done; }
-      void PrintStatus(int);
    };
 
    ChunkXfer   **chunks;
@@ -59,6 +49,8 @@ class pgetJob : public GetJob
 
    void free_chunks();
 
+   long total_eta;
+
 protected:
    void	 NextFile();
 
@@ -77,11 +69,14 @@ public:
       total_xfer_rate=0;
       no_parallel=false;
       max_chunks=5;
-      need_seek=true; // seek before writing
+      total_eta=-1;
    }
    ~pgetJob();
 
    void SetMaxConn(int n) { max_chunks=n; }
+
+   pgetJob::ChunkXfer *NewChunk(FileAccess *session,const char *remote,
+					  FDStream *local,long start,long limit);
 };
 
 #endif//PGETJOB_H
