@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1999 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1999-2000 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: resource.cc,v 1.11 1999/10/22 15:26:18 lav Exp $ */
+/* $Id: resource.cc,v 1.27 2000/04/29 17:21:13 lav Exp $ */
 
 #include <config.h>
 
@@ -86,31 +86,48 @@ static const char *FtpDefaultAnonPass()
 static ResDecl resources[]={
    ResDecl  ("ftp:anon-pass",	       "-lftp@",0,0),
    ResDecl  ("ftp:anon-user",	       "anonymous",0,0),
+   ResDecl  ("ftp:fxp-passive-source", "no",	ResMgr::BoolValidate,0),
+   ResDecl  ("ftp:list-options",       "",0,0),
    ResDecl  ("ftp:nop-interval",       "120",	ResMgr::UNumberValidate,0),
-   ResDecl  ("ftp:passive-mode",       "off",   ResMgr::BoolValidate,0),
+   ResDecl  ("ftp:passive-mode",       "on",    ResMgr::BoolValidate,0),
+   ResDecl  ("ftp:port-range",	       "full",	ResMgr::RangeValidate,0),
    ResDecl  ("ftp:proxy",	       "",	FtpProxyValidate,0),
    ResDecl  ("ftp:rest-list",	       "no",	ResMgr::BoolValidate,0),
+   ResDecl  ("ftp:rest-stor",	       "yes",	ResMgr::BoolValidate,0),
    ResDecl  ("ftp:skey-allow",	       "yes",   ResMgr::BoolValidate,0),
    ResDecl  ("ftp:skey-force",	       "no",    ResMgr::BoolValidate,0),
+   ResDecl  ("ftp:stat-interval",      "1",	ResMgr::UNumberValidate,0),
    ResDecl  ("ftp:sync-mode",	       "on",    ResMgr::BoolValidate,0),
+   ResDecl  ("ftp:use-abor",	       "yes",   ResMgr::BoolValidate,0),
+   ResDecl  ("ftp:use-fxp",	       "yes",   ResMgr::BoolValidate,0),
+   ResDecl  ("ftp:use-stat",	       "yes",   ResMgr::BoolValidate,0),
    ResDecl  ("ftp:verify-address",     "no",    ResMgr::BoolValidate,0),
    ResDecl  ("ftp:verify-port",	       "no",    ResMgr::BoolValidate,0),
+   ResDecl  ("ftp:web-mode",	       "off",	ResMgr::BoolValidate,0),
    ResDecl  ("hftp:cache",	       "yes",	ResMgr::BoolValidate,0),
    ResDecl  ("hftp:proxy",	       "",	HttpProxyValidate,0),
    ResDecl  ("hftp:use-head",	       "yes",	ResMgr::BoolValidate,0),
    ResDecl  ("http:cache",	       "yes",	ResMgr::BoolValidate,0),
    ResDecl  ("http:proxy",	       "",	HttpProxyValidate,0),
+   ResDecl  ("http:user-agent",	       PACKAGE"/"VERSION,0,0),
+   ResDecl  ("http:cookie",	       "",	0,0),
    ResDecl  ("net:idle",	       "180",   ResMgr::UNumberValidate,0),
    ResDecl  ("net:limit-max",	       "0",	ResMgr::UNumberValidate,0),
    ResDecl  ("net:limit-rate",	       "0",	ResMgr::UNumberValidate,0),
    ResDecl  ("net:limit-total-max",    "0",	ResMgr::UNumberValidate,0),
    ResDecl  ("net:limit-total-rate",   "0",	ResMgr::UNumberValidate,0),
    ResDecl  ("net:max-retries",	       "0",	ResMgr::UNumberValidate,0),
-   ResDecl  ("net:reconnect-interval", "30",	ResMgr::UNumberValidate,0),
-   ResDecl  ("net:relookup-always",    "off",   ResMgr::BoolValidate,0),
+   ResDecl  ("net:no-proxy",	       "",	0,0),
+   ResDecl  ("net:reconnect-interval-base","30",ResMgr::UNumberValidate,0),
+   ResDecl  ("net:reconnect-interval-multiplier","1.5",ResMgr::FloatValidate,0),
+   ResDecl  ("net:reconnect-interval-max","600",ResMgr::UNumberValidate,0),
    ResDecl  ("net:socket-buffer",      "0",	ResMgr::UNumberValidate,0),
    ResDecl  ("net:socket-maxseg",      "0",	ResMgr::UNumberValidate,0),
    ResDecl  ("net:timeout",	       "300",   ResMgr::UNumberValidate,0),
+   ResDecl  ("net:connection-limit",   "0",	ResMgr::UNumberValidate,0),
+   ResDecl  ("net:connection-takeover","yes",	ResMgr::BoolValidate,0),
+   ResDecl  ("mirror:time-precision",  "1s",    ResMgr::TimeIntervalValidate,0),
+   ResDecl  ("mirror:loose-time-precision","24h",ResMgr::TimeIntervalValidate,0),
 };
 
 void ResMgr::ClassInit()
@@ -135,6 +152,10 @@ void ResMgr::ClassInit()
 	 Set("hftp:proxy",0,ftp_proxy);
    }
 
+   const char *no_proxy=getenv("no_proxy");
+   if(no_proxy)
+      Set("net:no-proxy",0,no_proxy);
+
 #if INET6
    // check if ipv6 is really supported
    int s=socket(AF_INET6,SOCK_STREAM,IPPROTO_TCP);
