@@ -18,13 +18,14 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: rmJob.cc,v 1.4 1999/05/26 14:04:10 lav Exp $ */
+/* $Id: rmJob.cc,v 1.5 2000/01/22 16:27:31 lav Exp $ */
 
 #include <config.h>
 
 #include <errno.h>
 #include "rmJob.h"
 #include "plural.h"
+#include "url.h"
 
 rmJob::rmJob(FileAccess *s,ArgV *a) : TreatFileJob(s,a)
 {
@@ -57,5 +58,11 @@ void  rmJob::SayFinal()
 
 void rmJob::TreatCurrent()
 {
-   session->Open(curr,mode);
+   ParsedURL u(curr,true);
+   if(u.proto)
+      url_session=FA::New(&u);
+   if(url_session)
+      url_session->Open(u.path,mode);
+   else
+      session->Open(curr,mode);
 }
