@@ -24,6 +24,7 @@
 #include <ctype.h>
 #include "xalloca.h"
 #include "url.h"
+#include "ascii_ctype.h"
 
 /*
    URL -> [PROTO://]CONNECT[[:]/PATH]
@@ -32,7 +33,7 @@
 
 ParsedURL::ParsedURL(const char *url,bool proto_required)
 {
-   memory=(char*)xmalloc(strlen(url)*2+1);
+   memory=(char*)xmalloc(strlen(url)*2+20+1);
    strcpy(memory,url);
 
    proto=0;
@@ -44,7 +45,7 @@ ParsedURL::ParsedURL(const char *url,bool proto_required)
 
    char *base=memory;
    char *scan=base;
-   while(isalpha(*scan))
+   while(is_ascii_alpha(*scan))
       scan++;
    if(scan[0]==':' && scan[1]=='/' && scan[2]=='/')
    {
@@ -83,7 +84,27 @@ ParsedURL::ParsedURL(const char *url,bool proto_required)
       if(scan-1>=base && scan[-1]==':')
 	 scan[-1]=0;
       if(scan[1]!='~')
+      {
 	 memmove(scan+1,scan,strlen(scan)+1);
+	 if(!xstrcmp(proto,"ftp") || !xstrcmp(proto,"hftp"))
+	 {
+	    // special handling for ftp protocol.
+	    if(!strncasecmp(scan+2,"%2F",3))
+	    {
+	       char *p=scan+5+(scan[5]=='/');
+	       memmove(scan+2,p,strlen(p)+1);
+	    }
+	    else
+	    {
+	       if(!(is_ascii_alpha(scan[2]) && scan[3]==':' && scan[4]=='/'))
+	       {
+		  memmove(scan+3,scan+2,strlen(scan+2)+1);
+		  scan[1]='~';
+		  scan[2]='/';
+	       }
+	    }
+	 }
+      }
       *scan++=0;
       path=scan;
    }
@@ -134,6 +155,73 @@ decode:
    url::decode_string(path);
 }
 
+int url::path_index(const char *base)
+{
+   const char *scan=base;
+   while(is_ascii_alpha(*scan))
+      scan++;
+   if(scan[0]==':' && scan[1]=='/' && scan[2]=='/')
+   {
+      // found protocol
+      const char *slash=strchr(scan+3,'/');
+      if(slash)
+	 return slash-base;
+      return strlen(base);
+   }
+   else if(scan[0]==':' && !strncmp(base,"file:",5))
+   {
+      // special form for file protocol
+      return scan+1-base;
+   }
+   return 0;
+}
+
+void ParsedURL::Combine(char *url,const char *home)
+{
+   bool is_file=!strcmp(proto,"file");
+   bool is_ftp=(!strcmp(proto,"ftp") || !strcmp(proto,"hftp"));
+
+   strcpy(url,proto);
+   strcat(url,is_file?":":"://");
+   if(user && !is_file)
+   {
+      url::encode_string(user,url+strlen(url),URL_USER_UNSAFE);
+      if(pass)
+      {
+	 strcat(url,":");
+	 url::encode_string(pass,url+strlen(url),URL_PASS_UNSAFE);
+      }
+      strcat(url,"@");
+   }
+   if(host && !is_file)
+      url::encode_string(host,url+strlen(url),URL_HOST_UNSAFE);
+   if(port && !is_file)
+   {
+      strcat(url,":");
+      url::encode_string(port,url+strlen(url),URL_PORT_UNSAFE);
+   }
+   if(path==0)
+      return;
+   if(strcmp(path,"~"))
+   {
+      if(path[0]!='/' && !is_file) // e.g. ~/path
+	 strcat(url,"/");
+      int p_offset=0;
+      if(is_ftp)
+      {
+	 // some cruft for ftp urls...
+	 if(path[0]=='/' && xstrcmp(home,"/"))
+	 {
+	    strcat(url,"/%2F");
+	    p_offset=1;
+	 }
+	 else if(path[0]=='~' && path[1]=='/')
+	    p_offset=2;
+      }
+      url::encode_string(path+p_offset,url+strlen(url),URL_PATH_UNSAFE);
+   }
+}
+
 void url::decode_string(char *p)
 {
    if(!p)
@@ -158,7 +246,7 @@ void url::decode_string(char *p)
 
 /* Encodes the unsafe characters (listed in URL_UNSAFE) in a given
    string, returning a malloc-ed %XX encoded string.  */
-#define need_quote(c) (!unsafe || iscntrl((unsigned char)c) || strchr(unsafe,c))
+#define need_quote(c) (!unsafe || iscntrl((unsigned char)(c)) || strchr(unsafe,(c)))
 char *url::encode_string (const char *s,char *res,const char *unsafe)
 {
   char *p;
