@@ -18,12 +18,14 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: xmalloc.cc,v 1.4 1999/05/22 12:41:27 lav Exp $ */
+/* $Id: xmalloc.cc,v 1.7.2.1 2000/06/08 23:08:37 lav Exp $ */
 
 #include <config.h>
 #include <stdio.h>
 #include "xmalloc.h"
 
+static int memory_count=0;
+
 static void memory_error_and_abort(const char *fname,size_t size)
 {
    fprintf(stderr,"%s: out of virtual memory when trying to get %lu bytes\n",
@@ -38,8 +40,9 @@ void *xmalloc (size_t bytes)
    void *temp=(void*)malloc(bytes);
    if(temp==0)
       memory_error_and_abort("xmalloc",bytes);
+   memory_count++;
 #ifdef MEM_DEBUG
-   printf("xmalloc %p %lu\n",temp,(long)bytes);
+   printf("xmalloc %p %lu (count=%d)\n",temp,(long)bytes,memory_count);
 #endif
    return(temp);
 }
@@ -51,28 +54,68 @@ void *xrealloc(void *pointer,size_t bytes)
       return 0;
    if(bytes==0)
    {
+      memory_count--;
       free(pointer);
-      return 0;
+      temp=0;
+      goto leave;
    }
    if(pointer==0)
+   {
       temp=(void*)malloc(bytes);
+      memory_count++;
+   }
    else
       temp=(void*)realloc(pointer,bytes);
    if(temp==0)
       memory_error_and_abort ("xrealloc",bytes);
+leave:
 #ifdef MEM_DEBUG
-   printf("xrealloc %p %p %lu\n",pointer,temp,(long)bytes);
+   printf("xrealloc %p %p %lu (count=%d)\n",pointer,temp,(long)bytes,memory_count);
 #endif
    return(temp);
 }
 
 void xfree(void *p)
 {
-   if(p)
-   {
+   if(!p)
+      return;
 #ifdef MEM_DEBUG
-      printf("xfree %p\n",p);
+   printf("xfree %p (count=%d)\n",p,memory_count);
 #endif
-      free(p);
-   }
+   memory_count--;
+   free(p);
 }
+
+char *xstrdup(const char *s,int spare)
+{
+   if(!s)
+      return 0;
+#ifdef MEM_DEBUG
+   printf("xstrdup \"%s\"\n",s);
+#endif
+   int len=strlen(s)+1;
+   char *mem=(char*)xmalloc(len+spare);
+   memcpy(mem,s,len);
+   return mem;
+}
+
+void xmalloc_register_block(void *b)
+{
+   if(!b)
+      return;
+   memory_count++;
+#ifdef MEM_DEBUG
+   printf("xmalloc %p (count=%d)\n",b,memory_count);
+#endif
+}
+
+#ifdef MEM_DEBUG
+void *__builtin_new(size_t s)
+{
+   return xmalloc(s);
+}
+void __builtin_delete(void *p)
+{
+   xfree(p);
+}
+#endif
