@@ -19,7 +19,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
-/* $Id: xmalloc.h,v 1.5 1999/05/22 12:41:27 lav Exp $ */
+/* $Id: xmalloc.h,v 1.7.2.1 2000/06/08 23:08:37 lav Exp $ */
 
 #ifndef XMALLOC_H
 #define XMALLOC_H
@@ -34,11 +34,7 @@
 
 void *xmalloc(size_t);
 void *xrealloc(void *,size_t);
-static inline char *xstrdup(const char *s)
-{
-   if(!s) return 0;
-   return strcpy((char*)xmalloc(strlen(s)+1),s);
-}
+char *xstrdup(const char *s,int spare=0);
 #define alloca_strdup(s) ((s)?strcpy((char*)alloca(strlen((s))+1),(s)):0)
 
 static inline void *xmemdup(const void *m,int len)
@@ -50,4 +46,6 @@ static inline void *xmemdup(const void *m,int len)
 }
 void xfree(void *p);
 
+void xmalloc_register_block(void *);
+
 #endif /* XMALLOC_H */
